/*
 * lair2.cpp
 *
 * Copyright (C) 2003 Paul Blagay
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// Dragon's Lair 2  / Space Ace '91 Driver
// Coded by Paul Blagay
// Thanks to Robert DiNapoli
// Fabrice Frances & MAME team for I86 core

#ifdef WIN32
#pragma warning (disable:4100) // disable warning about unreferenced parameter
#endif

// Win32 doesn't use strcasecmp, it uses stricmp (lame)
#ifdef WIN32
#define strcasecmp stricmp
#endif

#ifdef _XBOX
#include "../xbox/xbox_grafx.h"
#include "../cpu/mamewrap.h"
#endif


#include <stdio.h> 
#include "lair2.h"
#include "../ldp-out/ldp.h"
#include "../ldp-in/ldp1000.h"
#include "../ldp-in/vp380.h"
#include "../io/conout.h"
#include "../io/error.h"
#include "../io/serial.h"	// for controlling a real LDP-1450
#include "../sound/sound.h"
#include "../cpu/cpu.h"
#ifndef _XBOX
#include "../cpu/x86/i86intf.h"
#else
#include "../cpu/i86intf.h"
#endif

#ifdef _XBOX
#include "../xbox/daphne_xbox.h"
#include "../xbox/xfile.h"
#include "../xbox/gamepad.h"
#endif

#ifdef DEBUG
#include "../io/numstr.h"
bool g_bVerbose = false;	// for debugging only
#endif

// Dragons Lair 2: Sound Samples
int g_sound_map[] = { 0, S_DL2_WARBLE, S_DL2_ERROR,
					  S_DL2_GOOD, S_DL2_BAD, S_DL2_TIC, S_DL2_TOC,
					  0,0,S_DL2_WARN,S_DL2_COIN1, S_DL2_COIN2,
					  S_DL2_COIN3, S_DL2_COIN4 };

int lair2_irq_callback(int irqline);

////////////////////////////////////////////////////////////////////////////
// Rom structs

// Dragons Lair 2   v2.11
static struct rom_def g_lair2_211_roms[] =
{
	{ "dl2_211.bin", NULL, 0, 0x10000, 0x9f2660a3 },
	{ NULL }
};

// Dragons Lair 2   v3.00
static struct rom_def g_lair2_300_roms[] =
{
	{ "dl2_300.bin", NULL, 0, 0x10000, 0xdec4f2e3 },
	{ NULL }
};

// Dragons Lair 2   v3.14
static struct rom_def g_lair2_314_roms[] =
{
	{ "dl2_314.bin", NULL, 0, 0x10000, 0xaf92b612 },
	{ NULL }
};

// Dragons Lair 2   v3.15
static struct rom_def g_lair2_315_roms[] =
{
	{ "dl2_315.bin", NULL, 0, 0x10000, 0x13ec0600 },
	{ NULL }
};

// Dragons Lair 2   v3.17
static struct rom_def g_lair2_317_roms[] =
{
	{ "dl2_317.bin", NULL, 0, 0x10000, 0xe9453a1b },
	{ NULL }
};

// Dragons Lair 2   v3.18
static struct rom_def g_lair2_318_roms[] =
{
	{ "dl2_318.bin", NULL, 0, 0x10000, 0x64706492 },
	{ NULL }
};

// Dragons Lair 2   v3.19
static struct rom_def g_lair2_319_roms[] =
{
	{ "dl2_319.bin", NULL, 0, 0x10000, 0xe9453a1b },
	{ NULL }
};

// Dragons Lair 2   v3.16 Euro
static struct rom_def g_lair2_316e_roms[] =
{
	{ "dl2_316e.bin", NULL, 0, 0x10000, 0xd68f1b13 },
	{ NULL }
};

// Dragons Lair 2   v3.19 Euro
static struct rom_def g_lair2_319e_roms[] =
{
	{ "dl2_319e.bin", NULL, 0, 0x10000, 0xcc23ad9f },
	{ NULL }
};

// Dragons Lair 2   v3.19 Spanish
static struct rom_def g_lair2_319es_roms[] =
{
	{ "dl2_319es.bin", NULL, 0, 0x10000, 0x4b9a811d},
	{ NULL }
};

// Space Ace 91
static struct rom_def g_ace91_roms[] =
{
	{ "ace91.bin", NULL, 0, 0x10000, 0xde93a213 },
	{ NULL }
};

// Space Ace 91 Euro
static struct rom_def g_ace91e_roms[] =
{
	{ "ace91e.bin", NULL, 0, 0x10000, 0x27dd0486 },
	{ NULL }
};

//  soundhook

// Dragons Lair 2  v211
int g_dl2vars_211[] =
{ 0x14CD0
};

// Dragons Lair 2  v300
int g_dl2vars_300[] =
{ 0x14BCE
};

// Dragons Lair 2  v314
int g_dl2vars_314[] =
{ 0x150FA
};

// Dragons Lair 2  v315
int g_dl2vars_315[] =
{ 0x15106
};

// Dragons Lair 2  v317
int g_dl2vars_317[] =
{ 0x156c2
};

// Dragons Lair 2  v318
int g_dl2vars_318[] =
{ 0x154ca
};

// Dragons Lair 2  v319
int g_dl2vars_319[] =
{ 0x156c2
};

// Dragons Lair 2  v316 Euro
int g_dl2vars_316e[] =
{ 0x1509e
};

// Dragons Lair 2  v319 Euro
int g_dl2vars_319e[] =
{ 0x153aa
};

// Dragons Lair 2  v319 Euro Spanish
int g_dl2vars_319es[] =
{ 0x15666
};

// Space Ace 91  (not complete)
int g_ace91vars[] =
{ 0x1be1a
};

// Space Ace 91 Euro  (not complete)
int g_ace91evars[] =
{ 0x1bb94
};

int	*g_Dv = NULL;
int	g_dl2_ver = 0;
int g_dl2_euro = 0;

//////////////////////////////////////////////////////////////////////////

// lair class constructor (default the rev F2 roms)
lair2::lair2() :
	m_uSerialBufSize(0),	// the serial buffer starts off being empty
	m_serial_int_enabled(false),	// serial interrupts start off disabled
	m_real_1450(false)	// no real LDP-1450 is attached by default
{
	m_shortgamename = "lair2";
	memset(m_cpumem, 0, CPU_MEM_SIZE);
	m_joyskill_val = 0xFF;	// all input cleared
	m_misc_val = 0x7F;	// bit 7 and 6 alternated, and bits 0-3 all set (bits 4-5 are unused)

	m_game_uses_video_overlay = false;	// this game doesn't (by default) use video overlay

	struct cpudef cpu;
	memset(&cpu, 0, sizeof(struct cpudef));
	cpu.type = CPU_I88;
	cpu.hz = LAIR2_CPU_HZ;
	cpu.irq_period[0] = LAIR2_IRQ_PERIOD;
	cpu.irq_period[1] = ((1000.0)*(8+1))/9600.0;	// serial port IRQ (8 data bits, 1 stop bit, 1000 ms pre sec, 9600 bits per second
	cpu.nmi_period = 0.0f;
	cpu.initial_pc = 0xFFFF0;	// the ROM copies itself to 1000:0000 and starts there, but the intial entry is here
	cpu.must_copy_context = false;
	cpu.mem = m_cpumem;
	add_cpu(&cpu);	// add this cpu to the list (it will be our only one)

	m_disc_fps = 29.97;
	m_game_type = GAME_LAIR2;
	ldp_status = 0x00;

	m_num_sounds = 11;
	m_sound_name[S_DL2_WARBLE] = "dl2_warble.wav";
	m_sound_name[S_DL2_ERROR] = "dl2_error.wav";
	m_sound_name[S_DL2_GOOD] = "dl2_good.wav";
	m_sound_name[S_DL2_BAD] = "dl2_bad.wav";
	m_sound_name[S_DL2_TIC] = "dl2_tic.wav";
	m_sound_name[S_DL2_TOC] = "dl2_toc.wav";
	m_sound_name[S_DL2_COIN1] = "dl2_coin1.wav";
	m_sound_name[S_DL2_COIN2] = "dl2_coin2.wav";
	m_sound_name[S_DL2_COIN3] = "dl2_coin3.wav";
	m_sound_name[S_DL2_COIN4] = "dl2_coin4.wav";
	m_sound_name[S_DL2_WARN] = "dl2_warn.wav";

	// MPO:	we need to default to some version so it doesn't crash ...
	set_version(6);	// v3.19 is the newest DL2 rom so it makes sense to have that as default

#ifdef _XBOX
	set_version(GameOpts[GameInfo.GameIndex].RomVersion);
#endif

	// set mem (yes, we load in at 0xF0000 because the program entry point is at 0xFFFF0)
	g_lair2_211_roms[0].buf = g_lair2_300_roms[0].buf  = 
	g_lair2_314_roms[0].buf = g_lair2_315_roms[0].buf = 
	g_lair2_317_roms[0].buf  = g_lair2_318_roms[0].buf =
	g_lair2_319_roms[0].buf  = g_lair2_316e_roms[0].buf =
	g_lair2_319e_roms[0].buf = g_lair2_319es_roms[0].buf =
	g_ace91_roms[0].buf = g_ace91e_roms[0].buf = &m_cpumem[0xF0000];

	// ldp1450 text stuff
	memset ( &g_LDP1450_TextControl, 0, sizeof(ldp_text_control));

	// clear the ldp strings
	for (int i=0; i<3; i++)
	{
		g_LDP1450_Strings[i].String[0] = 0;
	}

	i86_set_irq_callback(lair2_irq_callback);
}

// Space Ace 91 class constructor
ace91::ace91()
{
	m_shortgamename = "ace91";
	m_game_type = GAME_ACE91;	

#ifdef _XBOX
	set_version(GameOpts[GameInfo.GameIndex].RomVersion);
#endif

}

void ace91::set_version(int version)
{
	{
		switch (version)
		{
		case	0:
			m_rom_list = g_ace91_roms;
			g_dl2_euro = 0;
			g_Dv = g_ace91vars;
			break;
		case	1:
			m_rom_list = g_ace91e_roms;
			g_dl2_euro = 1;
			g_Dv = g_ace91evars;
			break;
		}
		g_dl2_ver = version;
	}
}

bool lair2::init()
{
	cpu_init();
	g_ldp->pre_play();	// the LDP-1450 automatically begins playback
	return true;
}

void lair2::set_version(int version)
{

	{
		switch (version)
		{
		case	0:		//315
			m_rom_list = g_lair2_315_roms;
			g_Dv = g_dl2vars_315;
			g_dl2_euro = 0;
			break;
		case	1:		//211
			m_rom_list = g_lair2_211_roms;
			g_Dv = g_dl2vars_211;
			g_dl2_euro = 0;
			break;
		case	2:		//300
			m_rom_list = g_lair2_300_roms;
			g_Dv = g_dl2vars_300;
			g_dl2_euro = 0;
			break;
		case	3:		//314
			m_rom_list = g_lair2_314_roms;
			g_Dv = g_dl2vars_314;
			g_dl2_euro = 0;
			break;
		case	4:		//317
			m_rom_list = g_lair2_317_roms;
			g_Dv = g_dl2vars_317;
			g_dl2_euro = 0;
			break;
		case	5:		//318
			m_rom_list = g_lair2_318_roms;
			g_Dv = g_dl2vars_318;
			g_dl2_euro = 0;
			break;
		case	6:		//319
			m_rom_list = g_lair2_319_roms;
			g_Dv = g_dl2vars_319;
			g_dl2_euro = 0;
			break;
		case	7:		//316 Euro
			m_rom_list = g_lair2_316e_roms;
			g_Dv = g_dl2vars_316e;
			g_dl2_euro = 1;
			break;
		case	8:		//319 Euro
			m_rom_list = g_lair2_319e_roms;
			g_Dv = g_dl2vars_319e;
			g_dl2_euro = 1;
			break;
		case	9:		//319 Euro Spanish
			m_rom_list = g_lair2_319es_roms;
			g_Dv = g_dl2vars_319es;
			g_dl2_euro = 0;
			break;
		default:		// default to 3.15
			m_rom_list = g_lair2_315_roms;
			g_Dv = g_dl2vars_315;
			g_dl2_euro = 0;
			break;
		}
		g_dl2_ver = version;
	}
}

// game-specific command line arguments handled here
bool lair2::handle_cmdline_arg(const char *arg)
{
	// instead of emulating a LDP-1450, just pass the serial port commands through to the real serial port
	// This is for people who actually have a real 1450 and want to play DL2 on it
	if (strcasecmp(arg, "-real1450")==0)
	{
		m_real_1450 = true;
	}

	return m_real_1450;
}

void lair2::patch_roms()
{
	#ifndef _XBOX
	if (strcasecmp(m_shortgamename, "lair2") == 0)
	#else
	if (strcmp(m_shortgamename, "lair2") == 0)
	#endif
	{
		switch (g_dl2_ver)
		{
		case	0:			//315
			// cut wait on coins
			m_cpumem[0x1120f + 1] = 0x1;
			m_cpumem[0x111fa] = 0x1;

			// return from eep funcs() - not needed.. 
			// these take FOREVER to execute.. seems to be
			// flash mem for saving stats
			m_cpumem[0x13ec6] = 0xc3;
			m_cpumem[0x13f0c] = 0xc3;
			break;
		case	1:			//211
			// cut down wait count for coins.
			m_cpumem[0x11a3e + 1] = 0x1;
			m_cpumem[0x11a27 + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x146f4] = 0xc3;
			m_cpumem[0x1473a] = 0xc3;
			break;
		case	2:			//300
			// cut down wait count for coins.
			m_cpumem[0x111db + 1] = 0x1;
			m_cpumem[0x111f2 + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x13d04] = 0xc3;
			m_cpumem[0x13d4a] = 0xc3;
			break;
		case	3:			//314
			// cut down wait count for coins.
			m_cpumem[0x111f9 + 1] = 0x1;
			m_cpumem[0x1120f + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x13ec6] = 0xc3;
			m_cpumem[0x13f0c] = 0xc3;
			break;
		case	4:			//317
			// cut down wait count for coins.
			m_cpumem[0x11227 + 1] = 0x1;
			m_cpumem[0x1123d + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x140b0] = 0xc3;
			m_cpumem[0x140f6] = 0xc3;
			break;
		case	5:		//318
			// cut down wait count for coins.
			m_cpumem[0x111f9 + 1] = 0x1;
			m_cpumem[0x1120f + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x13eb4] = 0xc3;
			m_cpumem[0x13efa] = 0xc3;
			break;
		case	6:   //319
			// cut down wait count for coins.
			m_cpumem[0x11227 + 1] = 0x1;
			m_cpumem[0x1123d + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x140f6] = 0xc3;
			m_cpumem[0x140b0] = 0xc3;

			// FIX compiler bugs that make laserdisc access problematic

			// FIX Send_Byte_2_LDP() function
			// NOTE : this first fix isn't necessary because the bug doesn't seem to cause any harm in this instance
			memmove(&m_cpumem[0xF0B93], &m_cpumem[0xF0B92], 85);	// make room for extra opcode
			// first jnz doesn't need to be changed 'cause label is inside moved code
			m_cpumem[0xF0b9a]--;	// jnz 10be8
			m_cpumem[0xF0ba0]--;	// call 10f20
			m_cpumem[0xf0ba7]--;	// jnz 10b82
			m_cpumem[0xf0bb1]--;	// call 13732
			m_cpumem[0xf0bbb]--;	// call 13642
			m_cpumem[0xf0bc5]--;	// call 10f20
			m_cpumem[0xf0bcc]--;	// call 10b7f
			m_cpumem[0xf0bd6]--;	// call 13732
			m_cpumem[0xf0be0]--;	// call 10f20
			m_cpumem[0xf0be6]--;	// call 10b46
			m_cpumem[0xf0b91] = 0xFE;	// INC AL instead of INC AX
			m_cpumem[0xf0b92] = 0xC0;

			// FIX Wait_4_LDP() function
			// This fix is necessary to get passed LDP initialization reliably
			memmove(&m_cpumem[0xF0C39], &m_cpumem[0xF0C38], 17);	// make room for extra opcode
			m_cpumem[0xF0c37] = 0xFE;	// INC AL instead of INC AX
			m_cpumem[0xF0c38] = 0xC0;
			m_cpumem[0xF0C3A]--;	// jnz 10C4A
			m_cpumem[0xF0C40]--;	// time_delay function
			m_cpumem[0xF0C47]--;	// jnz 10C28
			m_cpumem[0xF0C49]--;	// jmp 10c52
			break;
		case	7:    //316 euro
			// cut down wait count for coins.
			m_cpumem[0x10ef9 + 1] = 0x1;
			m_cpumem[0x10F0f + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x13ba4] = 0xc3;
			m_cpumem[0x13bea] = 0xc3;
			break;
		case	8:    //319 euro
			// cut down wait count for coins.
			m_cpumem[0x10F27 + 1] = 0x1;
			m_cpumem[0x10F3d + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x13dae] = 0xc3;
			m_cpumem[0x13df4] = 0xc3;
			break;
		case	9:    //319 euro spanish
			// cut down wait count for coins.
			m_cpumem[0x11227 + 1] = 0x1;
			m_cpumem[0x1123d + 1] = 0x1;

			// return from eep funcs() - not needed.. 
			m_cpumem[0x14056] = 0xc3;
			m_cpumem[0x1409c] = 0xc3;
			break;
		}
	}	
	else
	{
		switch (g_dl2_ver)
		{
		case	0:
			// cut down wait count for coins.
//			m_cpumem[0x0F28 + 1] = 0x1;
//			m_cpumem[0x0F3E + 1] = 0x1;

			// return from eep funcs() - not needed.. 
//			m_cpumem[0x3dae] = 0xc3;
//			m_cpumem[0x3df4] = 0xc3;
			break;
		case	1:
			// cut down wait count for coins.
//			m_cpumem[0x0F28 + 1] = 0x1;
//			m_cpumem[0x0F3E + 1] = 0x1;

			// return from eep funcs() - not needed.. 
//			m_cpumem[0x3dae] = 0xc3;
//			m_cpumem[0x3df4] = 0xc3;
			break;
		}
	}
}


///////////////////////////////////////////////////////////

int g_dl2_irq_val = 0;	// which IRQ the i86 code should activate
// this needs to be global because the callback needs to access it

// irqline will always be 0, it won't help us unfortunately
int lair2_irq_callback(int irqline)
{
	i86_set_irq_line(0, CLEAR_LINE);	// first order of business, clear the IRQ line
	return g_dl2_irq_val;
}

// raise IRQ line
void lair2::do_irq(unsigned int which_irq)
{
	// cpu timer IRQ
	if (which_irq == 0)
	{
		g_dl2_irq_val = 0x1C;	// the value of the TIMER_INT
		i86_set_irq_line(0, ASSERT_LINE);
	}

	// serial port IRQ (COM 2)
	else if (which_irq == 1)
	{
		// if we have room in our buffer to receive more stuff
		if (m_uSerialBufSize < DL2_BUF_SIZE)
		{
			int serial_val = -1;

			if (g_dl2_euro)
			{
				serial_val = read_vp380();
			}
			else
			{
				// if we're connected to a real 1450 ... read the value now
				if (m_real_1450)
				{
					// if there is a character waiting in the buffer to be read ...
					if (serial_rx_char_waiting())
					{
						serial_val = serial_rx();
					}
				}

				// otherwise just read from our emulated 1450
				else
				{
					serial_val = read_ldp1450();
				}
			}

			// if we got a new character from the serial port, then add it to our buffer
			if (serial_val != -1)
			{
				#ifdef DEBUG
//					if (g_bVerbose) printline("COM2 IRQ tick, read data!");
				#endif

				m_u8SerialBuf[m_uSerialBufSize] = serial_val;
				++m_uSerialBufSize;

				/*
				char s[81];
				sprintf(s, "received %x from port\n", serial_val);
				printline(s);
				*/
			}

			// if our buffer is not empty, generate an IRQ to encourage ROM to empty it :)
			// NOTE : after reviewing DL2 source code, it seems ok to generate interrupts
			//  if the buffer is not empty, even if no new character has been received.
			// Not only that, buf this seems like the only way to ensure that the buffer is
			//  emptied in a timely manner.
			if (m_uSerialBufSize > 0)
			{
				// if serial port should generate interrupts
				if (m_serial_int_enabled)
				{
					g_dl2_irq_val = 0x0B;	// COM2
					i86_set_irq_line(0, ASSERT_LINE);
				}
				// else wait until serial interrupts have been enabled
			}
		}
		// else we have no room
	}
}


void lair2::port_write(Uint16 port, Uint8 value)
{
	// ignore other port accesses for speed
	// x202 - eep saving.. TODO
	if (port == 0x202 || port == 0x201)
		return;
	
	if (port < 0x2f8 || port > 0x300)
		return;

    switch(port)
	{

	// COM2 port 0x2F8
	case	0x2F8:			// transmit/receive
	{
#ifdef DEBUG
//		if (value == 0x67) g_bVerbose = true;
#endif
		// write to LDP
		// and set our interupt
		if (g_dl2_euro)
		{
			write_vp380(value);
		}
		else
		{
			// if we're hooked up to a real LDP-1450, send directly to the serial port
			if (m_real_1450)
			{
				serial_tx(value);
			}

			// otherwise send to our emulated LDP
			else
			{
				write_ldp1000(value);
			}
		}
		break;
	}
	case	0x2F8 + DL2_IER:		// Interrupt enable
		m_serial_int_enabled = (value != 0);	// 0 disables interrupts, non-zero enables them
		break;
	case	0x2F8 + DL2_IIR:		// Interrupt ID
	case	0x2F8 + DL2_LCR:		// Line control
	case	0x2F8 + DL2_MCR:		// Modem control
	case	0x2F8 + DL2_LSR:		// Line Status
	case	0x2F8 + DL2_MSR:		// Modem Status
		break;						// divisor latch high

	default:
        break;
	}		
}

Uint8 lair2::port_read(Uint16 port)
{
    unsigned char result = 0;

	if (port < 0x201 || port > 0x300)
		return (result);

	// handle ports	
	switch(port)
    {
	
	case	0x201:
	{
		int		val = 0x00;

		// send the joystick vals back
		if (!(m_joyskill_val & 1))			// up button
			val += ~1;
		
		if (!(m_joyskill_val & 2))			// down button
			val += ~2;

		if (!(m_joyskill_val & 4))			// left button
			val += ~4;
		
		if (!(m_joyskill_val & 8))			// right button
			val += ~8;

		if (!(m_joyskill_val & 16))		// sword button
			val += ~64;

		if (!(m_misc_val & 1))				// Player 1 start
			val += ~16;

		if (!(m_misc_val & 2))				// Player 2 start
			val += ~32;

		if (!(m_joyskill_val & 128))		// Service button
			val += ~128;

		if (val == 0x00)
			val = 0xff;
		
		result = val;

		break;
	}

	// COIN_PORT  (0x30 - coin1, 0x50 - coin2)
	// also doubles as EEP read.. TODO
	case	0x202:
	{
		// EEP expects 1 back (shouldnt affect coins)
		int	val = 0x01;

		// coin slot 1
		if (!(m_misc_val & 4))			
			val += 4;
		// coin slot 2
		if (!(m_misc_val & 8))
			val += 8;
		
		result = val;
		break;
	}

	// COM2 port 0x2F8
	case	0x2F8:
		// if we have a character waiting, then return the character
		if (m_uSerialBufSize > 0)
		{
			result = m_u8SerialBuf[0];	// first char in buf is always the oldest in the buffer
			memmove(m_u8SerialBuf, m_u8SerialBuf+1, m_uSerialBufSize - 1);	// move remaining bytes to front
			--m_uSerialBufSize;	// buffer now has 1 less byte in it
#ifdef DEBUG
			if (g_bVerbose)
			{
//				string s = "READ 0x2F8 : " + numstr::ToStr(result, 16);
//				printline(s.c_str());
			}
#endif
		}

		// this should never happen, so it's safe to put an error to notify us if something weird is going on
		else printline("LAIR2.CPP WARNING : tried to read from serial port when no char was waiting");
		break;

	case	0x2F8 + DL2_IER:		// Interrupt enable
		if (m_serial_int_enabled) result = 1;	// this reg will have a 1 if the interrupt is enabled
		break;
	case	0x2F8 + DL2_LCR:		// Line control
		break;

	case	0x2F8 + DL2_IIR:		// Interrupt ID
		// if we have a character from the LDP waiting to be read then return a 4 (not sure why 4, but... that's what the code expects)
		if (m_uSerialBufSize > 0)
		{
			result = 4;
		}
		// else a 0 is returned as usual
		break;

	case	0x2F8 + DL2_LSR:		// Line Status (returning DSR is safe)
	case	0x2F8 + DL2_MSR:		// Modem Status (returning CTS is safe)
		result = DL2_CTS | DL2_DSR;
		break;

    default:
        break;
    }

    return(result);
}

void lair2::cpu_mem_write(Uint32 Addr, Uint8 Value)
{
	m_cpumem[Addr] = Value; // always store to RAM if we write to it

	// sound hooks
	if (Addr == (Uint32) g_Dv[DL2_SOUND])		//315 roms
	{
		if (Value >= 0x31)
			sound_play (g_sound_map[Value - 39 ]);
		else
			sound_play (g_sound_map[Value]);
	}
}

// used to set dip switch values
bool lair2::set_bank(unsigned char which_bank, unsigned char value)
{
	bool result = true;
	
	printline("ERROR: Dragon's Lair 2 uses onscreen setup");
	result = false;

	return result;
}


void lair2::input_enable(Uint8 move)
{
	switch(move)
	{
	case SWITCH_UP:
		m_joyskill_val &= (unsigned char) ~1;	// clear bit 0
		break;
	case SWITCH_DOWN:
		m_joyskill_val &= (unsigned char) ~2;	// clear bit 1
		break;
	case SWITCH_LEFT:
		m_joyskill_val &= (unsigned char) ~4;	// clear bit 2
		break;
	case SWITCH_RIGHT:
		m_joyskill_val &= (unsigned char) ~8;	// clear bit 3
		break;
	case SWITCH_START1: // PLAYER 1
		m_misc_val &= ~1;	// clear bit 0 ...
		break;
	case SWITCH_START2:
		m_misc_val &= ~2;	// clear bit 1
		break;
	case SWITCH_BUTTON1: // SWORD
		m_joyskill_val &= (unsigned char) ~0x10;	// clear bit 4
		break;
	case SWITCH_COIN1: 
		m_misc_val &= (unsigned char) ~0x04;	
		break;
	case SWITCH_COIN2: 
		m_misc_val &= (unsigned char) ~0x08;	
		break;
	case SWITCH_SKILL1:	// cadet
		m_joyskill_val &= (unsigned char) ~0x20;	// clear bit 5
		break;
	case SWITCH_SKILL2:	// captain
		m_joyskill_val &= (unsigned char) ~0x40;	// clear bit 6
		break;
	case SWITCH_SKILL3:	// space ace
		m_joyskill_val &= (unsigned char) ~0x80;	// clear bit 7
		break;
	default:
		break;
	}
}


void lair2::input_disable(Uint8 move)
{
	switch(move)
	{
	case SWITCH_UP:
		m_joyskill_val |= 1;	// set bit 0
		break;
	case SWITCH_DOWN:
		m_joyskill_val |= 2;	// set bit 1
		break;
	case SWITCH_LEFT:
		m_joyskill_val |= 4;	// set bit 2
		break;
	case SWITCH_RIGHT:
		m_joyskill_val |= 8;	// set bit 3
		break;
	case SWITCH_START1:
		m_misc_val |= 1; // PLAYER 1
		break;
	case SWITCH_START2:
		m_misc_val |= 2;	// set bit 1
		break;
	case SWITCH_BUTTON1: // SWORD
		m_joyskill_val |= 0x10;	// set bit 4
		break;
	case SWITCH_COIN1: 
		m_misc_val |= 0x04;	
		break;
	case SWITCH_COIN2: 
		m_misc_val |= 0x08;	
		break;
	case SWITCH_SKILL1:	// cadet
		m_joyskill_val |= (unsigned char) 0x20;	// set bit 5
		break;
	case SWITCH_SKILL2:	// captain
		m_joyskill_val |= (unsigned char) 0x40;	// set bit 6
		break;
	case SWITCH_SKILL3:	// space ace
		m_joyskill_val |= (unsigned char) 0x80;	// set bit 7
		break;
	default:
		break;
	}
}
