/*
 * ldp1000.cpp
 *
 * Copyright (C) 2001 Mark Broadhead
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// ldp1000.c
// part of the DAPHNE emulator
// written by Mark Broadhead
//
// lDP-1450 command and text handling added by Paul Blagay 2003
//
// This code emulates the Sony LDP-1000 laserdisc player which is used in
// Bega's Battle and Cobra Command (dedicated). 

#include <stdio.h>
#include <stdlib.h>	// for atoi
#include <string.h>	// memset, strcpy
#include "../game/game.h"
#include "ldp1000.h"
#include "../io/conout.h"
#include "../ldp-out/ldp.h"

#define LDP1000_STACKSIZE 10 // the largest response is 5 bytes
#define FRAME_SIZE 5

unsigned char ldp1000_output_stack[LDP1000_STACKSIZE];
int ldp1000_output_stack_pointer = 0;
unsigned int ldp1000_curframe;
unsigned int ldp1000_repeat_frame;
unsigned int ldp1000_repeat_start_frame;
bool ldp1000_repeat = false;
int ldp1000_times_to_repeat = 0;

char ldp1000_frame[FRAME_ARRAY_SIZE] = {0}; // holds the digits sent to the LDP1000
int ldp1000_frame_index = 0; 

// LDP1450 - Text Display (up to 3 lines)
// NOTE, you will need to handle the rendering of these strings
// in your engine.
ldp_text			g_LDP1450_Strings[3];
ldp_text_control	g_LDP1450_TextControl;

// default status
static unsigned char enter_status = 0x00;
///////////////////////////////////////////

// clear all the crap
void reset_ldp1450_globals()
{
	// ldp1450 text stuff
	memset ( &g_LDP1450_TextControl, 0, sizeof(ldp_text_control));

	// clear the ldp strings
	for (int i=0; i<3; i++)
		g_LDP1450_Strings[i].String[0] = 0;
}

// retrieves the status from our virtual LDP1000
unsigned char read_ldp1000()
{
	unsigned char result = 0x00;

	if (ldp1000_output_stack_pointer < 1)
	{
		printline("ERROR: LDP1000 stack read when empty");
	}
	else 
	{
		ldp1000_output_stack_pointer--;
		result = ldp1000_output_stack[ldp1000_output_stack_pointer];
	}
	
	return(result);
}

// MPO : returns -1 if the buffer is empty or a value between 0-0xFF if we are returning a legitimate value
int read_ldp1450()
{
	if (ldp1000_output_stack_pointer < 1)
		return(-1);
	
	ldp1000_output_stack_pointer--;
	return (ldp1000_output_stack[ldp1000_output_stack_pointer]);
}


// pushes a value on the ldp1000 stack, returns 1 if successful or 0 if stack is full
int ldp1000_stack_push(unsigned char value)
{
	int result = 0;

	// if we still have room to push
	if (ldp1000_output_stack_pointer < LDP1000_STACKSIZE-1)
	{
		ldp1000_output_stack[ldp1000_output_stack_pointer++] = value;
		result = 1;
	}
	else
	{
		printline("ERROR: LDP1000 stack overflow (increase its size)");
	}

	return(result);
}

// sends a byte to our virtual LDP1000
void write_ldp1000 (unsigned char value)
{
//	char s[81] = { 0 };

	// text control
	// XY Position tags
	if (g_LDP1450_TextControl.TextGetXY)
	{
		if (g_LDP1450_TextControl.TextGotXpos == 0)
		{
			g_LDP1450_TextControl.TextGotXpos = 1;
			g_LDP1450_TextControl.TextXPos = value;
		}
		else if (g_LDP1450_TextControl.TextGotYpos == 0)
		{
			g_LDP1450_TextControl.TextGotYpos = 1;
			g_LDP1450_TextControl.TextYPos = value;
		}
		else
		{
			// Text Mode.. we are using this for scale
			// for now.. 
			g_LDP1450_TextControl.TextMode = value;
			g_LDP1450_TextControl.TextGetXY = 0;

			// set scale..
			if (value == 0x030)
				g_LDP1450_TextControl.Scale = 2.0f;		// 3 lines of 10 chars
			else
				g_LDP1450_TextControl.Scale = 3.0f;		// 1 line of 20 chars
		}
		return;
	}

	// string to fill out??
	// ignore first char, then store the rest off
	if (g_LDP1450_TextControl.TextGotString)
	{
		// 00,0a,14 - 1st,2nd,3rd lines
		if ( (value == 0x00 || value == 0x0a || value == 0x14) && g_LDP1450_TextControl.TextGot00 == 0)
		{
			// set our line position
			g_LDP1450_TextControl.TextLine = value / 10;
			g_LDP1450_TextControl.TextGot00 = 1;
		}
		else
		{
			// store off chars until we get a null
			if (value != 0x1a)	//string complete cmd
			{
				// make sure we have a valid ascii char
				if (value>=32)
				{
					g_LDP1450_TextControl.TextString[g_LDP1450_TextControl.TextStringPointer] = value;
					g_LDP1450_TextControl.TextStringPointer++;
				}
			}
			else
			{
				int spaces = 0;
				int i = 0;
				
				// if its a string of spaces, then clear all strings
				for (i=0; i<32; i++)
				{
					if (g_LDP1450_TextControl.TextString[i] == 32)
						spaces++;
				}

				// clear the text strings
				if (spaces > 20)
				{
					for (i=0; i<3; i++)
						g_LDP1450_Strings[i].String[0] = 0;
				}
				else	// set it to draw
				{
					// copy the string over
					strcpy (g_LDP1450_Strings[g_LDP1450_TextControl.TextLine].String, g_LDP1450_TextControl.TextString);
					g_LDP1450_Strings[g_LDP1450_TextControl.TextLine].x = 20 + (g_LDP1450_TextControl.TextXPos * 7	);
					g_LDP1450_Strings[g_LDP1450_TextControl.TextLine].y = 20 + ((g_LDP1450_TextControl.TextYPos * 5) + 
																				g_LDP1450_TextControl.TextLine * 38);
				}
				// clear vars
				g_LDP1450_TextControl.TextStringPointer = 0;
				memset(g_LDP1450_TextControl.TextString, 0, 128);
				g_LDP1450_TextControl.TextGotString = 0;
				g_LDP1450_TextControl.TextGot00 = 0;
				g_LDP1450_TextControl.TextGotXpos = 0;
				g_LDP1450_TextControl.TextGotYpos = 0;
				g_LDP1450_TextControl.TextGetXY = 0;
			}
		}
		return;
	}

	// regular commands
	switch (value)
	{
	case 0x00:	//LDP1450 - text handling
		if (g_LDP1450_TextControl.TextCommand == true)
		{			
			g_LDP1450_TextControl.TextGetXY = 1;
			g_LDP1450_TextControl.TextCommand = false;
		}
		break;
	case 0x01:	//LDP1450 - Text handling
		if (g_LDP1450_TextControl.TextCommand == true)
		{
			g_LDP1450_TextControl.TextGotString = 1;
			g_LDP1450_TextControl.TextCommand = false;
		}
		break;
	case 0x0a:	// ldp1450 - 2nd line
	case 0x14:	// ldp1450 - 3rd line
	case 0x1a:	// ldp1450 - text sent..
		break;
	case 0x24:	// audio_mute_on
		g_ldp->disable_audio1();	
		g_ldp->disable_audio2();
		break;
	case 0x25:	// audio_mute_off
		g_ldp->enable_audio1();	
		g_ldp->enable_audio2();
		break;
	case 0x28:	// Stop Codes Enable
		break;
	case 0x29:	// Stop Codes Disable
		break;
	case 0x30: // '0'
	case 0x31: // '1'
	case 0x32: // '2'
	case 0x33: // '3'
	case 0x34: // '4'
	case 0x35: // '5'
	case 0x36: // '6'
	case 0x37: // '7'
	case 0x38: // '8'
	case 0x39: // '9'
		ldp1000_add_digit(value);
		ldp1000_stack_push(0x0a); // ack
		break;
	case 0x3a: // Play
		g_ldp->pre_play();
		ldp1000_stack_push(0x0a); // ack
		break;
	case 0x40: // Enter
		ldp1000_enter();
		break;
	case 0x43: // Search 
		enter_status = LDP1000_SEARCH;
		ldp1000_stack_push(0x0a); // ack
		break;
	case 0x44: // Repeat
		enter_status |= LDP1000_REPEAT;
		ldp1000_stack_push(0x0a); // ack
		break;
	case 0x46: // Channel 1 on
		g_ldp->enable_audio1();
		ldp1000_stack_push(0x0a);	// ack
		break;
	case 0x47: // Channel 1 off
		g_ldp->disable_audio1();
		ldp1000_stack_push(0x0a);
		break;
	case 0x48: // Channel 2 on
		g_ldp->enable_audio2();
		ldp1000_stack_push(0x0a);	// ack
		break;
	case 0x49: // Channel 2 off
		g_ldp->disable_audio2();
		ldp1000_stack_push(0x0a);
		break;
 	case 0x4F: // LDP1450 - Still()
		g_ldp->pre_pause();	// return 0xB if the disc isn't playing
		break;
	case 0x56: // C.L. (reset)
		ldp1000_stack_push(0x0a); // ack
		break;
	case 0x60: // Addr Inq (get current frame number)
		ldp1000_curframe = g_ldp->get_current_frame();
		char f[6];
		//itoa(ldp1000_curframe, &f[0], 10);	// should be faster than sprintf
		//safe_itoa(ldp1000_curframe, f, sizeof(f));	// MPO : linux doesn't have itoa
		sprintf(f, "%05u", ldp1000_curframe);  //WDO:  Should have leading zeros, so we're back to sprintf() =]
		ldp1000_stack_push(f[4]); // M5
		ldp1000_stack_push(f[3]); // M4
		ldp1000_stack_push(f[2]); // M3
		ldp1000_stack_push(f[1]); // M2
		ldp1000_stack_push(f[0]); // M1
		break;
	case 0x67:	// LDP1450 - Status Inquiry

		// if the disc is paused
		if (g_ldp->get_status() == LDP_PAUSED)
		{
			ldp1000_stack_push(0x20); // 
			ldp1000_stack_push(0x00); // disc in, door closed,paused
			ldp1000_stack_push(0x10); // 
			ldp1000_stack_push(0x00); // 
			ldp1000_stack_push(0x80); // 
		}
		else if (g_ldp->get_status() == LDP_PLAYING)
		{
			ldp1000_stack_push(1);
			ldp1000_stack_push(00);
			ldp1000_stack_push(0x10);
			ldp1000_stack_push(0);
			ldp1000_stack_push(0x80);
		}
		// 8a 00 00 00 00 = no disc in the player
		else
		{
			// if this happens, you need to account for the other status codes
			printline("ERROR : ldp1000 status was queried but we aren't prepared to handle it");

			// return "paused" status just so we don't get caught in an endless loop
			ldp1000_stack_push(0x20); // 
			ldp1000_stack_push(0x00); // disc in, door closed,paused
			ldp1000_stack_push(0x10); // 
			ldp1000_stack_push(0x00); // 
			ldp1000_stack_push(0x80); // 			

		}
		break;
	case 0x6e:	// CX_ON	??
		break;
	case 0x80:	// text start
		g_LDP1450_TextControl.TextCommand = true;
		break;
	case 0x81:	// LDP1450 - Turn on text
		g_LDP1450_TextControl.LDP1450_DisplayOn = true;
		break;
	case 0x82:	// LDP1450 - Turn off text
		g_LDP1450_TextControl.LDP1450_DisplayOn = false;
		break;
	default:
		ldp1000_stack_push(0x0a); // ack
		break;
	}
}


bool ldp1000_result_ready(void)
{
	bool result = false;

	if (ldp1000_repeat)
	{
		if (g_ldp->get_current_frame() >= ldp1000_repeat_frame)
		{
			// we finished one repeat
			ldp1000_times_to_repeat--;
			
			// start over if we still need to repeat
			if (ldp1000_times_to_repeat)
			{
				char f[6] = {0};
				sprintf(f, "%i", ldp1000_repeat_start_frame);
				g_ldp->pre_search(f, true);
				g_ldp->pre_play();
			}
			// or if we are done, pause
			else
			{
				g_ldp->pre_pause();
				ldp1000_repeat = false;
			}

			// we got to the desired frame
			ldp1000_stack_push(0x01); // completion
		}
	}
	
	if (ldp1000_output_stack_pointer)
	{
		result = true;
	}
	return result;
}

void ldp1000_enter(void)
{
//	char s[81] = {0};
	
	if (enter_status == LDP1000_SEARCH)
	{
		ldp1000_repeat = false;
		
		ldp1000_frame[ldp1000_frame_index] = 0;
		
		// TODO: check for search failure
		g_ldp->pre_search(ldp1000_frame, true);

		ldp1000_frame_index = 0; // reset frame index

		ldp1000_stack_push(0x01); // completion
		ldp1000_stack_push(0x0a); // ack
		enter_status = 0x00;
	}

	else if (enter_status == LDP1000_REPEAT_NUMBER)
	{
		if (ldp1000_frame_index == 0)
		{
			ldp1000_times_to_repeat = 1;
		}
		// TODO: check for number of repeats other than 1
//		sprintf(s, "LDP1000: Repeat %i times\n", ldp1000_times_to_repeat); 
//		printline(s);

		g_ldp->pre_play();

		ldp1000_frame_index = 0; // reset frame index
		ldp1000_repeat = true;

		ldp1000_stack_push(0x0a); // ack
		enter_status = 0x00;
	}
	
	else if (enter_status == LDP1000_REPEAT)
	{
		ldp1000_frame[ldp1000_frame_index] = 0;
		
//		sprintf(s, "LDP1000: Repeat at frame %i\n", atoi(ldp1000_frame)); 
//		printline(s);

		ldp1000_repeat_frame = atoi(ldp1000_frame);
		ldp1000_repeat_start_frame = g_ldp->get_current_frame();

		ldp1000_frame_index = 0; // reset frame index
		
		ldp1000_stack_push(0x0a); // ack

		enter_status = LDP1000_REPEAT_NUMBER;
	}
}

// adds a digit to  the frame array that we will be seeking to
// digit should be in ASCII format
void ldp1000_add_digit(char digit)
{
	if (ldp1000_frame_index < FRAME_SIZE)
	{
		ldp1000_frame[ldp1000_frame_index] = digit;
		ldp1000_frame_index++;
	}
	else
	{
//		char s[81] = { 0 };

//		sprintf(s, "Too many digits received for frame! (over %d)", FRAME_SIZE);
//		printline(s);
	}
}
