
/*
 * ldp.cpp
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// LDP.CPP
// by Matt Ownby

// The code in this file translates general LDP functions into LDP-specific functions
// Part of the Daphne emulator

#ifdef DEBUG
#include <assert.h>
#include "../io/numstr.h"
#include "../cpu/cpu-debug.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../io/serial.h"
#include "ldp.h"
#include "../timer/timer.h"
#include "../io/conout.h"
#include "framemod.h"
#include "../game/game.h"
#include "../game/boardinfo.h"
#include "../cpu/cpu.h"
#include "../cpu/generic_z80.h"

#ifdef WIN32
#pragma warning (disable:4100)	// disable the warning about unreferenced parameters
#endif

// generic ldp constructor
ldp::ldp() :
	need_serial(false),
	serial_initialized(false),
	player_initialized(false),
	blitting_allowed(true),
	skipping_supported(false),
	skip_instead_of_search(false),
	max_skippable_frames(0),
	m_last_try_frame(0),
	m_last_seeked_frame(0),
	m_play_cycles(0),
	m_play_time(0),
	m_status(LDP_STOPPED),
	search_latency(0),
	m_stop_on_quit(false),
	m_discvideo_width(640),
	m_discvideo_height(480),
	m_use_nonblocking_searching(false),
	m_dont_get_search_result(false),
	m_sram_continuous_update(false),
	m_noldp_timer(0)
{
	m_bug_log.clear();	// probably redundant, but what the heck..
}

ldp::~ldp()
{
	pre_shutdown();
}

// Initializes and "pings" the LDP/VLDP and returns true if the player is online and detected
// or false if the LDP could not be initialized
bool ldp::pre_init()
{

	bool result = true;	// assume everything works until we find out otherwise
	bool temp = true;	// needed to make && work the way we want below

	// If we are controlling a real LDP,
	// or if we are controlling a combo and DVD initialized properly,
	// then initialize the serial port here
	if (need_serial)
	{
		printline("NOTE : You are attempting to use DAPHNE with a real laserdisc player!");
		printline("If you don't have DAPHNE plugged into a real laserdisc player,");
		printline("you should be using VLDP instead.");
		serial_initialized = serial_init(get_serial_port(), get_baud_rate());
		temp = serial_initialized;
	}
	
	player_initialized = init_player();
	result = temp && player_initialized;
	
	return(result);

}

// initializes the player after serial has been initialized
// this is so that the main init function can initialize the serial without having each LDP
// initialize its own serial
bool ldp::init_player()
{
	return(true);
}

// may stop the player, shuts down the serial if it's open, then calls shutdown_player for player-specific stuff
void ldp::pre_shutdown()
{
	if (player_initialized)
	{
		// if stop on quit has been requested stop the player from playing
		if (m_stop_on_quit)
		{
			pre_stop();
		}
		
		// if serial has been initialized, shut it down now
		if (serial_initialized)
		{
			serial_close();
			serial_initialized = false;
		}
	
		shutdown_player();
		player_initialized = false;
	}
	// else we were never initialized
}

// does any player specific stuff  
void ldp::shutdown_player()
{
}

// see .h file for usage
bool ldp::pre_search(const char* pszFrame, bool block_until_search_finishes)
{
	char frame[FRAME_ARRAY_SIZE] = { 0 };
	Uint16 frame_number = 0;	// frame to search to
	bool result = false;
	char s1[81] = { 0 };

	// safety check, if they try to search without checking the search result ...
	if (m_status == LDP_SEARCHING)
	{
		printline("LDP : tried to search without checking for search result first! that's bad!");
		printline(frame);
		
		// this is definitely a Daphne bug if this happens, so log it!
		m_bug_log.push_back("LDP.CPP, pre_search() : tried to search without checking for search result first!");
		return false;
	}
	// end safety check

	// Get the frame that we are on now, before we do the seek
	// This is needed in order to calculate artificial seek delay
	// We must do this before we change 'm_status' due to the way get_current_frame is designed
	m_last_seeked_frame = get_current_frame();

	m_status = LDP_SEARCHING;	// searching can take a while

	// we copy here so that we can safely null-terminate
	strncpy(frame, pszFrame, 5);
	frame[5] = 0;	// terminate the string ...
	frame_number = (Uint16) atoi(frame);

	// If we need to alter the frame # before searching
	if (need_frame_conversion())
	{
		Uint16 unadjusted_frame = frame_number;
		frame_number = (Uint16) do_frame_conversion(frame_number);
		framenum_to_frame(frame_number, frame);
		sprintf(s1, "Search to %d (formerly %d) received", frame_number, unadjusted_frame);
	}
	else
	{
		sprintf(s1, "Search to %d received", frame_number);
	}

	outstr(s1);

	// if it's Dragon's Lair/Space Ace, print the board we are on
	if ((g_game->get_game_type() == GAME_LAIR) || (g_game->get_game_type() == GAME_ACE))
	{
		Uint8 *cpumem = get_cpu_mem(0);	// get the memory for the first (and only)
		outstr(" - ");
		print_board_info(cpumem[0xA00E], cpumem[0xA00F], cpumem[Z80_GET_IY]);
	}
	else
	{
		newline();
	}

	// If the user requested a delay before seeking, make it now
	if (search_latency > 0)
	{
		make_delay(get_search_latency());
	}

	m_last_try_frame = (Uint16) atoi(frame);	// the last frame we tried to seek to becomes this current one

	// HERE IS WHERE THE SEARCH ACTUALLY TAKES PLACE

	int difference = frame_number - get_current_frame();	// how many frames we'd have to skip

	// if the player supports skipping AND user has requested the we skip instead of searching when possible
	// AND if we can skip forward instead of seeking ...
	if (skipping_supported && skip_instead_of_search && ((difference <= max_skippable_frames) && (difference > 1)))
	{
		result = pre_skip_forward((Uint16) difference);
	}

	// otherwise do a regular search
	else
	{
		result = nonblocking_search(frame);
		m_dont_get_search_result = false;	// it's now ok to get the search result

		// if we are to block until the search finishes, then wait here (this is bad, don't do this!)
		// This is only here for backward compatibility!
		if (block_until_search_finishes)
		{
			// if search command was accepted
			if (result)
			{
				unsigned int cur_time = refresh_ms_time();
				int ldp_stat = -1;

				// we know we may be waiting for a while, so we pause the cpu timer to avoid getting a flood after the seek completes
				cpu_pause();

				// wait for player to change its status or for us to timeout
				while (elapsed_ms_time(cur_time) < 7000)
				{
					ldp_stat = get_status();	// get_status does some stuff that needs to get done, so we don't just read m_status instead

					// if the search is finished
					if (ldp_stat != LDP_SEARCHING)
					{
						break;
					}
					make_delay(1);	// don't hog CPU
				}

				cpu_unpause();	// done with the delay, so we can unpause

				// if we didn't succeed, then return an error
				if (ldp_stat != LDP_PAUSED)
				{
					printline("LDP : blocking search didn't succeed");
					result = false;
				}
			} // if the initial search command was accepted
			else
			{
				printline("LDP : search command was rejected");
				m_status = LDP_ERROR;
			}
		} // end if we were doing a blocking styled search
	} // end regular search (instead of skipping)

	return(result);
}

// don't call this function directly, call pre_search instead
bool ldp::nonblocking_search(char *new_frame)
{
	m_noldp_timer = refresh_ms_time();
	return(true);
}

// don't call this function directly, call get_status() instead
int ldp::get_search_result()
{
	int result = SEARCH_BUSY;
	
	// stall for a couple of seconds to simulate search delay
	if (elapsed_ms_time(m_noldp_timer) > 2000)
	{
		printline("Search success!");
		result = SEARCH_SUCCESS;
	}
	return result;
}

// prepare to skip forward a certain # of frames and continue playing
// Call this function, do not call skip_forward directly
bool ldp::pre_skip_forward(Uint16 frames_to_skip)
{
	bool result = false;

	// only skip if the LDP is playing
	if (m_status == LDP_PLAYING)
	{
		char s[81] = { 0 };
		Uint32 ms_to_skip = 0;
		Uint32 completed_frames = 0;	// # of frames that have been displayed since we last played
		Uint16 target_frame = (Uint16) (get_current_frame() + frames_to_skip);

		completed_frames = (Uint32) (elapsed_ms_time(m_play_time) / g_game->get_disc_ms_per_frame());
		// find out how many frames we've completed up to this point, yes we want result truncated so we don't add +0.5
		completed_frames++;	// increment by 1 to account for the current frame we're on
		ms_to_skip = (Uint32) ((completed_frames * g_game->get_disc_ms_per_frame())+0.5);
		// compute what to add to our play_timer

		Uint32 cpu_timer = get_cpu_timer();
		m_play_time = m_play_time + ms_to_skip;	// adjust book-keeping
		m_last_seeked_frame = target_frame;	// adjust book-keeping

		// if cpu timers were not reset, we can do a perfect re-calculation of the cycles
		if (m_play_time > cpu_timer)
		{
			m_play_cycles = (Uint64) ((m_play_time - cpu_timer) * get_cpu_cycles_per_ms(0));
		}
		// otherwise we'll have to settle for a good approximation (this shouldn't ever happen)
		else
		{
			printline("LDP.CPP approximating m_play_cycles, you should try to avoid this");
			m_play_cycles = m_play_cycles + (Uint64) ((ms_to_skip * get_cpu_cycles_per_ms(0))+0.5);
		}

		result = skip_forward(frames_to_skip, target_frame);

		sprintf(s, "Skipped forward %d frames", frames_to_skip);
		printline(s);
	}
	else
	{
		printline("LDP ERROR: Skip forward command was called when the disc wasn't playing");
	}

	return(result);
}

// prepare to skip forward a certain # of frames backward and continue playing forward
// do not call skip_backward directly
bool ldp::pre_skip_backward(Uint16 frames_to_skip)
{
	bool result = false;

	// only skip if the LDP is playing
	if (m_status == LDP_PLAYING)
	{
		result = skip_backward(frames_to_skip, (Uint16) (get_current_frame() - frames_to_skip));

		char s[81];
		sprintf(s, "Skipped backward %d frames", frames_to_skip);
		printline(s);
	}
	else
	{
		printline("LDP ERROR: Skip backward command was called when the disc wasn't playing");
	}

	return(result);
}

// steps forward 1 frame
void ldp::pre_step_forward()
{
	// NOTE : right now we don't have LDP-specific implementations of this function since it is rarely used
	char frame[6];
	Uint16 new_frame = g_ldp->get_current_frame();

	// bounds check (if we haven't overflowed)
	if (new_frame < ((Uint16) -1))
	{
		framenum_to_frame(g_ldp->get_current_frame() + 1, frame);
		printline("LDP : Stepping forward one frame");
		g_ldp->pre_search(frame, true);
	}
	else
	{
		printline("LDP: pre_step_forward failed bounds check");
	}
}

// steps backward 1 frame
void ldp::pre_step_backward()
{
	// NOTE : right now we don't have LDP-specific implementations of this function since it is rarely used
	char frame[6];
	Uint16 new_frame = g_ldp->get_current_frame();

	// don't step backward before the first frame on the disc
	if (new_frame > 0)
	{
		new_frame--;
	}

	framenum_to_frame(new_frame, frame);
	printline("LDP : Stepping backward one frame");
	g_ldp->pre_search(frame, true);
}

// skips forward a certain number of frames and continues playing
// Do not call this function directly!  Call pre_skip_forward instead
bool ldp::skip_forward(Uint16 frames_to_skip, Uint16 target_frame)
{
	bool result = false;
	
	char frame[6] = {0};
	sprintf(frame, "%05d", target_frame);

	result = pre_search(frame, true);
	pre_play();

	return(result);
}

// DO NOT CALL THIS FUNCTION DIRECTLY
// call pre_skip_backward instead
bool ldp::skip_backward(Uint16 frames_to_skip, Uint16 target_frame)
{
	bool result = false;
	
	char frame[6] = {0};
	sprintf(frame, "%05d", target_frame);

	result = pre_search(frame, true);
	pre_play();

	return(result);
}

// prepares to play the disc
void ldp::pre_play()
{
	Uint32 cpu_hz;	// used to calculate elapsed cycles

	// safety check, if they try to play without checking the search result ...
	// THIS SAFETY CHECK CAN BE REMOVED ONCE ALL LDP DRIVERS HAVE BEEN CONVERTED OVER TO NON-BLOCKING SEEKING
	if (m_status == LDP_SEARCHING)
	{
		printline("LDP : tried to play without checking to see if we were still seeking! that's bad!");
		
		// if this ever happens, it is a bug in Daphne, so log it
		m_bug_log.push_back("LDP.CPP, pre_play() : tried to play without checking to see if we're still seeking!");
		
		return;
	}

	// we only want to refresh the frame calculation stuff if the disc is
	// not already playing
	if (m_status != LDP_PLAYING)
	{
		// if the disc may need to take a long time to spin up, then pause the cpu timers while the disc spins up
		if (m_status == LDP_STOPPED)
		{
			cpu_pause();
			m_play_time = play();
			cpu_unpause();
		}
		else
		{
			m_play_time = play();
		}

		cpu_hz = get_cpu_hz(0);

		// if the game is using a cpu, then we want to calculate current LDP frame based
		// off of elapsed cpu cycles
		if (cpu_hz)
		{
			Uint32 cpu_timer = get_cpu_timer();
			
			// IMPORTANT NOTE about m_play_cycles!
			// This method works as long as the total # of cycles elapsed is
			//  accurate.  My tests show that converting doubles to uint64's
			//  should be accurate for about 53 bits, which gives about 71 years
			//  of playing time on a 4 MHz system.  Should be good enough.
			// The advantage of this method is it tends to keep the video and the cpu
			//  in sync better, since m_play_time is given to us by our laserdisc
			//  driver.
			m_play_cycles = (Uint64) ((m_play_time - cpu_timer) * get_cpu_cycles_per_ms(0));
			
			// The advantage of this method is it is more stable.  The disadvantage is
			//  it may not be too compatible with skipping (which favors the other
			//  method) and it may not be as in sync with the laserdisc video.
			//m_play_cycles = get_total_cycles_executed(0);
		} // end if we're using cpu cycles to calculate current laserdisc frame
		
		m_status = LDP_PLAYING;
	}
	else
	{
		printline("LDP : disc is already playing, play command ignored");
	}

	printline("Play");	// moved to the end of the function so as to not cause lag before play command could be issued
}

// starts playing the laserdisc
// returns the timer value that indicates when the playback actually started (used to calculate current frame)
unsigned int ldp::play()
{
	return refresh_ms_time();
}

// sets the speed at which we playback at (1 is standard speed)
void ldp::set_play_speed(Uint8 speed)
{
	char s[81] = { 0 };

	sprintf(s, "LDP: Playback speed of %d was requested (ignored)", speed);
	printline(s);
}

// prepares to pause
void ldp::pre_pause()
{
	// only send pause command if disc is playing
	// some games (Super Don) repeatedly flood with a pause command and this doesn't work well with the Hitachi
	if (m_status == LDP_PLAYING)
	{
		m_last_seeked_frame = get_current_frame();
		pause();
		m_status = LDP_PAUSED;
		printline("Pause");
	}
	else
	{
		printline("LDP : Received pause while disc was not playing, ignoring");
	}
}

// pauses the disc
void ldp::pause()
{
}

// prepares to stop the disc
// "stop" is defined as going to frame 0 and stopping the motor so that
// the player has to spin up again to begin playing
void ldp::pre_stop()
{
	m_last_seeked_frame = 0;
	stop();
	m_status = LDP_STOPPED;
	printline("Stop");
}

// stops the disc
void ldp::stop()
{
}

// (see .h for description)
Uint16 ldp::get_current_frame()
{
	Uint16 result = 0;

	// if the disc is playing, compute which frame # we're on
	// IMPORTANT:
	// ldp-vldp.cpp's nonblocking_seek() RELIES on this function NOT calling get_status()!!!
	// Be very careful about changing this 'm_status' to a get_status()
	if (m_status == LDP_PLAYING)
	{
		Uint32 cpu_hz = get_cpu_hz(0);
		double elapsed_secs = 0.0;	// how many seconds have elapsed since the disc began playing

		// if our game has a cpu, then we use its elapsed cycles to compute the elapsed time.
		// This prevents major problems with games that depend on their cpu/irq's to sync with the laserdisc
		// player, such as Super Don Quix-ote
		if (cpu_hz)
		{
			Uint64 cur_total_cycles = get_total_cycles_executed(0);

			// safety check : sometimes m_play_cycles will be greater than cur_total_cycles if the cpu is a little lagged
			if (cur_total_cycles > m_play_cycles)
			{
				// it seems kind of silly to cast to Uint32 then to double, but Uint64 can't go straight to double
				elapsed_secs = ((double)((Uint32)(cur_total_cycles - m_play_cycles))) / cpu_hz;

			}

			// else the CPU is momentarily lagged
			else
			{
				//			printline("ldp.cpp get_current_frame() ERROR : timers were flushed, this should NEVER happen, fix it now!");
				//			elapsed_secs = elapsed_ms_time(m_play_time) * 0.001; // compute elapsed seconds
				elapsed_secs = 0.0;
			}

		}
		// if our game has no cpu (seektest for example) then we cannot compute the current frame based on
		// cpu cycles and must the less safe (but more accurate) method of elapsed time instead
		else
		{
			elapsed_secs = elapsed_ms_time(m_play_time) * 0.001; // compute elapsed seconds
		}

		result = (Uint16) (m_last_seeked_frame + (Uint16) (elapsed_secs * g_game->get_disc_fps()));
		// do not add 0.5, we do NOT want to round to the nearest number in this case

#ifdef DEBUG
		// SAFETY CHECK
		// If we are using cpu_hz to calculate the current frame, this ensures that
		// the number that is calculated is not far off what it should be.
		if (cpu_hz)
		{
			elapsed_secs = elapsed_ms_time(m_play_time) * 0.001; // compute elapsed seconds
			Uint16 time_result = (Uint16) (m_last_seeked_frame + (Uint16) (elapsed_secs * g_game->get_disc_fps()));
			Uint16 diff = 0;

			if (result > time_result) diff = result - time_result;
			else diff = time_result - result;

			// if the difference is noticeable, then alert the developer
			if (diff > 5)
			{
				static unsigned int last_warning = 0;

				if (elapsed_ms_time(last_warning) > 1000)
				{
					string s = "WARNING : cycle frame is ";
					s += numstr::ToStr(result) + " but time frame is ";
					s += numstr::ToStr(time_result) + " which has a diff of ";
					s += numstr::ToStr(diff);
					printline(s.c_str());
					last_warning = refresh_ms_time();
				}
			}
		}
#endif

	}
	// otherwise the disc is idle, so return the frame that we last seeked to
	else
	{
		result = m_last_seeked_frame;
	}

	return(result);
}

// returns value of blitting_allowed.  VLDP does not allow blitting!
bool ldp::is_blitting_allowed()
{
	return (blitting_allowed);
}

// returns the current LDP status
// This should ONLY be called if we are waiting for a search to complete, or
// if we don't have read access to 'm_status'
int ldp::get_status()
{
	// if we are in the middle of a search, find out if the search has completed
	if (m_status == LDP_SEARCHING)
	{
#ifdef DEBUG
		// verify that get_search_result() is not being called when it ought not be
		assert(!m_dont_get_search_result);
#endif
		int stat = get_search_result();

		// if the search was successful
		if (stat == SEARCH_SUCCESS)
		{
			m_dont_get_search_result = true;
			m_last_seeked_frame = m_last_try_frame;
			m_status = LDP_PAUSED;

			// Update sram after every search if user desires it
			//  (so that the DAPHNE can be improperly terminated, if it's inside a cab and powered off, for example)
			// This is the best place to do this because it's right after a seek which can take a second or two anyway.
			if (m_sram_continuous_update)
			{
				g_game->save_sram();
			}

		}

		// if the search failed
		else if (stat == SEARCH_FAIL)
		{
			m_status = LDP_ERROR;
		}
		// else the search is busy and still going, so we needn't act ...
	}

	return m_status;
}

// converts an integer frame number into ASCII (with leading zeroes)
// NOTE : 'f' should be an array of 6 characters (5 numbers plus null terminator)
void ldp::framenum_to_frame(Uint16 num, char *f)
{
    sprintf(f, "%05d", num);
}

Uint32 ldp::get_search_latency()
{
	return(search_latency);
}

void ldp::set_search_latency(Uint32 value)
{
	search_latency = value;
}

void ldp::set_stop_on_quit(bool value)
{
	m_stop_on_quit = value;
}

// causes LDP to blank video while searching
void ldp::set_search_blanking(bool enabled)
{
	printline("NOTE : Search blanking cannot be modified with this laserdisc player!");
}

// causes LDP to blank video while skipping
void ldp::set_skip_blanking(bool enabled)
{
	printline("NOTE : Skip blanking cannot be modified with this laserdisc player!");
}

void ldp::set_seek_frames_per_ms(double value)
{
	printline("NOTE : Seek delay is not supported with this laserdisc player!");
}

void ldp::set_min_seek_delay(unsigned int value)
{
	printline("NOTE : Seek delay is not supported with this laserdisc player!");
}

// causes sram to be saved after every seek
void ldp::set_sram_continuous_update(bool value)
{
	m_sram_continuous_update = value;
}

void ldp::enable_audio1()
{
	printline("Audio1 enable received (ignored)");
}

void ldp::enable_audio2()
{
	printline("Audio2 enable received (ignored)");
}

void ldp::disable_audio1()
{
	printline("Audio1 disable received (ignored)");
}

void ldp::disable_audio2()
{
	printline("Audio2 disable received (ignored)");
}

// asks LDP to take a screenshot if that's possible
// it's only possible with VLDP as of this time
void ldp::request_screenshot()
{
	printline("NOTE: current laserdisc player does not support taking screenshots, sorry");
}

// returns the width of the laserdisc video (only meaningful with mpeg)
Uint32 ldp::get_discvideo_width()
{
	return m_discvideo_width;
}

// returns the height of the laserdisc video (only meaningful with mpeg)
Uint32 ldp::get_discvideo_height()
{
	return m_discvideo_height;
}

// ordinarily does nothing unless we're VLDP
bool ldp::lock_overlay(Uint32 timeout)
{
	return true;
}

// does nothing unless we're VLDP
bool ldp::unlock_overlay(Uint32 timeout)
{
	return true;
}

// sets the value of this boolean
void ldp::set_use_nonblocking_searching(bool val)
{
	m_use_nonblocking_searching = val;
}

// returns the value of this boolean
bool ldp::get_use_nonblocking_searching()
{
	return m_use_nonblocking_searching;
}

// this is called by cmdline.cpp if it gets any cmdline parameters that it doesn't recognize
// returns true if this argument was recognized and processed,
// or false is this argument wasn't recognized (and therefore the command line is bad on a whole)
bool ldp::handle_cmdline_arg(const char *arg)
{
	return false;
}

void ldp::get_bug_log(list<string> &log)
{
	log = m_bug_log;
	m_bug_log.clear();
}
