/*
 * ldp.h
 *
 * Copyright (C) 2001 Matt Ownby
 *
 * This file is part of DAPHNE, a laserdisc arcade game emulator
 *
 * DAPHNE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DAPHNE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef LDP_PRE_H
#define LDP_PRE_H

// different states that the laserdisc player could be in
enum
{
	LDP_ERROR, LDP_SEARCHING, LDP_STOPPED, LDP_PLAYING, LDP_PAUSED
};

// different non-blocking search results
enum
{
	SEARCH_FAIL,	// search failed
	SEARCH_SUCCESS,	// search succeeded
	SEARCH_BUSY		// search is still taking place, no change yet, my frenid
};

#include <SDL.h>	// needed for datatypes

// for bug logging
#include <string>
#include <list>

using namespace std;

#define FRAME_SIZE 5

// the size to make your frame array (the frame + the NULL terminator)
#define FRAME_ARRAY_SIZE FRAME_SIZE + 1

class ldp
{
public:
	ldp();
	virtual ~ldp();
	
	// Call this function to initialize the ldp (the constructor does _not_ call it).
	// It's safe to call this function mulitple times.
	bool pre_init();
	
	// player-specific init stuff
	// NOTE : this function should not be called directly, call pre_init instead
	virtual bool init_player();
	
	// Call this function to shutdown the ldp (the destructor also calls it).
	// It's safe to call this function multiple times.
	void pre_shutdown();

	// player-specific shutdown stuff	
	// NOTE : this function should not be called directly, call pre_shutdown instead
	virtual void shutdown_player();
	
	void clear();	// clears any received frames

	// searches to the 5-digit ASCII frame specified by 'pszFrame'
	// if 'block_until_search_finished' is true, this function will return after the search has completed and will return
	//    true if the search was successful or false on error.
	// if 'block_until_search_finished' is false, this function will return after the search has _begun_ and will return true
	//    if the player is currently executing the search or false if there was an error
	// Call this instead of calling the search function directly!
	// WARNING : if 'block_until_search_finished' is false, then the calling function
	// MUST also call get_status() to determine when a seek has finished.
	// If you aren't prepared to do this, make 'block_until_search_finished' true.
	bool pre_search(const char *, bool block_until_search_finished);

	// does the actual search. This func returns once the search has begun and you need to call get_search_result
	// until the search isn't busy anymore.
	// This function returns true if the LDP acknowledged the search
	virtual bool nonblocking_search(char *);

	// do not call this function directly! (get get_status instead)
	// This is the LDP-specific function that checks to see if a search (issued previously) finished
	// it returns one of the SEARCH_* enumerations ...
	virtual int get_search_result();

	bool pre_skip_forward(Uint16);
	bool pre_skip_backward(Uint16);

	// steps 1 frame forward
	void pre_step_forward();

	// steps 1 frame backward
	void pre_step_backward();

	virtual bool skip_forward(Uint16 frames_to_skip, Uint16 target_frame);
	// NOTE : frames_to_skip and target_frame are both included for your convenience
	// use frames_to_skip only if you cannot use target_frame (which is more accurate)

	virtual bool skip_backward(Uint16 frames_to_skip, Uint16 target_frame);
	// NOTE : frames_to_skip and target_frame are both included for your convenience
	// use frames_to_skip only if you cannot use target_frame (which is more accurate)
	
	void pre_play();
	void pre_pause();
	void pre_stop();

	// LASERDISC-PLAYER-SPECIFIC IMPLEMENTATIONS OF VARIOUS FUNCTIONALITY
	virtual unsigned int play();
	virtual void set_play_speed(Uint8);	// sets the speed we playback at
	virtual void pause();
	virtual void stop();

	// returns the current frame number that the disc is on
	// this is a generic function which computes the current frame number using the elapsed time
	// and the framerate of the disc.  Obviously querying the laserdisc player would be preferable
	// if possible (some laserdisc players don't like to be queried too often)
	// SEEKING NOTES:
	// if get_current_frames() is called during a NON-BLOCKING seek,
	// it returns the frame the LDP was at right 
	// when seek was begun. It DOES NOT check to see if a seek is finished, which
	// means that any driver that uses non-blocking seeking MUST also call get_status()
	// to check for the completion of that seek.
	// A laserdisc driver that supports non-blocking seeking (LD-V1000) SHOULD call
	// get_status() to check to see if a search is complete.  If an LDP isn't designed
	// to return its status (ie the PR-8210) then either blocking seeking must be used
	// or each game driver which uses this player must call get_status().
	virtual Uint16 get_current_frame();

	// causes sram to be saved after every seek
	virtual void set_sram_continuous_update(bool value);

	virtual void enable_audio1();
	virtual void enable_audio2();
	virtual void disable_audio1();
	virtual void disable_audio2();
	virtual void request_screenshot();
	virtual void set_search_blanking(bool enabled);
	virtual void set_skip_blanking(bool enabled);
	virtual void set_seek_frames_per_ms(double value);
	virtual void set_min_seek_delay(unsigned int value);

	// END LDP-SPECIFIC SECTION

	bool is_blitting_allowed();	// returns value of blitting_allowed
	int get_status();	// returns status of laserdisc player
	void framenum_to_frame(Uint16, char *);	// converts int to 5-digit string
	Uint32 get_search_latency();
	void set_search_latency(unsigned int);
	void set_stop_on_quit(bool);	// enables the stop_on_quit bool flag

	Uint32 get_discvideo_height();	// gets the height of the laserdisc video (only meaningful with mpeg)
	Uint32 get_discvideo_width();	// gets the width of the laserdisc video (only meaningful with mpeg)
	virtual bool lock_overlay(Uint32);	// prevents yuv callback from being called (only meaningful with mpeg)
	virtual bool unlock_overlay(Uint32);

	// sets the value of this boolean
	void set_use_nonblocking_searching(bool);

	// returns the value of this boolean
	bool get_use_nonblocking_searching();

	// handles LDP-specific command-line arguments
	virtual bool handle_cmdline_arg(const char *arg);
	
	// Copies m_bug_log into 'log' and clears m_bug_log.
	// Used by releasetest.
	void get_bug_log(list<string> &log);

protected:
	bool need_serial;	// whether this LDP driver needs the serial port initialized
	bool serial_initialized; // whether serial has been initialized
	bool player_initialized; // whether the LDP has been properly initialized
	bool blitting_allowed;	// whether it's ok to blit directly to the screen (SMPEG forbids this)
	bool skipping_supported;	// whether the laserdisc player supports skipping
	bool skip_instead_of_search;	// whether we should skip instead of search if searching forward a short distance
	Uint16 max_skippable_frames;	// maximum # of frames that player can skip (if skipping is supported)
	Uint16 m_last_try_frame;	// the last frame we _tried_ to seek to
	Uint16 m_last_seeked_frame;	// the last frame we successfully seeked to (used with m_play_time to calculate current frame)
	Uint64 m_play_cycles;	// # of elapsed cpu cycles from when we last issued a play command
	Uint32 m_play_time;	// current time when we last issued a play command
	int m_status;	// the current status of the laserdisc player
	Uint32 search_latency;	// how many ms to stall before searching (to simulate slow laserdisc players)
	bool m_stop_on_quit;	// should the LDP stop when it quits?
	Uint32 m_discvideo_width;	// width of laserdisc video (only meaningful with mpeg)
	Uint32 m_discvideo_height;	// height of laserdisc video (only meaningful with mpeg)
	bool m_use_nonblocking_searching;	// true if ldp-in drivers should use pre_search in non-blocking mode (as of now, blocking mode is more stable but non-blocking mode is more accurate)
	bool m_dont_get_search_result;	// if we should not be calling get_search_result()
	bool m_sram_continuous_update;  // if sram is to be updated on a regular basis
	
	// timer to be used to simulate search delay when in 'noldp' mode (for debugging)
	Uint32 m_noldp_timer;
	
	// used by 'releasetest' to do automatic self-testing
	list<string> m_bug_log;
};

extern ldp *g_ldp;	// our global ldp class.  Defined here so that every .cpp file doesn't have to define it.

#endif
