// DaphneLoader C++ version (uses wxWindows for cross-platform compatibility)
// by Matt Ownby

#include "parse.h"

#include "main.h"
#include "transfers.h"
#include "prefs.h"
#include <mpolib/mpo_numstr.h>
#include <mpolib/mpo_fileio.h>
#include <mpolib/mpo_client.h>
#include <mpolib/mpo_net.h>
#include <mpolib/mpo_misc.h>
#include <mpolib/md5_global.h>
#include <mpolib/md5.h>

#ifdef WIN32
#include <windows.h>
#include <direct.h>	// for _mkdir
#define MPO_MKDIR(path) _mkdir(path)

// how many pixels to expand certain troubling controls (wxChoice)
#define WIDTH_FUDGE 0

#else
#include <sys/types.h>	// for fork
#include <unistd.h>		// for fork
#include <sys/wait.h>	// for waitpid
#define MPO_MKDIR(path) mkdir(path, 0777)
#define WIDTH_FUDGE 0
#endif

#if defined(__DARWIN__)
#include <mach-o/dyld.h>	// for _NSGetExecutablePath
#endif

#ifdef PREFS_READONLY
const char *PREFS_PATH = "wx/frontend/prefs.xml";
#else
const char *PREFS_PATH = "DaphneLoaderPrefs.xml";
#endif

///////////////////////////////////////////////////

struct the_vars g_the_vars;
class prefs g_prefs(&g_the_vars, &g_config_xml, &g_control_xml);
MyFrame *g_mainframe = NULL;	// a pointer to our MyFrame instance in case we need to access its functions
MyApp *g_mainapp = NULL;	// pointer to MyApp instance (to call wxApp::Yield from other places)

XferControl *g_xfer_control = NULL;	// accessible by all, but allocated/de-allocated by MyApp
///////////////////////////////////////////////////

BEGIN_EVENT_TABLE(MyFrame, wxFrame)
EVT_MENU(ID_Preferences, MyFrame::OnPreferences)
EVT_CLOSE(MyFrame::OnClose)
EVT_MENU(ID_Quit, MyFrame::OnQuit)
EVT_MENU(ID_About, MyFrame::OnAbout)
EVT_MENU(ID_Start, MyFrame::OnStart)
EVT_MENU(ID_Config, MyFrame::OnConfig)
EVT_MENU(ID_Transfers, MyFrame::OnTransfers)
EVT_LISTBOX(ID_List_Select, MyFrame::OnSelected)
EVT_BUTTON(ID_Config, MyFrame::OnConfig)	
EVT_BUTTON(ID_Start, MyFrame::OnStart)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(MyPreConfig, wxDialog)
EVT_BUTTON(ID_CFG_KEYS, MyPreConfig::OnConfigKeys)
EVT_BUTTON(ID_CFG_BUTTONS, MyPreConfig::OnConfigButtons)
EVT_BUTTON(ID_CFG_OK, MyPreConfig::OnOK)
EVT_BUTTON(ID_CFG_CANCEL, MyPreConfig::OnCancel)
EVT_BUTTON(ID_CFG_APPLY, MyPreConfig::OnApply)
END_EVENT_TABLE()

BEGIN_EVENT_TABLE(MyConfig, wxNotebook)
EVT_NOTEBOOK_PAGE_CHANGED(ID_Notebook, MyConfig::OnTabChange)
EVT_CHOICE(ID_GAMEVER_CHOICES, MyConfig::OnGameVerType)
EVT_CHOICE(ID_PLAYER_TYPE, MyConfig::OnPlayerType)
EVT_CHOICE(ID_VLDP_CHOICES, MyConfig::OnVLDPChoices)
EVT_BUTTON(ID_BROWSE_FRAMEFILE, MyConfig::OnBrowseFramefile)
END_EVENT_TABLE()

IMPLEMENT_APP(MyApp)

#include <string.h>

// takes a pathname (relative or absolute) + a filename and removes the filename itself
// therefore it returns just the path (stored in 'path')
// !!! We assume that the 'path' buffer's size is >= 'file_with_path' size
// returns false if there is an error (such as the file having no path)
bool get_path_of_file(char *file_with_path, char *path)
{
	bool success = false;
	int index = strlen(file_with_path) - 1;	// start on last character

	// make sure the file_with_path is at least 2 characters long
	// because otherwise our proceeding calculations could cause a segfault
	if (index > 0)
	{
		// locate the preceeding / or \ character
		while ((index >= 0) && (file_with_path[index] != '/') && (file_with_path[index] != '\\'))
		{
			index--;
		}

		// if we found a leading / or \ character
		if (index >= 0)
		{
			strcpy(path, file_with_path);
			path[index+1] = 0;	// chop off filename from string (we already know
			success = true;
		}
	}

	return success;
}

bool MyApp::OnInit()
{
	bool result = false;

	g_mainapp = this;	// to make calling wxApp::Yield possible ...

	net_init();	// needs to be called to get our networking started

#if defined(__DARWIN__)
	char file[160], path[160];
	unsigned long l = sizeof(file);
	_NSGetExecutablePath(file, &l);
	get_path_of_file(file, path);	// isolate the path ...
	chdir(path);	// change to the directory where the loader is stored
	chdir("../../..");	// move out of application bundle schlop
	//	printf("Changed to dir %s\n", path);
#else
	// this may work for MacOSX also, we need to see ...
	wxSetWorkingDirectory(wxPathOnly(argv[0]));	// set working directory to the executable's directory
	// We do this so that we can access other files relative to our .exe's position
#endif

	// load the config file
	if (load_config("config.xml"))
	{
		m_frame = new MyFrame( "Daphne Loader v2.0 (C++ version)", wxPoint(50,50), wxSize(420,285) );
		m_frame->Show(TRUE);
		SetTopWindow(m_frame);

#ifdef DEBUG
		// after frame has been displayed, try to do this ...
		if (m_frame->load_control_file())
		{
			// we only can load prefs if the control file loaded properly
			m_frame->load_prefs_file();	// this needs to be done after control file has been displayed
			m_frame->apply_the_vars();	// now that prefs are loaded, apply them
		}
#endif

		/*
		// hopefully this warning will deter some questions ...
		wxMessageBox("This program is VERY incomplete, almost all features are missing.\n"
		"It will have more features when someone decides to do the work to add them.",
		"WARNING!", wxOK | wxICON_ERROR, m_frame);
		*/

		wxMessageBox("This is just a bit torrent test. Click on Game->View Transfers to watch the progress.\n"
		"You don't need to do anything except leave this program running until your download finishes.",
		"RELAX!", wxOK, m_frame);

		// XferControl cannot be created if g_mainframe is NULL, so it must be done inside here
		g_xfer_control = new XferControl();	// start transferring in the background

		result = true;
	}
	else
	{
		wxMessageBox("Could not open DaphneLoader config file!",
			"Fatal error!", wxOK | wxICON_ERROR, 0);
	}

	return result;
}

// gets called right before application exits
int MyApp::OnExit()
{
	g_prefs.save(PREFS_PATH);

	if (g_xfer_control != NULL)
	{
		delete g_xfer_control;
		g_xfer_control = NULL;
	}

	net_shutdown();	// needs to be called when our program is done with networking

	return 0;
}

void MyApp::do_yield()
{
	Yield();
}

bool MyApp::load_config(const char *filename)
{
	bool result = false;
	unsigned char *buf = NULL;
	mpo_io *io = mpo_open(filename, MPO_OPEN_READONLY);

	if (io)
	{
		buf = new unsigned char[io->size];
		if (buf)
		{
			if (mpo_read(buf, (size_t) io->size, NULL, io))
			{
				string err_msg = "";
				result = parse_cfg_file(buf, (unsigned int) io->size, err_msg);
				if (!result)
				{
					wxMessageBox(err_msg.c_str(),
						"Config File Error!", wxOK | wxICON_ERROR, 0);
				}
			}
			delete [] buf;
		}
		mpo_close(io);
	}

	return result;
}

// runs a new program in a new process
// the name of the executable should be the first parameter in the cmd_line
// if 'wait_for_process' is true, then this function won't return until the child process
// has finished (the exit code will be stored in 'exit_code'
// otherwise, 'exit_code' is undefined
// 'exit_code' may be NULL if you don't care about what it is
bool spawn_executable(vector<string> cmd_line, bool wait_for_process, int *exit_code)
{
	bool result = false;
	exit_code = 0;

	// windows and unix have different methods for launching a new process
#ifdef WIN32
	char cmd_line_buf[1024];	// because createprocess won't accept const char *

	// windows wants the command line in one big string
	string cmd_s = "";
	// add command line to one big line win32 string ...
	for (int i = 0; i != cmd_line.size(); i++)
	{
		cmd_s += cmd_line[i];
		cmd_s += " ";
	}

	strncpy(cmd_line_buf, cmd_s.c_str(), sizeof(cmd_line_buf));
	cmd_line_buf[sizeof(cmd_line_buf) - 1] = 0;	// terminate string

	PROCESS_INFORMATION process_info;
	STARTUPINFO startup_info;
	memset(&startup_info, 0, sizeof(startup_info));
	startup_info.cb = sizeof(startup_info);
	BOOL b_result = CreateProcess(cmd_line[0].c_str(),
		cmd_line_buf, 
		NULL,	// process attributes
		NULL,	// thread attributes
		FALSE,	// inherit handles?
		0,	// creation flags
		NULL,	// new environment
		NULL,	// current directory (same as us)
		&startup_info,	// startup info
		&process_info);	// process information
	if (b_result != 0) result = true;

	// if we are supposed to wait for the process to complete
	if (wait_for_process)
	{
		DWORD wait_result = WAIT_TIMEOUT;

		// wait for process to exit
		while (wait_result == WAIT_TIMEOUT)
		{
			wait_result = WaitForSingleObject(process_info.hProcess, 125);
			wxSafeYield();	// update screen
		}
		// if the exit code has been requested
		if (exit_code != NULL)
		{
			DWORD tmp = 0;
			GetExitCodeProcess(process_info.hProcess, &tmp);
			*exit_code = tmp;	// may lose precision here, but exit codes are usually small
		}
		CloseHandle(process_info.hProcess);	// not sure if this is necessary, but some example code showed it being used
	}

#else
	const char **argv = (const char **) malloc(sizeof (const char *) * (cmd_line.size() + 1));
	// allocate memory to store command line, add an extra slot for NULL termination

	// add command line to one big line win32 string ...
	for (size_t i = 0; i != cmd_line.size(); i++)
	{
		argv[i] = cmd_line[i].c_str();
	}
	argv[cmd_line.size()] = NULL;	// final null terminating entry

	// fork and exec
	pid_t pid = fork();

	// if we are the child process
	if (pid == 0)
	{
		// if exec failed ...
		if (execv(argv[0], (char* const*) argv) < 0)
		{
			perror("Execv failed, error is");

			// get rid of wxWindows callbacks
			atexit(NULL);
			//			on_exit(NULL, NULL);
			exit(1);	// critical shutdown ...
		}
	}

	// parent process
	else
	{
		result = true;	// FIXME : we always succeed for now, but is that really what we want to do?
		
		// if we are supposed to wait for child process to complete
		if (wait_for_process)
		{
			for (;;)
			{
				pid_t p = waitpid(pid, NULL, WNOHANG);	// check to see if child has exited, but don't wait
				
				// some error
				if (p == -1)
				{
					result= false;
					break;
				}
				// else if the child we were waiting for is done, we're done too
				else if (p == pid) break;
					// else p is either 0 ('timed out') or an unknown
				else
				{
					wxSafeYield();	// update screen
					wxUsleep(125);
				}
			} // end for loop
		} // end if we are to wait

	}	

	free(argv);
#endif

	return result;
}

MyFrame::MyFrame(const wxString& title, const wxPoint& pos, const wxSize& size)
: wxFrame((wxFrame *)NULL, -1, title, pos, size),
m_menuFile(NULL), m_menuHelp(NULL), m_menuBar(NULL),
m_panel(NULL), m_listbox(NULL), m_start_button(NULL), m_config_button(NULL),
m_canvas(NULL), m_xfer_frame(NULL)
{
	m_the_vars = &g_the_vars;

	g_mainframe = this;

	m_menuFile = new wxMenu();
	m_menuFile->Append( ID_Preferences, "Preferences" );
	m_menuFile->AppendSeparator();
	m_menuFile->Append( ID_Quit, "E&xit" );

	m_menuGame = new wxMenu();
	m_menuGame->Append( ID_Start, "Start the game");
	//	m_menuGame->Append( ID_Watch, "Watch the laserdisc associated with selected game");
	m_menuGame->AppendSeparator();
	m_menuGame->Append( ID_Config, "Configure DAPHNE options");
	m_menuGame->Append( ID_Transfers, "View Transfers");

	m_menuHelp = new wxMenu();
	//	m_menuHelp->Append( ID_Web_Site, "Daphne Web Site" );
	//	m_menuHelp->Append( ID_Documentation, "Daphne Documentation" );
	//	m_menuHelp->AppendSeparator();
	m_menuHelp->Append( ID_About, "About Daphne Loader" );

	m_menuBar = new wxMenuBar();
	m_menuBar->Append( m_menuFile, "&File" );
	m_menuBar->Append( m_menuGame, "&Game" );
	m_menuBar->Append( m_menuHelp, "&Help" );

	SetMenuBar( m_menuBar );

	// child controls
	m_panel = new wxPanel(this);
	m_listbox = new wxListBox(m_panel, ID_List_Select, wxPoint(5, 5), wxSize(235, 120));

	m_start_button = new wxButton(m_panel, ID_Start, "&Start!");
	m_config_button = new wxButton(m_panel, ID_Config, "&Configure");

	m_canvas = new MyCanvas(m_panel, wxPoint(-1,-1), wxSize(160, 120));

	m_daphne_logo = new MyCanvas(m_panel, wxPoint(-1, -1), wxSize(164, 52));	// fixme, find a way to fit around image
	m_daphne_logo->new_image("logo.jpg");

	size_t game_count = g_config_xml.m_games.size();
	size_t cur_game = 0;

	// BEGIN GAME VARIABLE INITIALIZATION IN PREPARATION FOR LOADING USER SETTINGS
	// add each game name to our listbox and initialize it
	for (cur_game = 0; cur_game < game_count; cur_game++)
	{
		m_listbox->Append(g_config_xml.m_games[cur_game].m_name.c_str());

		game_vars g;
		g.adv_display_cmd_line = true;
		g.adv_nocrc = false;
		g.adv_noserversend = true;
		g.aud_buffer = 2048;
		g.aud_nosound = 0;
		g.game_version_index = 0;	// default to 'auto'
		g.disc_name = "??";	// OnSelected handles this

		g.ldp_type = LDP_TYPE_VLDP;	// default to VLDP
		g.ldp_driver = "vldp";	// default to VLDP
		g.ldp_port = 0;	// default serial port

		g.vid_fullscreen = false;
		g.vid_blank_searches = false;
		g.vid_blank_searches = true;
		g.vid_blend = false;
		g.vid_nohwaccel = false;
		g.vid_scanlines = false;
		g.vid_x = 640;
		g.vid_y = 480;

		g.vldp_download_index = 0;	// default to automatic version
		g.vldp_framefile = "";
		g.vldp_autodownload = true;	// default to automatic downloads

		m_the_vars->games.push_back(g);	// add this game to our variable vector
	}

	// END GAME VARIABLE INITIALIZATION

	// BEGIN CONFIGS INITIALIZATION IN PREPARATION FOR LOADING USER SETTINGS
	for (size_t conf_index = 0; conf_index < g_config_xml.m_games_cfg.size(); conf_index++)
	{
		const struct cfg_gamecfg *cfg = &g_config_xml.m_games_cfg[conf_index];
		conf_vars c;
		c.conf_name = cfg->m_name;

		// go through every list box and create an entry
		for (size_t lb_index = 0; lb_index < cfg->m_listboxes.size(); lb_index++)
		{
			size_t val = 0;	// initialize to "Auto"
			c.listbox_indices.push_back(val);	// add to this listbox
		}

		m_the_vars->configs.push_back(c);	// add our new config
	}
	// END CONFIGS INITIALIZATION

	// BEGIN GLOBAL PREFS INITIALIZATION IN PREPARATION FOR LOADING USER SETTINGS
	
	m_the_vars->game_index = 0;	// select first item
	m_the_vars->tab_index = 0;	// select first tab

	m_the_vars->ul_bytes_per_s = 0;	// no limit
	m_the_vars->dl_bytes_per_s = 0;	// " " "
	m_the_vars->leech_resistance = 0;	// 'auto'
	m_the_vars->max_peers = 50;	// 50 max peers	
	m_the_vars->bt_listen_port = 0;	// port to listen for incoming bit torrent connections (0 means scan a default range)

	// END GLOBAL PREFS INITIALIZATION
	
	// DO SIZERS NOW
	m_sizer_main = new wxBoxSizer(wxHORIZONTAL);
	m_sizer_right = new wxBoxSizer(wxVERTICAL);
	m_sizer_buttons = new wxBoxSizer(wxHORIZONTAL);

	// the button row ...
	m_sizer_buttons->Add(m_start_button, 0, 0, 0);
	m_sizer_buttons->Add(m_config_button, 0, 0, 0);

	// the right portion
	m_sizer_right->Add(m_canvas, 0, 0, 0);
	m_sizer_right->Add(0, 10, 0);	// spacer
	m_sizer_right->Add(m_sizer_buttons, 0, wxALIGN_CENTER_HORIZONTAL, 0);
	m_sizer_right->Add(0, 10, 0);	// spacer
	m_sizer_right->Add(m_daphne_logo, 0, wxALIGN_CENTER, 0);
	// add more stuff later like a little logo

	m_sizer_main->Add(m_listbox, 1, wxALL | wxEXPAND | wxGROW, 5);	// add the list of games here ...
	m_sizer_main->Add(m_sizer_right, 0, wxALL, 5);	// add our right panel

	m_panel->SetSizer(m_sizer_main);
	m_panel->Fit();
	this->Fit();
}

void MyFrame::OnPreferences(wxCommandEvent& WXUNUSED(event))
{
	GUIPrefs my_prefs(this);
	my_prefs.ShowModal();	// show the modal dialog box ...
}

void MyFrame::OnClose(wxCloseEvent& WXUNUSED(event))
{
	// destroy xfer window if it is open
	if (m_xfer_frame != NULL)
	{
		m_xfer_frame->Destroy();
	}

	Destroy();
}

void MyFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
	Close(TRUE);
}

void MyFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
	wxMessageBox("Daphne Loader (C++ version) by Matt Ownby",
		"About Daphne Loader", wxOK | wxICON_INFORMATION, this);
}

void MyFrame::OnStart(wxCommandEvent &event)
{
	// get the game that we currently have selected
	const struct game_vars *g = get_active_gamevars();

	// check and update dependencies, and if it fails, do NOT try to start daphne!
	if (!UpdateDependencies()) return;

	// make sure the queue is empty (it should be)
	m_cmd_line.clear();

#ifdef WIN32
	AddCmdLine("daphne.exe");	// name of the executable
#else
	AddCmdLine("daphne");	// name of the executable
#endif

	// verify that the EXE exists before we proceed any further ...
	FILE *F = fopen(m_cmd_line[0].c_str(), "rb");

	// if the file exists where we expect it ...
	if (F)
	{
		fclose(F);
	}

	// if the file does not exist ...
	else
	{
		wxMessageBox("The DAPHNE executable was not found in the current directory.",
			"Unable to start!", wxOK | wxICON_ERROR, this);
		return;
	}

	// add the name of the game driver (this should never fail if config file isn't corrupt)
	AddCmdLine(get_active_game_version(g->game_version_index)->m_driver);

	// add the name of the LDP driver
	AddCmdLine(g->ldp_driver);

	// if we also need to include a framefile ...
	if (g->ldp_driver == "vldp")
	{
		// UpdateDependencies already verified that everything is in order here ...

		AddCmdLine("-framefile");

		string framefile = g->vldp_framefile;
		// if they have specified auto downloading, the auto framefile overrides the manual one
		if (g->vldp_autodownload)
		{
			const ctrl_version *v = get_active_disc_version();
			// this should always succeed since UpdateDeps returned true
			framefile = make_autodownload_path(v) + "/" + v->m_framefile;
		}

		// windows receives its command line as one big string, so it needs quotes surrounding
		// the framefile path (in case of spaces)		
#ifdef WIN32
		string inquotes = "\"";
		inquotes += framefile + "\"";
		AddCmdLine(inquotes);
#else
		AddCmdLine(framefile);
#endif

		// VLDP-specific stuff which is otherwise ignored
		if (g->vid_blend) AddCmdLine("-blend");
		if (g->vid_scanlines) AddCmdLine("-scanlines");
		if (g->vid_blank_searches) AddCmdLine("-blank_searches");
		if (g->vid_blank_skips) AddCmdLine("-blank_skips");
	}

	// various command line options
	if (g->adv_nocrc) AddCmdLine("-nocrc");
	if (g->vid_nohwaccel) AddCmdLine("-nohwaccel");
	if (g->aud_nosound) AddCmdLine("-nosound");
	if (g->adv_noserversend) AddCmdLine("-noserversend");
	if (g->vid_fullscreen) AddCmdLine("-fullscreen");

	// setting up the command line like this is used by WIN32 AND if the user has requested to
	// display the command line, so we will always do it even if we're in linux and they didn't request it
	string cmd_line = "";
	// add command line to one big line win32 string ...
	for (size_t i = 0; i != m_cmd_line.size(); i++)
	{
		cmd_line += m_cmd_line[i];
		cmd_line += " ";
	}

	if (g->adv_display_cmd_line)
	{
		wxMessageBox(cmd_line.c_str(), "The command line is:", wxOK, this);
	}

	spawn_executable(m_cmd_line, false, NULL);	// do it!
}

const struct ctrl_version *MyFrame::get_active_disc_version()
{
	const struct ctrl_version *result = NULL;
	const ctrl_disc *disc = control_find_disc(&g_control_xml, get_active_gamevars()->disc_name);
	if (disc)
	{
		// -1 because first item is 1, auto is 0
		int v_index = get_active_gamevars()->vldp_download_index - 1;	// THIS MUST BE SIGNED!

		// if they chose auto, choose first item
		if (v_index < 0) v_index = 0;

		// range check: make sure our selected index exists
		if (disc->m_versions.size() > (unsigned int) v_index)
		{
			result = &disc->m_versions[v_index];
		}
		// else this disc has no versions or the version is out of range for whatever reason
	}
	// else the disc entry can't be found, hence the version cannot be found
	return result;
}

const struct cfg_gamever *MyFrame::get_active_game_version(int gamever_index)
{
	const struct cfg_gamever *result = NULL;

	// -1 because first item is 1, auto is 0
	int g_index = gamever_index - 1;	// THIS MUST BE SIGNED!

	// if they chose auto, choose first item
	if (g_index < 0) g_index = 0;

	// range check: make sure our selected index exists
	if (g_config_xml.m_games[m_the_vars->game_index].m_versions.size() > (unsigned int) g_index)
	{
		result = &g_config_xml.m_games[m_the_vars->game_index].m_versions[g_index];
	}
	// else out of range

	return result;
}

const struct cfg_gamecfg *MyFrame::get_active_gamever_cfg(int gamever_index)
{
	const struct cfg_gamecfg *result = NULL;
	const struct cfg_gamever *g = get_active_game_version(gamever_index);
	size_t i = 0;

	// safety check
	if (g)
	{
		// look for matching config
		// go until we run out of configs or find a match
		for (i = 0; (i < g_config_xml.m_games_cfg.size()) && (!result); i++)
		{
			// if the config this game version uses matches up with the config we're currently examining
			if (g_config_xml.m_games_cfg[i].m_name == g->m_cfg)
			{
				result = &g_config_xml.m_games_cfg[i];
			}
		} // end for loop
	}
	// else we'll be returning NULL anyway

	return result;
}

const struct ctrl_version *MyFrame::get_rom_version()
{
	const struct ctrl_version *result = NULL;

	// FIXME : do not assume there is 1 entry and do not assume it's the first one in the list!
	result = &g_control_xml.m_roms[0];

	return result;
}

const struct game_vars *MyFrame::get_active_gamevars()
{
	return &m_the_vars->games[m_the_vars->game_index];
}

const struct conf_vars *MyFrame::get_confvars(const char *name)
{
	const conf_vars *result = NULL;
	size_t i;

	for (i = 0; i < m_the_vars->configs.size(); i++)
	{
		if (m_the_vars->configs[i].conf_name == name)
		{
			result = &m_the_vars->configs[i];
			break;
		}
	}

	return result;
}

//
bool MyFrame::update_helper(const ctrl_version *v, const char *file_description)
{
	bool result = false;

	if (v)
	{
		dl_status stat = get_autodownload_status(v);
		switch (stat)
		{
		case DLSTAT_FILE_MISSING:
		case DLSTAT_LENGTH_WRONG:
		case DLSTAT_CHECKSUM_WRONG:
			// prompt them, and if download and verification succeeded, we're ready to run daphne!
			if (prompt_download_and_verify(v, file_description))
			{
				result = true;
			}
			// else they canceled
			break;
		case DLSTAT_FILE_UNVERIFIED:
			// prompt them.. and if verification succeeds, we're ready to run daphne
			if (prompt_verify(v, file_description))
			{
				result = true;
			}
			break;
		case DLSTAT_FILE_VERIFIED:
			result = true;
			break;
		default:	// DLSTAT_ERROR
			wxMessageBox("Unknown error when checking dependencies!\n",
				"Fatal Error", wxOK | wxICON_ERROR, this);
			break;
		}
	}

	return result;
}

bool MyFrame::UpdateDependencies()
{
	bool result = false;	// it makes sense to default to false, that makes sure nothing unauthorized slips through the cracks

	const struct game_vars *g = get_active_gamevars();

	// START LDP TEST

	// if we're VLDP
	if (g->ldp_type == LDP_TYPE_VLDP)
	{
		// if we've requested auto-downloading
		if (g->vldp_autodownload)
		{
			const ctrl_version *v = get_active_disc_version();
			if (v)
			{
				result = update_helper(v, "Video/Audio");
			}
			// else version could not be found
			else
			{
				// if there is a MOTD, display it to clue the user in on what's going on
				const ctrl_disc *disc = control_find_disc(&g_control_xml, g->disc_name);
				if (disc && (disc->m_motd != ""))
				{
					string msg = "DAPHNE cannot get video/audio files for this game automatically.\n\n";
					msg += disc->m_motd;
					do_error(msg.c_str());
				}
				// no MOTD, so just print the usual message
				else
				{
					// Instead of taking newbs to the config section, just
					// give them an error so they will give up.
					// It's not that their config is wrong, it's that something on the
					// server may be wrong.
					do_error("This game cannot be run.  Required files are missing.");
				}
			}
		}
		// else make sure framefile is present
		else
		{
			if (g->vldp_framefile == "")
			{
				string msg = "You have chosen to provide video/audio files manually,\n"
					"but you have not specified a framefile.\n"
					"You will be taken to the configuration section to fix the problem.";
				if (wxMessageBox(msg.c_str(),
					"Framefile Missing", wxOK | wxCANCEL, this) == wxOK)
				{
					wxCommandEvent dummy;	// unused
					OnConfig(dummy);	// pop-open config screen for them to fiddle around with
				}
				// else they didn't wanna go to the config screen ...
			}
			// else, framefile is present ...
			else
			{
				// ADDME : verify that the framefile is legitimate? NAHHH that's what Daphne is for :)
				result = true;
			}
		}
	}
	// else no VLDP

	// else if this is a serial connection, there is nothing we can test, so just let it pass
	else if (g->ldp_type == LDP_TYPE_SERIAL) result = true;
	
	// else if they have requested testing mode, allow them to pass
	else if (g->ldp_type == LDP_TYPE_NONE) result = true;

	// else their ldp_type is LDP_TYPE_UNKNOWN which should never happen, so do an error :)

	// END LDP CHECK

	// only keep doing checks if we've passed the previous tests
	if (result)
	{
		result = false;	// reset ...

		// CHECK ROMS
		const ctrl_version *roms = get_rom_version();

		// if we have found some roms that we can validate
		if (roms)
		{
			result = update_helper(roms, "ROM Image");
		}
		// END ROM CHECK
	}

	return result;
}

bool MyFrame::prompt_download_and_verify(const struct ctrl_version *v, const char *file_description)
{
	bool result = false;
	string msg = "DAPHNE will now try to get required files from the internet.";
	string title = file_description;

	title += " Files Needed";

	if (wxMessageBox(msg.c_str(),
		title.c_str(), wxOK | wxCANCEL, this) == wxOK)
	{
		// delete all incorrect *.md5 files because they will become obsolete after bit torrent has run
		for (size_t index = 0; index < v->m_files.size(); index++)
		{
			string path = make_autodownload_path(v);

			// we don't want to delete .md5 files if they indicate that the file is correct,
			// because if the file is already correct, it won't be modified
			if (get_status_one_file(path, &v->m_files[index]) != DLSTAT_FILE_VERIFIED)
			{
				string fullpath = make_md5_fullpath(v, index);
				unlink(fullpath.c_str());	// delete file
			}
		}

		result = do_download(v);
		if (result)
		{
			result = false;	// new results...
			dl_status stat = get_autodownload_status(v);
			switch (stat)
			{
				// this is the status that we expect
			case DLSTAT_FILE_UNVERIFIED:
				result = prompt_verify(v, file_description);	// now move to verification stage
				break;
				// this should never happen, but if it does.. maybe we can find out about it
			case DLSTAT_FILE_VERIFIED:
				do_error("Bug in DaphneLoader detected.\nFiles were verified after download, before verification.\nPlease report this.");
				break;
			default:
				msg = "DAPHNE's attempt to get files from the internet failed.\nYou should try again later.";
				do_error(msg.c_str());
				break;
			}
		}
		// else it didn't work ...
	}
	// else they didn't want to ...
	return result;
}

bool MyFrame::download_torrent(const struct ctrl_version *v)
{
	bool result = false;
	size_t count = v->m_torrents.size();
	string err_msg = "";	// so we can give a descriptive err msg to the user

	// if we have at least 1 torrent for this version
	if (count > 0)
	{
		string tmp_buf = "";
		// start at a random mirror
		unsigned int start_index = (unsigned int) (wxGetLocalTime() % count);
		unsigned int cur_index = start_index;

		// go through the list of available torrents until we successfully download one,
		// or run out of torrents to try downloading
		do
		{
			// if we found a URL and if the download was successful
			if ((v->m_torrents[cur_index].m_url != "") &&
				(download_file(v->m_torrents[cur_index].m_url.c_str(), "Downloading...", tmp_buf)))
			{
#ifdef DEBUG
				// when debugging, we may be altering the .torrents constantly
				if (true)
#else
				// if the .torrent file is the proper length
				// AND
				// if upper-case MD5 of what we downloaded is equal to
				// the upper-case MD5 of what the server says we should've downloaded
				if ((numstr::ToStr(tmp_buf.size()) == v->m_torrents[cur_index].m_size)
					&&
					(mpom::str_toupper(mpom::compute_md5((const unsigned char *) tmp_buf.data(),
					tmp_buf.size())) == mpom::str_toupper(v->m_torrents[cur_index].m_md5))
					)
#endif
				{
					MPO_MKDIR(v->m_rootdir.c_str());	// make sure root dir exists
					string subdir = g_mainframe->make_autodownload_path(v);
					MPO_MKDIR(subdir.c_str());	// make it if it doesn't exist

					// put torrent inside root VLDP directory, name it after the directory it will create
					string filename = v->m_rootdir;
					filename += "/" + v->m_dir + ".torrent";

					// try to write the file to disk
					mpo_io *io = mpo_open(filename.c_str(), MPO_OPEN_CREATE);
					if (io)
					{
						unsigned int bytes_written = 0;
						if (mpo_write(tmp_buf.data(), tmp_buf.size(), &bytes_written, io))
						{
							if (bytes_written == tmp_buf.size())
							{
								result = true;
							}
							else err_msg = "Wrote file, but expected bytes were " +
								numstr::ToStr(tmp_buf.size()) + " and actual bytes were " +
								numstr::ToStr(bytes_written);
						}
						else err_msg = "Failed to write file: " + filename;
						mpo_close(io);
					}
					else err_msg = "Could not create file: " + filename;
				} // end if integrity check passed
				// else integrity check failed
				else err_msg = ".torrent file was downloaded, but its integrity check failed";
			}
			// else the download failed, so we'll try the next mirror (if available)
			else err_msg = ".torrent file could not be downloaded";
			
			cur_index++;
			if (cur_index >= count) cur_index = 0;	// wraparound
		} while ((cur_index != start_index) && (!result));
	} // end if there is at least 1 torrent to try to download
	else err_msg = "Could not find any associated .torrent files to download";

	if (!result)
	{
		wxMessageBox(err_msg.c_str(),
				"Error", wxOK | wxICON_ERROR, this);
	}

	return result;
}

bool MyFrame::do_download(const struct ctrl_version *v)
{
	bool result = false;

	// try to download the .torrent file to disk and verify it.  If this succeeds, keep going ...
	if (download_torrent(v))
	{
		vector<string> cl;	// holds our command-line
		cl.clear();
		string arg = "";
		string exename = "";
#ifdef WIN32
		exename = "bittorrent/btdownloadgui.exe";
#else
		exename = "bittorrent/btdownloadgui.py";	// assume wrongly unix users have python installed here
#endif
		cl.push_back(exename);	// add to the vector
		arg = "--responsefile";
		cl.push_back(arg);
		arg = v->m_rootdir;	// torrent goes into root VLDP dir
		arg += "/" + v->m_dir + ".torrent";	// the name of the .torrent file
		cl.push_back(arg);
		MPO_MKDIR(v->m_rootdir.c_str());	// create directory if it does not exist (if it exists, that's fine too)
		arg = "--saveas";
		cl.push_back(arg);
		arg = v->m_rootdir;
		arg += "/";
		// commented out: windows version of BitTorrent (latest) will create
		// the directory even with --saveas,
		// whereas the linux version i have installed won't...
		// I need to find some consistency here!! :(
		//arg += v->m_dir;
		cl.push_back(arg);

		int exit_code = 0;

		wxBeginBusyCursor();

		// show progress dialog so user doesn't wonder if the loader is locked up
		wxProgressDialog *progress = new wxProgressDialog("Waiting",
			"Waiting for BitTorrent to finish", 100, this,
			wxPD_AUTO_HIDE | wxPD_APP_MODAL );

		// run bit torrent without returning until it is complete
		if (spawn_executable(cl, true, &exit_code))
		{
			// ADDME : we can check to see if the exit code is not 0, but
			// it appears that bit torrent may always return 0 (at least for win32)
			result = true;
		}
		else
		{
			wxString msg = exename.c_str();
			wxMessageBox(msg + " could not be executed!  You may need to tweak your preferences.",
				"ERROR!", wxOK | wxICON_ERROR, this);
		}

		progress->Update(100);
		progress->Destroy();	// get rid of it ...

		wxEndBusyCursor();

	} // else download failed
	else
	{
		wxMessageBox("Could not get .torrent file!\n"
			"This file is needed to automatically get files from the internet!",
			"Error", wxOK | wxICON_ERROR, this);
	}

	return result;
}

bool MyFrame::prompt_verify(const struct ctrl_version *v, const char *file_description)
{
	bool result = false;
	string title = file_description;
	title += " Verification Needed";
	if (wxMessageBox("DAPHNE will now verify its files.",
		title.c_str(), wxOK | wxCANCEL, this) == wxOK)
	{
		result = do_verify(v);
	}
	return result;
}

// NOTE : *.md5 files are created only if user does not abort the procedure
bool MyFrame::do_verify(const struct ctrl_version *v)
{
	bool result = false;
	const int MAX = 1024;	// our progress updater resolution (better if it's 2^n)
	bool abort = false;	// whether we need to abort early
	
	wxBeginBusyCursor();
	
	// check every file in our list
	for (size_t f_index = 0; (f_index < v->m_files.size()) && !abort; f_index++)
	{
		mpo_io *io;
		string path = make_autodownload_path(v);
		
		// we want to skip files that are already verified to save time
		if (get_status_one_file(path, &v->m_files[f_index]) != DLSTAT_FILE_VERIFIED)
		{
			path += "/" + v->m_files[f_index].m_name;
			
			// show a nice cool progress dialog ...
			wxProgressDialog *progress = new wxProgressDialog("Please Wait",
				("Verifying " + path).c_str(), MAX, this,
				wxPD_AUTO_HIDE | wxPD_APP_MODAL | wxPD_CAN_ABORT | wxPD_REMAINING_TIME );
			
			io = mpo_open(path.c_str(), MPO_OPEN_READONLY);

			// if we can open the file ...
			if (io)
			{
				unsigned char minibuf[16384];	// increase this for more efficiency
				MPO_BYTES_READ bytes = 0;
				MPO_UINT64 total_bytes = 0;
				oMD5_CTX md5stuff;
				unsigned char md5_digest[16];
				
				MD5Init(&md5stuff);
				
				// go through entire file until we hit EOF or user aborts
				while (!io->eof && !abort)
				{
					if (mpo_read(minibuf, sizeof(minibuf), &bytes, io))
					{
						MD5Update(&md5stuff, minibuf, bytes);
					}
					// something weird happened, abort
					else
					{
						do_error("Could not read the file to verify it!");
						abort = true;
					}
					
					total_bytes += bytes;
					
					// some files can be 0 bytes (bit torrent creates them as 0 bytes)
					// protect against divide by 0 error
					if (io->size > 0)
					{
						// update them on our progress, and break out of the loop if they cancel
						if (progress->Update((total_bytes * MAX) / io->size) == false)
						{
							result = false;	// because an earlier file may have passed this test
							abort = true;
						}
					}
					// else they don't need any progress anyway...
				} // end while
				
				// if we ended the loop because of EOF ...
				if (io->eof)
				{
					MD5Final(md5_digest, &md5stuff);	// finalize md5 hash
					
					mpo_io *md5_io = mpo_open((path + ".md5").c_str(), MPO_OPEN_CREATE);
					
					// try to create *.md5 file now that we've gone to the trouble to calculate it
					if (md5_io)
					{
						// try to write file to disk
						// if write fails, print error message
						if (!mpo_write(md5_digest, sizeof(md5_digest), NULL, md5_io))
						{
							do_error("Could not write verification file!\n"
								"Do you have write permissions?");
						}
						mpo_close(md5_io);
					}
					// else could not create file
					else
					{
						do_error("Could not create verification file!\n"
							"Without this file, DAPHNE will have to verify the files every time.");
						// the verification can still succeed without a file being created,
						// so we don't set the abort flag...
					}
					
					// NOW verify md5 hash with what control file says the md5 ought to be
					if (compare_md5s(md5_digest, v->m_files[f_index].m_md5))
					{
						result = true;	// woohoo, verification on this file succeeded!
					}
					// verification failed!
					else
					{
						string msg = "Verification failed for file: ";
						msg += path + "\nDAPHNE probably needs to get this file from the internet again.\n";
						msg += "This will happen automatically the next time you try to start this game.";
						do_error(msg.c_str());
						result = false;	// because previous file's verification may have succeeded
						abort = true;	// no need to verify the other files ...
					}
				}
				// else we ended due to user abortment, no error message needed
				
				mpo_close(io);	// close file
			}
			// else file could not be opened
			else
			{
				do_error("Could not open file to verify it!");
				abort = true;
			}
			
			progress->Update(MAX);
			progress->Destroy();
		} // end if file was not already verified
		// else file is verified, so no work needed
	} // end for loop
	
	wxEndBusyCursor();
	
	return result;
}

bool MyFrame::compare_md5s(unsigned char *challenge_digest, string good_md5)
{
	bool result = false;
	string challenge = mpom::bin2hex(challenge_digest, 16);
	if (mpom::str_toupper(challenge) == mpom::str_toupper(good_md5))
	{
		result = true;
	}
	return result;
}

void MyFrame::do_error(const char *msg)
{
	wxMessageBox(msg,
		"Error!", wxOK | wxICON_ERROR, this);
}

void MyFrame::AddCmdLine(string s)
{
	m_cmd_line.push_back(s);	// add to the queue ...
}

void MyFrame::OnConfig(wxCommandEvent & WXUNUSED(event))
{
	// now create the config dialog box ...
	MyPreConfig pre_config(this);
	pre_config.ShowModal();	// show the modal dialog box ...
}

void MyFrame::OnTransfers(wxCommandEvent & WXUNUSED(event))
{
	// if the frame isn't currently being displayed
	if (m_xfer_frame == NULL)
	{
		m_xfer_frame = new XferFrame( "Transfers", wxPoint(50,50), wxSize(420,285) );
		m_xfer_frame->Show(TRUE);
	}
}

// called by XferFrame's destructor so we know when it is being killed
void MyFrame::XferFrameDying()
{
	m_xfer_frame = NULL;	// wxWindows garbage collection should make this OK
}

void MyFrame::OnSelected(wxCommandEvent &event)
{
	// set the game index
	m_the_vars->game_index = m_listbox->GetSelection();

	// get the disc name
	m_the_vars->games[m_the_vars->game_index].disc_name = g_config_xml.m_games[m_the_vars->game_index].m_disc;

	// get the image name
	string val = g_config_xml.m_games[m_the_vars->game_index].m_image;

	// if an image exists for this game
	if (val != "")
	{
		m_canvas->new_image(val.c_str());
	}

	// otherwise, display a blank image
	else
	{
		m_canvas->clear_image();
	}

	m_canvas->Refresh();
}

bool MyFrame::download_file(const char *url, string msg, string &buf)
{
	bool result = false;
	mpo_client c;
	mpo_http h;

	h.init(&c, "DaphneLoader");

	wxBeginBusyCursor();

	// show a nice cool progress dialog ...
	wxProgressDialog *progress = new wxProgressDialog("Please Wait",
		msg.c_str(), 100, this,
		wxPD_AUTO_HIDE | wxPD_APP_MODAL | wxPD_CAN_ABORT );	// show progress

	net_result res = NET_ERROR;
	string host = "", uri = "";
	unsigned int port = 0;

	if (net_parse_url(url, host, port, uri))
	{
		res = c.connect_to_host(host.c_str(), port, 500);
	}

	int counter = 0;

	// can be user-aborted
	while (res == NET_TIMEOUT)
	{
		counter += 10;
		if (counter > 90) counter = 0;

		res = c.wait_connect(500);
		// update them on our progress, and break out of the loop if they cancel
		if (progress->Update(counter, (msg + "Connecting...").c_str()) == false)
		{
			c.disconnect();
			res = NET_DISCONNECT;
			break;
		}
	}

	// if we're connected ...
	if (res == NET_OK)
	{
		unsigned char minibuf[1024];
		unsigned int total_bytes = 0;
		unsigned int expected_size = 0;
		buf = "";

		res = h.StartGet(host, uri, true, 5000);

		// if we're ready to start receiving AND we know the size of the file we're downloading
		if (res == NET_OK)
		{
			bool aborted = false;

			// get HTTP headers
			while (!aborted)
			{
				res = h.GetHeaders(1000);
				if (res == NET_OK)
				{
					expected_size = h.GetContentLength();
					break;	// time to call receive
				}

				// if we get some error, bail out ...
				else if (res != NET_TIMEOUT) aborted = true;

				if (progress->Update(0, (msg + "Receiving Headers...").c_str()) == false)
				{
					aborted = true;
				}
			}

			while (!aborted && (expected_size > 0))
			{
				int bytes_read = 0;
				unsigned int bytes_to_read = expected_size - total_bytes;

				// don't overflow
				if (bytes_to_read > sizeof(minibuf)) bytes_to_read = sizeof(minibuf);

				if (bytes_to_read > 0)
				{
					res = h.Receive(minibuf, bytes_to_read, &bytes_read, 1000);

					// we shouldn't get EOF if we're keeping track of our size ...
					if (res == NET_OK)
					{
						total_bytes += bytes_read;
						buf.append((const char *) minibuf, bytes_read);
					}
					// if we get some error
					else if (res != NET_TIMEOUT)
					{
						aborted = true;
					}
				}
				// we've hit EOF, we're done!
				else
				{
					result = true;
					break;
				}

				// give them a progress update
				// (if this returns false, it means they canceled)
				if (progress->Update((total_bytes * 100) / expected_size, (msg + "Receiving...").c_str()) == false)
				{
					aborted = true;
				}
			} // end loop
		} // end if HTTP request came back favorable

		// else something weird happened ...

		c.disconnect();	// regardless of pass/fail, this is a good place to disconnect
	}
	// else it failed

	progress->Update(100);
	progress->Destroy();

	wxEndBusyCursor();

	return result;
}

string MyFrame::make_autodownload_path(const struct ctrl_version *v)
{
	string path = v->m_rootdir;
	path += "/" + v->m_dir;
	return path;
}

string MyFrame::make_md5_fullpath(const struct ctrl_version *v, int which_file)
{
	string path = make_autodownload_path(v);
	path += "/" + v->m_files[which_file].m_name;
	path += ".md5";
	return path;
}

dl_status MyFrame::get_status_one_file(const string autodownload_path, const struct ctrl_file *f)
{
	dl_status result = DLSTAT_ERROR;
	string path = autodownload_path + "/" + f->m_name;

	mpo_io *io = mpo_open(path.c_str(), MPO_OPEN_READONLY);

	// try to open the file (verifies existence and length)
	if (io)
	{
		// if the sizes are correct
		if (numstr::ToUint64(f->m_size.c_str()) == io->size)
		{
			mpo_io *md5_io = mpo_open((path+".md5").c_str(), MPO_OPEN_READONLY);

			// if the file's *.md5 checksum file exists, it means that a checksum was performed
			if (md5_io)
			{
				// make sure that the *.md5 file is as new as or newer than the file it has verified
				if (md5_io->time_last_modified >= io->time_last_modified)
				{
					unsigned char digest[16];
					// if we are able to read the *.md5 file ok
					if (mpo_read(digest, 16, NULL, md5_io))
					{
						// if MD5's match up
						if (compare_md5s(digest, f->m_md5))
						{
							result = DLSTAT_FILE_VERIFIED;	// woohoo!
						}
						// md5 match failed
						else
						{
							result = DLSTAT_CHECKSUM_WRONG;
						}
					}
					// for some reason, couldn't read the file
					else
					{
						result = DLSTAT_ERROR;
					}
				} // end if 'last_modified' check succeeded
				// else the .md5 file is out of date and cannot be used
				else
				{
					result = DLSTAT_FILE_UNVERIFIED;
				}
				mpo_close(md5_io);
			}
			// else the *.md5 file is missing
			else
			{
				result = DLSTAT_FILE_UNVERIFIED;
			}
		}
		// size mismatch
		else
		{
			result = DLSTAT_LENGTH_WRONG;
		}
		mpo_close(io);
	}
	// file cannot be opened, so we assume it does not exist
	else
	{
		result = DLSTAT_FILE_MISSING;
	}

	return result;
}

dl_status MyFrame::get_autodownload_status(const struct ctrl_version *v)
{
	dl_status result = DLSTAT_ERROR;
	size_t index = 0;
	bool got_error = false;
	bool one_file_unverified = false;	// true if at least 1 file is unverified

	// sanity check
	// if v exists and is not empty
	if (v && (v->m_files.size() > 0))
	{
		// verify all files
		for (index = 0; (index < v->m_files.size()) && !got_error; index++)
		{
			result = get_status_one_file(make_autodownload_path(v), &v->m_files[index]);
			switch (result)
			{
			case DLSTAT_FILE_UNVERIFIED:
				one_file_unverified = true;
				// we don't set got_error to true here because we'd rather return a more severe
				// error message than just "unverified" because verifying is expensive and
				// if we can find another problem without verifying, so much the better
				break;
			case DLSTAT_FILE_VERIFIED:
				break;
				// anything else, we want to abort ...
			default:
				got_error = true;
				break;
			}
		} // end for loop
	} // end sanity check

	// if the last file in the set verified properly, but a previous file did not
	if (DLSTAT_FILE_VERIFIED && one_file_unverified)
	{
		result = DLSTAT_FILE_UNVERIFIED;	// then report that at least one file did not verify
	}

	return result;
}

bool MyFrame::load_control_file()
{
	bool result = false;
	string tmp_buf = "";

	const char *CONTROL_NAME = "controlptr1.bin";

	string url = "http://www.daphne-emu.com/updates/";
	url += CONTROL_NAME;

	if (download_file(url.c_str(), "Checking for updates...", tmp_buf))
	{
		// convert to an INI that we can refer to
		g_controlptr_ini.use_buf((const unsigned char *) tmp_buf.data(), tmp_buf.size());

		if (g_controlptr_ini.find_key("ControlMirrors"))
		{
			string md5_hash = g_controlptr_ini.get_var_value("md5");
			string count_s = g_controlptr_ini.get_var_value("num");
			unsigned int count = atoi(count_s.c_str());	// string can be empty, in which case atoi will result in 0

			if (count > 0)
			{
				unsigned int start_index = (unsigned int) (wxGetLocalTime() % count);	// start at a random mirror
				unsigned int cur_index = start_index;

				do
				{
					string s = "mirror";
					s += numstr::ToStr(cur_index);
					s = g_controlptr_ini.get_var_value(s.c_str());	// get URL of mirror

					// if we found a URL and if the download was successful
					if ((s != "") && (download_file(s.c_str(), s + "Updating...", tmp_buf)))
					{
#ifdef DEBUG
						// when debugging, we may be altering the control file constantly
						if (true)
#else
						// if upper-case MD5 of what we downloaded is equal to
						// the upper-case MD5 of what the server says we should've downloaded
						if (mpom::str_toupper(mpom::compute_md5((const unsigned char *) tmp_buf.data(),
							tmp_buf.size())) == mpom::str_toupper(md5_hash))
#endif
						{
							string err_msg = "";
							// convert the file we've downloaded to our control file
							result = parse_control_file((const unsigned char *) tmp_buf.data(), tmp_buf.size(), err_msg);
							if (!result)
							{
								do_error(err_msg.c_str());
							}
							break;	// we're done ...
						}
						// notify them what is going on
						else
						{
							wxMessageBox("Daphne found an update, but the integrity test failed.",
								"Checksum Error", wxOK | wxICON_ERROR, this);
						}
					}
					// else we're in the loop anyway ...

					cur_index++;
					if (cur_index >= count) cur_index = 0;	// wraparound
				} while (cur_index != start_index);				
			}
		}
	}

	if (!result)
	{
		wxMessageBox("Daphne could not be updated!\n"
			"Update checking can be disabled in the preferences.",
			"Networking Error", wxOK | wxICON_ERROR, this);
	}

	return result;
}

// PRECONDITION : g_control_xml and g_config_xml MUST be loaded and populated
bool MyFrame::load_prefs_file()
{
	bool result = false;
	string tmp_buf;
	string err_msg = "";
	mpo_io *io = mpo_open(PREFS_PATH, MPO_OPEN_READONLY);

	if (io)
	{
		unsigned char *buf = new unsigned char [io->size];
		if (buf)
		{
			if (mpo_read(buf, io->size, NULL, io))
			{
				result = g_prefs.parse(buf, io->size, err_msg);
			}
			delete [] buf;
		}
		mpo_close(io);
	}

	// if we have an error to report to the user
	if (!result && (err_msg != ""))
	{
		do_error(err_msg.c_str());
	}

	return result;
}

void MyFrame::apply_the_vars()
{
	m_listbox->SetSelection(m_the_vars->game_index);	// select stored game index
	wxCommandEvent tmp;
	OnSelected(tmp);	// force the graphic to display
}

BEGIN_EVENT_TABLE(MyCanvas, wxScrolledWindow)
EVT_PAINT(MyCanvas::OnPaint)
END_EVENT_TABLE()

// Define a constructor for my canvas
MyCanvas::MyCanvas(wxWindow *parent, const wxPoint& pos, const wxSize& size):
wxScrolledWindow(parent, -1, pos, size, wxSUNKEN_BORDER),
m_bitmap(NULL)
{
	// wxSUNKEN_BORDER, wxSIMPLE_BORDER
	m_handler = new wxJPEGHandler;
	m_image.AddHandler(m_handler);
	m_bitmap = new wxBitmap();
}

MyCanvas::~MyCanvas()
{
	// de-allocate m_bitmap and m_handler somehow?
}

// tells the canvas to display a new image
void MyCanvas::new_image(const wxString &filename)
{
	m_bitmap->LoadFile("images/" + filename, wxBITMAP_TYPE_JPEG);
}

// erases bitmap
void MyCanvas::clear_image()
{
	m_bitmap->Create(1, 1);
}

// Define the repainting behaviour
void MyCanvas::OnPaint(wxPaintEvent& WXUNUSED(event))
{
	wxPaintDC dc(this);

	if ( m_bitmap && m_bitmap->Ok() )
	{
		wxMemoryDC memDC;

		memDC.SelectObject(* m_bitmap);

		// Normal, non-transparent blitting
		dc.Blit(0, 0, m_bitmap->GetWidth(), m_bitmap->GetHeight(), & memDC, 0, 0, wxCOPY, FALSE);

		memDC.SelectObject(wxNullBitmap);
	}
}

//////////////////////

MyPreConfig::MyPreConfig(wxWindow *parent)
: wxDialog(parent, -1, wxString(_T("Config")))
{
	m_config = new MyConfig(this, ID_Notebook);
	m_config->CreateInitialPages();

	m_ok = new wxButton(this, ID_CFG_OK, "&OK");
	m_cancel = new wxButton(this, ID_CFG_CANCEL, "Cancel");
	m_apply = new wxButton(this, ID_CFG_APPLY, "Apply");

	// sizers
	m_sizer_main = new wxBoxSizer(wxVERTICAL);
	m_sizer_bottom_row = new wxBoxSizer(wxHORIZONTAL);

	int flags = wxALL;
	m_sizer_bottom_row->Add(m_ok, 0, flags, 4);
	m_sizer_bottom_row->Add(m_cancel, 0, flags, 4);
	m_sizer_bottom_row->Add(m_apply, 0, flags, 4);

	m_sizer_config = new wxNotebookSizer(m_config);
	m_sizer_main->Add(m_sizer_config, 1, wxEXPAND);
	m_sizer_main->Add(m_sizer_bottom_row, 0, wxEXPAND);

	wxSize real_min_size(0, 0);	// this'll eventually be the true minimum size

	// there are basically 5 dynamic possibilities that we are concerned with
	for (int i = 0; i < 5; i++)
	{
		// display possibility 1
		m_config->display_dynamic_possibility(i);

		// compute minimum size
		m_sizer_main->Layout();

		// update our real min size as we found bigger minimum sizes
		wxSize tmp = m_sizer_main->GetMinSize();
		if (tmp.GetWidth() > real_min_size.GetWidth()) real_min_size.SetWidth(tmp.GetWidth());
		if (tmp.GetHeight() > real_min_size.GetHeight()) real_min_size.SetHeight(tmp.GetHeight());
	}

	// we know truly know the minimum size so we can safely set it
	m_sizer_main->SetMinSize(real_min_size);

	this->SetSizer(m_sizer_main);
	this->Fit();

	m_config->Initialize();	// get ready to go ...
}

void MyPreConfig::RedoLayout()
{
	/*
	// THIS CODE DOESN'T WORK
	//m_sizer_main->Layout();
	//wxSize tmp = m_sizer_main->GetMinSize();
	//this->SetSize(tmp);
	*/

	// this DOES work
	this->Fit();
}

void MyPreConfig::OnConfigKeys(wxCommandEvent& event)
{
}

void MyPreConfig::OnConfigButtons(wxCommandEvent& event)
{
}

void MyPreConfig::OnOK(wxCommandEvent& event)
{
	int which_tab = 0;

	m_config->apply_changes();
	if (m_config->validate(which_tab))
	{
		EndModal(0);
	}

	// print warning message
	else
	{
		m_config->ChangeTab(which_tab);	// go to the problem tab so they can clearly see what's wrong
		wxMessageBox("DAPHNE still requires some configuration.\n"
			"Items marked in red show what to do.",
			"More Configuration Needed!", wxOK | wxICON_ERROR, m_config);
	}
}

void MyPreConfig::OnCancel(wxCommandEvent& event)
{
	EndModal(0);
}

void MyPreConfig::OnApply(wxCommandEvent& event)
{
	m_config->apply_changes();
}

MyConfig::MyConfig(wxWindow *parent, wxWindowID id,
				   const wxPoint &pos, const wxSize &size, long style)
				   : wxNotebook(parent, id, pos, size, style)
{
	m_the_vars = g_mainframe->get_readonly_vars();
	m_game_vars = &m_the_vars->games[m_the_vars->game_index];

	m_disc = control_find_disc(&g_control_xml, m_game_vars->disc_name);	// this is as good a place as any to define this
	m_preconfig = (MyPreConfig *) parent;	// make the assignment
}

// just a test to make sure that all the memory we were allocating was getting de-allocated properly
MyConfig::~MyConfig()
{
}

void MyConfig::CreateInitialPages()
{
	wxPanel *panel = NULL;

	panel = CreateLaserdiscPage();
	AddPage(panel, "Laserdisc", FALSE);

	panel = CreateGamePage();
	AddPage(panel, "Game", FALSE);

	panel = CreateVideoPage();
	AddPage(panel, "Video", FALSE);

	panel = CreateAudioPage();
	AddPage(panel, "Audio", FALSE);

	panel = CreateInputPage();
	AddPage(panel, "Input", FALSE);

	panel = CreateAdvancedPage();
	AddPage(panel, "Advanced", FALSE);
}

// called after tabs have been created, and dialog box has been sized.. clears validators, sets start tab
void MyConfig::Initialize()
{
	// LASERDISC TAB START

	// clear validators
	m_val_player_type->SetLabel("");
	m_val_vldp_choices->SetLabel("");
	m_val_vldp_framefile->SetLabel("");

	// set defaults
	// HandlePlayerTypeSelection need not be called here because HandleTabChange calls it
	m_player_type->SetSelection(m_game_vars->ldp_type);	// ldp_type corresponds to the ordering of m_player_Type

	// value of framefile string (if used)
	m_vldp_framefile->SetValue(m_game_vars->vldp_framefile.c_str());

	if (m_game_vars->vldp_autodownload) m_vldp_choices->SetSelection(1);	// auto download
	else m_vldp_choices->SetSelection(2);	// manual install

	// m_vldp_version is not allocated if m_disc is NULL
	if (m_disc)
	{
		// handle VLDP auto-download version
		m_vldp_version->SetSelection(m_game_vars->vldp_download_index);	// the version they last chose
	}

	// LASERDISC TAB END

	// GAME TAB START
	// get game options drawn
	m_gamever_choices->SetSelection(m_game_vars->game_version_index);
	HandleGameVerSelection(m_game_vars->game_version_index);
	// GAME TAB END

	ChangeTab(m_the_vars->tab_index);	// go to the tab we were last at

	// DEBUG: disabled auto validation for testing purposes, it can be re-enabled later
	/*
	// if there are problems in the configuration, go there instead so user can clearly see what's going on
	int problem_tab = -1;
	if (!validate(problem_tab))
	{
		ChangeTab(problem_tab);	// go to the rogue tab
	}
	*/
}

// if the laserdisc player type is selected ...
void MyConfig::OnPlayerType(wxCommandEvent &event)
{
	if (event.m_eventType == wxEVT_COMMAND_CHOICE_SELECTED)
	{
		HandlePlayerTypeSelection(event.m_commandInt);
		m_val_player_type->SetLabel("");	// clear validator
	}
}

// if they vldp autodownload/manual wxChoice is changed
void MyConfig::OnVLDPChoices(wxCommandEvent &event)
{
	if (event.m_eventType == wxEVT_COMMAND_CHOICE_SELECTED)
	{
		HandleVldpChoice(event.m_commandInt);
		m_val_vldp_choices->SetLabel("");	// clear validator
	}
}

// this isn't part of OnPlayerType because this function is also called when we first initialize the Laserdisc Page
void MyConfig::HandlePlayerTypeSelection(int player_index)
{
	switch (player_index)
	{
	default:	// choose or nothing
		m_ldp_driver = "??";	// undefined
		m_ldp_type = LDP_TYPE_UNKNOWN;	// undefined
		break;
	case 1:	// VLDP
		m_ldp_driver = "vldp";
		m_ldp_type = LDP_TYPE_VLDP;
		break;
	case 2:	// LDP
		m_ldp_driver = "???";	// not specified yet
		m_ldp_type = LDP_TYPE_SERIAL;
		break;
	case 3:	// NOLDP
		m_ldp_driver = "noldp";
		m_ldp_type = LDP_TYPE_NONE;
		break;
	}

	ResetLaserdiscPageLayout();
}

// this is in its own function because display_dynamic_possibilities and HandlePlayerTypeSelection accept
// different arguments.
void MyConfig::ResetLaserdiscPageLayout()
{

	// hide none message
	m_ldp_main_sizer->Show(m_none_sizer, false);

	m_vldp_box->Show(false);	// the box must be hidden separately because it is somewhat independent
	m_ldp_main_sizer->Show(m_vldp_sizer, false);

	// hide physical LDP box and sizer
	m_ldp_box->Show(false);
	m_ldp_main_sizer->Show(m_ldp_sizer, false);

	// now make the selected one visible
	if (m_ldp_type == LDP_TYPE_NONE)
	{
		m_ldp_main_sizer->Show(m_none_sizer, true);
	}

	else if (m_ldp_type == LDP_TYPE_VLDP)
	{
		m_vldp_box->Show(true);
		m_ldp_main_sizer->Show(m_vldp_sizer, true);
		HandleVldpChoice(m_vldp_choices->GetSelection());	// the VLDP choices may need to be updated when the VLDP options are displayed
	}

	else if (m_ldp_type == LDP_TYPE_SERIAL)
	{
		m_ldp_box->Show(true);
		m_ldp_main_sizer->Show(m_ldp_sizer, true);
	}
	// else we haven't chosen a type yet so show nothing

	m_centering_sizer->Layout();	// this must be called any time visibility is altered

}

// this isn't part of OnVldpChoices because this function may also be called when we first initialize the Laserdisc Page
void MyConfig::HandleVldpChoice(int index)
{
	// hide autodownload options
	m_vldp_sizer->Show(m_vldp_auto_sizer, false);

	// hide manual options
	m_vldp_sizer->Show(m_vldp_ff_sizer, false);

	switch (index)
	{
	case 0:	// choose
		break;
	case 1:	// auto-download
		m_vldp_sizer->Show(m_vldp_auto_sizer, true);
		break;
	case 2:	// manual
		// show manual options
		m_vldp_sizer->Show(m_vldp_ff_sizer, true);
		break;
	default:	// unknown
		break;
	}

	m_centering_sizer->Layout();	// this must be called any time visibility is altered
	// m_vldp_sizer->Layout cannot be called because it does not resize m_vldp_box for some unknown reason!
}

const char *G_VAL_PLAYER_TYPE_MSG = "You must choose whether to simulate a laserdisc player!";
const char *G_VAL_VLDP_CHOICES_MSG = "You must choose whether to get missing files automatically!";
const char *G_VAL_VLDP_FRAMEFILE_MSG = "You must specify a framefile when installing the video/audio files manually.";

// bring up a file browsing window so user can select their framefile
void MyConfig::OnBrowseFramefile(wxCommandEvent &event)
{
	wxFileDialog file_dialog(this, "Find your framefile",
		".",	// default directory
		"",	// default filename
		"*.txt",
		wxOPEN | wxHIDE_READONLY);

	int result = file_dialog.ShowModal();

	// if they pressed ok
	if (result == wxID_OK)
	{
		m_vldp_framefile->SetValue(file_dialog.GetPath());
		m_val_vldp_framefile->SetLabel("");	// clear validator
	}
}

void MyConfig::OnTabChange(wxNotebookEvent &event)
{
	HandleTabChange(event.GetSelection());	
	event.Skip();	// this seems to be required
}

void MyConfig::ChangeTab(int which_tab)
{
	SetSelection(which_tab);
	HandleTabChange(which_tab);
}

void MyConfig::HandleTabChange(int new_tab_index)
{
	// if they click on the laserdisc tab, we need to make sure that it's up to date
	if (new_tab_index == 0)
	{
		HandlePlayerTypeSelection(m_player_type->GetSelection());
	}
	// else if they click on the video tab, we need to make sure that it's up to date
	else if (new_tab_index == 2)
	{
		// if they are using VLDP, they need certain options enabled
		if (m_ldp_type == LDP_TYPE_VLDP)
		{
			m_blend->Enable();
			m_scanlines->Enable();
			m_vldp_blank_on_searches->Enable();
			m_vldp_blank_on_skips->Enable();
			m_disable_hwaccel->Enable();
		}
		// otherwise they need those options disabled because they do not apply
		else
		{
			m_blend->Disable();
			m_scanlines->Disable();
			m_vldp_blank_on_searches->Disable();
			m_vldp_blank_on_skips->Disable();
			m_disable_hwaccel->Disable();
		}
	}
}

void MyConfig::apply_gamecfg_changes()
{
	// holds the config section that this game version uses
	const cfg_gamecfg *cfg = g_mainframe->get_active_gamever_cfg(m_gamever_choices->GetSelection());

	// if this game has an assigned config
	if (cfg != NULL)
	{
		// these variables tell us what to initialize the listboxes to
		// NOTE : we are changing read-only to read-write for this purpose
		conf_vars *cv = (conf_vars *) g_mainframe->get_confvars(cfg->m_name.c_str());

		// go through all listboxes and save their values
		for (size_t i = 0; i < m_game_listboxes.size(); i++)
		{
			size_t val = m_game_listboxes[i].m_choice->GetSelection();
			cv->listbox_indices[i] = val;
		}
	}
}

// stores all widget variables into our variable struct because this class will be destroyed before the game is run
void MyConfig::apply_changes()
{
	struct the_vars *v = g_mainframe->get_readwrite_vars();
	struct game_vars *g = &v->games[v->game_index];	// needs to be read/write

	g->adv_nocrc = m_disable_crc_check->GetValue();
	g->adv_display_cmd_line = m_display_cmd_line->GetValue();
	g->adv_noserversend = m_dont_report_stats->GetValue();

	g->aud_nosound = m_disable_sound->GetValue();
	//g->aud_buffer = 0;

	g->game_version_index = m_gamever_choices->GetSelection();

	g->ldp_driver = m_ldp_driver;
	g->ldp_type = m_ldp_type;

	g->vid_blend = m_blend->GetValue();
	g->vid_nohwaccel = m_disable_hwaccel->GetValue();
	g->vid_fullscreen = m_fullscreen->GetValue();
	g->vid_scanlines = m_scanlines->GetValue();
	g->vid_blank_searches = m_vldp_blank_on_searches->GetValue();
	g->vid_blank_skips = m_vldp_blank_on_skips->GetValue();
	//g->vid_x = 0;
	//g->vid_y = 0;

	// if they want to automatically download from the internet...
	if (m_vldp_choices->GetSelection() == 1) g->vldp_autodownload = true;
	else g->vldp_autodownload = false;
	g->vldp_framefile = m_vldp_framefile->GetValue();

	// m_vldp_Version is not allocated if m_disc is NULL
	if (m_disc)
	{
		g->vldp_download_index = m_vldp_version->GetSelection();
	}

	// remember which tab we were last on
	v->tab_index = GetSelection();

	apply_gamecfg_changes();
}

// which_tab = which tab has problems that have caused us not to validate properly
bool MyConfig::validate(int &which_tab)
{
	bool is_valid = false;

	// check for problems on the first tab
	which_tab = 0;

	// check LDP type selection
	switch (m_player_type->GetSelection())
	{
	default:	// unknown / nothing has been chosen
		m_val_player_type->SetLabel(G_VAL_PLAYER_TYPE_MSG);
		break;
	case 1:	// VLDP
		// check VLDP choice
		switch (m_vldp_choices->GetSelection())
		{
		default:	// unknown / nothing has been chosen
			m_val_vldp_choices->SetLabel(G_VAL_VLDP_CHOICES_MSG);
			break;
		case 1:	// automatically download
			// if there is no download available, then auto-downloading is not valid
			if (!control_download_available(m_disc))
			{
				// m_vldp_motd will only be defined if there is a MOTD
				if (m_disc && (m_disc->m_motd != ""))
				{
					m_vldp_motd->SetForegroundColour( *wxRED ); // red so user knows what the problem is
				}
				m_val_vldp_choices->SetLabel("You must choose 'No' here because the files cannot be found on the internet");
				wxSize s = m_val_vldp_choices->GetBestSize();
				m_vldp_sizer->SetItemMinSize(m_val_vldp_choices, s.x, s.y);	// this must be done when an item is resized
				m_preconfig->RedoLayout();	// anytime an item is resized, this must be done
			}
			else is_valid = true;
			break;
		case 2:	// manual download
			// if a framefile path has been defined
			// FIXME : this could be improved to ignore whitespace
			if (m_vldp_framefile->GetValue() != "")
			{
				is_valid = true;
			}
			else
			{
				m_val_vldp_framefile->SetLabel(G_VAL_VLDP_FRAMEFILE_MSG);
			}
			break;
		}
		break;
		// serial and noldp need no validation
	case 2:	// serial
	case 3:	// NOLDP
		is_valid = true;
	}

	return is_valid;
}

// the purpose of this function is to display all of the possible options that our laserdisc tab can display,
// so that the dialog box can be sized to the proper size
void MyConfig::display_dynamic_possibility(int which_one)
{
	switch (which_one)
	{
	case 0:
		m_ldp_type = LDP_TYPE_NONE;
		break;
	case 1:
		m_ldp_type = LDP_TYPE_VLDP;
		m_vldp_choices->SetSelection(2);	// auto-download
		break;
	case 2:
		m_ldp_type = LDP_TYPE_VLDP;
		m_vldp_choices->SetSelection(1);	// manual transfer
		break;
	case 3:
		m_ldp_type = LDP_TYPE_VLDP;
		m_vldp_choices->SetSelection(0);	// choose
	case 4:
		m_ldp_type = LDP_TYPE_SERIAL;
		break;
	default:
		assert(0);	// this should never happen
		break;
	}

	ResetLaserdiscPageLayout();
}

// if the game version is changed
void MyConfig::OnGameVerType(wxCommandEvent &event)
{
	if (event.m_eventType == wxEVT_COMMAND_CHOICE_SELECTED)
	{
		HandleGameVerSelection(event.m_commandInt);
	}
}

void MyConfig::HandleGameVerSelection(int new_ver_index)
{
	size_t i = 0;

	// STEP 1 : destroy any old game options
	// remove everything from our sizer
	for (i = 0; i < m_game_listboxes.size(); i++)
	{
		// remove everything until there is nothing left to be removed
		if (m_game_gsizer->Remove(0) == false) break;
	}

	// destroy old games
	for (i = 0; i < m_game_listboxes.size(); i++)
	{
		// mark for deletion
		m_game_listboxes[i].m_text->Destroy();
		m_game_listboxes[i].m_choice->Destroy();
	}
	m_game_listboxes.clear();	// empty this

	// holds the config section that this game version uses
	const cfg_gamecfg *cfg = g_mainframe->get_active_gamever_cfg(new_ver_index);

	// if this game has an assigned config
	if (cfg != NULL)
	{
		// these variables tell us what to initialize the listboxes to
		const conf_vars *cv = g_mainframe->get_confvars(cfg->m_name.c_str());

		// STEP 2 : create new game options
		// add all config listboxes
		for (i = 0; i < cfg->m_listboxes.size(); i++)
		{
			const struct cfg_listbox *lb = &cfg->m_listboxes[i];	// for convenience
			size_t li_index = 0;	// index for each list item in the list box
			m_listbox_holder lbh;	// this will be populated

			lbh.m_text = new wxStaticText(m_game_panel, -1, lb->m_title.c_str());
			lbh.m_choice = new wxChoice(m_game_panel, -1);
			lbh.m_choice->Append("[Auto]");	// first option is always auto

			// if a tooltip has been provided, then add it
			if (lb->m_tooltip != "") lbh.m_choice->SetToolTip(lb->m_tooltip.c_str());

			// now add each individual list item
			for (li_index = 0; li_index < cfg->m_listboxes[i].m_listitems.size(); li_index++)
			{
				lbh.m_choice->Append(lb->m_listitems[li_index].m_text.c_str());
			}
			fix_size(lbh.m_choice, WIDTH_FUDGE);

			// restore it to its previously selected value
			size_t val = cv->listbox_indices[i];
			lbh.m_choice->SetSelection(val);

			// add to our flex grid sizer
			m_game_gsizer->Add(lbh.m_text, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL, 0);
			m_game_gsizer->Add(lbh.m_choice, 0, wxEXPAND, 0);

			// add to our listbox vector
			m_game_listboxes.push_back(lbh);
		}
	} // end if game has a config
	// else game has no config, so there are no options to be displayed

	m_game_csizer->Layout();
	m_preconfig->RedoLayout();
}

wxPanel *MyConfig::CreateGamePage()
{
	wxPanel *panel = new wxPanel(this);

	wxStaticText *text = NULL;
	
	text = new wxStaticText(panel, -1, _T(
		"Game version:"));
	m_gamever_choices = new wxChoice(panel, ID_GAMEVER_CHOICES);
	m_gamever_choices->Append("[Auto]");	// default

	const cfg_game *g = &g_config_xml.m_games[m_the_vars->game_index];	// for convenience

	// add each version of the game to this dropdown list
	for (size_t gamever_idx = 0; gamever_idx < g->m_versions.size();
		gamever_idx++)
	{
		m_gamever_choices->Append(g->m_versions[gamever_idx].m_vername.c_str());
	}
	fix_size(m_gamever_choices, WIDTH_FUDGE);
	m_gamever_choices->SetSelection(0);	// FIXME : restore default value here

	// START SIZERS
	m_game_gsizer = new wxFlexGridSizer(0, 2, 5, 5);	// rows, cols, vgap, hgap

	wxStaticBox *sbox = new wxStaticBox(panel, -1, "Game Options");
	m_game_sbsizer = new wxStaticBoxSizer(sbox, wxVERTICAL);
	m_game_sbsizer->Add(m_game_gsizer, 0, wxALL, 10);

	m_game_vsizer = new wxBoxSizer(wxVERTICAL);
	m_game_vsizer->Add(text, 0, wxTOP | wxALIGN_CENTER_HORIZONTAL, 5);
	m_game_vsizer->Add(m_gamever_choices, 0, wxTOP | wxALIGN_CENTER_HORIZONTAL, 1);
	m_game_vsizer->Add(m_game_sbsizer, 0, wxTOP, 5);	// add the flex grid to this

	m_game_csizer = new wxBoxSizer(wxHORIZONTAL);
	m_game_csizer->Add(0, 0, 1);
	m_game_csizer->Add(m_game_vsizer, 0, wxALIGN_TOP, 0);
	m_game_csizer->Add(0, 0, 1);
	panel->SetSizer(m_game_csizer);
	// END SIZERS

	m_game_panel = panel;	// this is needed elsewhere
	return panel;
}

void MyConfig::resize(wxWindow *w, wxSizer *sz)
{
	wxSize s = w->GetSize();	// GetBestSize kind of croaks here on wxChoice
	sz->SetItemMinSize(w, s.x, s.y);
}

void MyConfig::fix_size(wxWindow *w, unsigned int padding)
{
	wxSize size = w->GetBestSize();
	size.SetWidth(size.GetWidth() + padding);	// add a bit of padding to the end
	w->SetSize(size);
}

// makes a validator (bold, red text)
wxStaticText *MyConfig::make_val(wxPanel *panel, const char *msg)
{
	wxStaticText *result = new wxStaticText(panel, -1, _T(msg));
	wxFont font = result->GetFont();
	font.SetWeight(wxFONTWEIGHT_BOLD);	// make font bold
	result->SetFont(font);	// make bold
	result->SetForegroundColour( *wxRED ); // make red
	return result;
}

wxPanel *MyConfig::CreateLaserdiscPage()
{
	wxPanel *panel = new wxPanel(this);

	m_player_type = new wxChoice(panel, ID_PLAYER_TYPE);
	m_player_type->SetToolTip("People who do not have a real laserdisc player will want to choose Virtual LDP.");

	wxStaticText *player_type_text = new wxStaticText(panel, -1, _T(
		"All games require a laserdisc player.\nMost people don't have a laserdisc player, and will need DAPHNE to simulate one.\n"
		"Should DAPHNE simulate a laserdisc player?"));

	m_player_type->Append("");	// empty slot
	m_player_type->Append("Yes, use Virtual Laserdisc Player. (recommended)");
	// I say 'definitely' here to push people who are uncertain toward VLDP
	m_player_type->Append("No, I definitely have a physical laserdisc player attached to my computer.");
	// I say 'dont want to play a game' here to push people who are uncertain toward VLDP
	m_player_type->Append("No, I don't want to play a game. I want to disable video/audio for testing.");

	fix_size(m_player_type, WIDTH_FUDGE);

	m_val_player_type = make_val(panel, G_VAL_PLAYER_TYPE_MSG);	// needs to be shown to get sized

	wxBoxSizer *expand_ldp_list = new wxBoxSizer(wxHORIZONTAL);
	expand_ldp_list->Add(m_player_type, 0, 0, 0);

	// START NONE
	m_none_text = new wxStaticText(panel, -1, _T("WARNING : This is probably not what you want,\n"
		"it's for testing only!\n"
		"If you don't have a laserdisc player, choose Virtual Laserdisc Player."));
	m_none_sizer = new wxBoxSizer(wxVERTICAL);
	m_none_sizer->Add(m_none_text, 0, 0, 0);
	// END NONE

	// START VLDP
	m_vldp_box = new wxStaticBox(panel, -1, "Virtual Laserdisc Player Options");
	m_vldp_sizer = new wxStaticBoxSizer(m_vldp_box, wxVERTICAL);
	m_vldp_auto_sizer = new wxBoxSizer(wxVERTICAL);	// automatic download options
	m_vldp_ff_sizer = new wxBoxSizer(wxVERTICAL);	// manual install options

	m_vldp_choose_text = new wxStaticText(panel, -1, _T(
		"DAPHNE's Virtual Laserdisc Player requires video/audio files from the original game's laserdisc.\n"
		"If these files are missing, should DAPHNE try to get them automatically from the internet?"
		));

	m_vldp_choices = new wxChoice(panel, ID_VLDP_CHOICES);
	m_vldp_choices->Append("");	// empty slot
	m_vldp_choices->Append("Yes, do it automatically (recommended)");
	m_vldp_choices->Append("No, I will install any missing video/audio files manually");
	fix_size(m_vldp_choices, WIDTH_FUDGE);
	m_val_vldp_choices = make_val(panel, G_VAL_VLDP_CHOICES_MSG);	// need to be shown to get sized

	if (m_disc)
	{
		// MOTD here
		// if there is a motd ...
		if (m_disc->m_motd != "")
		{
			m_vldp_motd_header = new wxStaticText(panel, -1, _T(
				"Message Of The Day:"));
			m_vldp_motd = new wxStaticText(panel, -1, _T(m_disc->m_motd.c_str()));
			wxFont motd_font = m_vldp_motd->GetFont();
			motd_font.SetWeight(wxFONTWEIGHT_BOLD);	// make font bold
			m_vldp_motd->SetFont(motd_font);	// make bold
			m_vldp_motd->SetForegroundColour( *wxBLUE ); // color it blue to indicate that all is well (for now)
		}

		// VERSION choice here
		m_vldp_version_header = new wxStaticText(panel, -1, _T(
			"Version to get"));
		m_vldp_version = new wxChoice(panel, ID_VLDP_VERSION);
		m_vldp_version->Append("Automatically choose the best one for me");	// default
		for (size_t ver_i = 0; ver_i < m_disc->m_versions.size(); ver_i++)
		{
			m_vldp_version->Append(m_disc->m_versions[ver_i].m_name.c_str());
		}
		fix_size(m_vldp_version, WIDTH_FUDGE);
	}

	// if we have an entry available for automatic downloading
	if (m_disc)
	{
		// add MOTD only if there is one
		if (m_disc->m_motd != "")
		{
			m_vldp_auto_sizer->Add(m_vldp_motd_header, 0, 0, 0);	// motd header
			m_vldp_auto_sizer->Add(m_vldp_motd, 0, wxTOP, 5);	// motd itself
		}
		// add VERSION only if there is one
		m_vldp_auto_sizer->Add(m_vldp_version_header, 0, wxTOP, 10);
		m_vldp_auto_sizer->Add(m_vldp_version, 0, wxTOP, 0);
	}
	// FIXME : maybe add something blank so the m_vldp_auto_sizer isn't empty?

	m_vldp_framefile_text = new wxStaticText(panel, -1, _T("MPEG Framefile Location"));
	m_vldp_framefile = new wxTextCtrl(panel, -1, "", wxPoint(-1, -1), wxSize(350, -1));
	m_vldp_framefile->SetToolTip("The DAPHNE online documentation explains what a framefile is.");
	m_vldp_browse = new wxButton(panel, ID_BROWSE_FRAMEFILE, "Browse");
	m_val_vldp_framefile = make_val(panel, G_VAL_VLDP_FRAMEFILE_MSG);	// need to show to be sized

	m_vldp_ff_sizer->Add(m_vldp_framefile_text, 0, wxTOP, 5);
	m_vldp_ff_sizer->Add(m_vldp_framefile, 0, 0, 0);
	m_vldp_ff_sizer->Add(m_vldp_browse, 0, wxTOP, 5);
	m_vldp_ff_sizer->Add(m_val_vldp_framefile, 0, wxTOP, 5);

	m_vldp_sizer->Add(m_vldp_choose_text, 0, wxTOP, 5);
	m_vldp_sizer->Add(m_vldp_choices, 0, wxTOP, 5);
	m_vldp_sizer->Add(m_val_vldp_choices, 0, wxTOP, 5);	// add validator
	m_vldp_sizer->Add(m_vldp_auto_sizer, 0, wxTOP, 5);
	m_vldp_sizer->Add(m_vldp_ff_sizer, 0, wxTOP, 5);
	m_vldp_sizer->Add(0, 15, 0);	// spacer
	// END VLDP

	// START LDP
	m_ldp_box = new wxStaticBox(panel, -1, "LDP Options");
	m_ldp_sizer = new wxStaticBoxSizer(m_ldp_box, wxVERTICAL);

	wxStaticText *text = new wxStaticText(panel, -1, "Serial Port");
	wxString choices[] = 
	{
		"Port 1 (COM1)",
		"Port 2 (COM2)"
	};
	int choices_count = sizeof(choices) / sizeof(wxString);
	m_serial_port = new wxChoice(panel, -1, wxDefaultPosition, wxDefaultSize, choices_count, choices);
	m_serial_port->Disable();

	wxFlexGridSizer *flex_sizer = new wxFlexGridSizer(2, 2, 15, 5);
	flex_sizer->Add(text, 0, 0, 0);
	flex_sizer->Add(m_serial_port, 0, 0, 0);

	text = new wxStaticText(panel, -1, "Baud Rate");
	wxString choices2[] =
	{
		"9600",
		"4800"
	};
	choices_count = sizeof(choices2) / sizeof(wxString);
	m_baud_rate = new wxChoice(panel, -1, wxDefaultPosition, wxDefaultSize, choices_count, choices2);
	m_baud_rate->Disable();

	flex_sizer->Add(text, 0, 0, 0);
	flex_sizer->Add(m_baud_rate, 0, 0, 0);
	m_ldp_sizer->Add(flex_sizer, 0, wxTOP, 15);

	// END LDP

	m_ldp_main_sizer = new wxBoxSizer(wxVERTICAL);
	m_ldp_main_sizer->Add(0, 10, 1);
	m_ldp_main_sizer->Add(player_type_text, 0, wxALIGN_CENTER_HORIZONTAL, 0);
	m_ldp_main_sizer->Add(expand_ldp_list, 0, wxALIGN_CENTER_HORIZONTAL, 0);
	m_ldp_main_sizer->Add(m_val_player_type, 0, wxALIGN_CENTER_HORIZONTAL, 0);
	m_ldp_main_sizer->Add(m_none_sizer, 0, wxTOP | wxALIGN_CENTER_HORIZONTAL, 10);
	m_ldp_main_sizer->Add(m_vldp_sizer, 0, wxTOP | wxALIGN_CENTER_HORIZONTAL, 10);
	m_ldp_main_sizer->Add(m_ldp_sizer, 0, wxTOP | wxALIGN_CENTER_HORIZONTAL, 10);

	m_centering_sizer = new wxBoxSizer(wxHORIZONTAL);
	m_centering_sizer->Add(0, 0, 1);
	m_centering_sizer->Add(m_ldp_main_sizer, 0, 0, 0);
	m_centering_sizer->Add(0, 0, 1);

	panel->SetSizer(m_centering_sizer);

	return panel;
}

wxPanel *MyConfig::CreateVideoPage()
{
	wxPanel *panel = new wxPanel(this);

	m_fullscreen = new wxCheckBox(panel, -1, "Run Fullscreen");
	m_fullscreen->SetToolTip(_T("If you want to run fullscreen, instead of in a window, check this box."));
	//	m_fullscreen->Disable();

	wxStaticText *screen_size_text = new wxStaticText(panel, -1, "Screen Size");
	wxString choices[] = 
	{
		"(default)",
			"640x480",
			"720x480",
			"800x600",
			"1024x768",
			"1152x864",
			"1280x1024",
			"1600x1200"
	};

	int choices_count = sizeof(choices) / sizeof(wxString);
	m_screen_size = new wxChoice(panel, -1, wxDefaultPosition, wxDefaultSize, choices_count, choices);
	m_screen_size->Disable();

	m_vldp_blank_on_skips = new wxCheckBox(panel, -1, "Blank on skips");
	m_vldp_blank_on_skips->SetToolTip("This blanks the screen on every skip.\nThis is not authentic behavior, more of a gimmick.");
	m_vldp_blank_on_searches = new wxCheckBox(panel, -1, "Blank on searches");
	m_vldp_blank_on_searches->SetToolTip("This blanks the screen on every search.\nThis is authentic behavior.");
	m_disable_hwaccel = new wxCheckBox(panel, -1, "Safe Mode (disable hardware acceleration)");
	m_disable_hwaccel->SetToolTip(_T("Disabling MPEG acceleration is safer but slower."));

	m_blend = new wxCheckBox(panel, -1, "Blend Fields (de-interlace)");
	m_blend->SetToolTip(_T("This will help smooth out any 'interlacing lines' in your video, at the cost of some detail."));

	m_scanlines = new wxCheckBox(panel, -1, "Simulate scanlines");
	m_scanlines->SetToolTip(_T("This may make your video look more authentic."));

	wxBoxSizer *screen_size_sizer = new wxBoxSizer(wxHORIZONTAL);
	screen_size_sizer->Add(screen_size_text, 0, wxRIGHT, 5);
	screen_size_sizer->Add(m_screen_size, 0, 0, 0);

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(screen_size_sizer, 0, wxTOP, 0);
	sizer->Add(m_fullscreen, 0, wxTOP, 10);
	sizer->Add(m_disable_hwaccel, 0, wxTOP, 10);
	sizer->Add(m_vldp_blank_on_searches, 0, wxTOP, 10);
	sizer->Add(m_vldp_blank_on_skips, 0, wxTOP, 10);
	sizer->Add(m_blend, 0, wxTOP, 10);
	sizer->Add(m_scanlines, 0, wxTOP, 10);

	wxBoxSizer *center = new wxBoxSizer(wxHORIZONTAL);
	center->Add(0, 0, 1);
	center->Add(sizer, 0, wxALIGN_CENTER_VERTICAL, 0);
	center->Add(0, 0, 1);

	panel->SetSizer(center);

	// restore vars here
	m_fullscreen->SetValue(m_game_vars->vid_fullscreen);
	m_vldp_blank_on_searches->SetValue(m_game_vars->vid_blank_searches);
	m_vldp_blank_on_skips->SetValue(m_game_vars->vid_blank_skips);
	m_disable_hwaccel->SetValue(m_game_vars->vid_nohwaccel);
	m_blend->SetValue(m_game_vars->vid_blend);
	m_scanlines->SetValue(m_game_vars->vid_scanlines);

	return panel;
}

wxPanel *MyConfig::CreateAudioPage()
{
	wxPanel *panel = new wxPanel(this);

	m_disable_sound = new wxCheckBox(panel, -1, "Disable Sound");
	m_disable_sound->SetToolTip(_T("If you don't want sound, or have no sound card, check this."));
	//	m_disable_sound->Disable();
	m_sound_buffer_modify = new wxCheckBox(panel, -1, "Sound Buffer Size");
	m_sound_buffer_modify->Disable();

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(m_disable_sound, 0, wxTOP, 0);
	sizer->Add(m_sound_buffer_modify, 0, wxTOP, 10);

	wxBoxSizer *center = new wxBoxSizer(wxHORIZONTAL);
	center->Add(0, 0, 1);
	center->Add(sizer, 0, wxALIGN_CENTER_VERTICAL, 0);
	center->Add(0, 0, 1);
	panel->SetSizer(center);

	// restore vars here
	m_disable_sound->SetValue(m_game_vars->aud_nosound);

	return panel;
}

wxPanel *MyConfig::CreateInputPage()
{
	wxPanel *panel = new wxPanel(this);

	return panel;
}

wxPanel *MyConfig::CreateAdvancedPage()
{
	wxPanel *panel = new wxPanel(this);

	m_ld_latency_modify = new wxCheckBox(panel, -1, "LD latency adjust");
	m_ld_latency_modify->Disable();
	m_dont_report_stats = new wxCheckBox(panel, -1, "Don't Report Stats");
	m_dont_report_stats->SetToolTip(_T("By default, DAPHNE will send anonymous stats to a central server. "
		"If you don't have an internet connection, or if you don't wish to participate, "
		"check this box."));
	m_disable_crc_check = new wxCheckBox(panel, -1, "Disable ROM CRC check");
	m_disable_crc_check->SetToolTip(_T("By default, DAPHNE performs CRC32 checking on all ROM images. "
		"The only time you don't want this is if you are experimenting with ROMs that DAPHNE doesn't expect. "
		"If in doubt, leave this unchecked."));
	m_display_cmd_line = new wxCheckBox(panel, -1, "Display command line before launching");

	wxBoxSizer *main_sizer = new wxBoxSizer(wxVERTICAL);
	int space = 10;
	main_sizer->Add(m_ld_latency_modify, 0, 0, 0);
	main_sizer->Add(m_dont_report_stats, 0, wxTOP, space);
	main_sizer->Add(m_disable_crc_check, 0, wxTOP, space);
	main_sizer->Add(m_display_cmd_line, 0, wxTOP, space);

	wxBoxSizer *centering_sizer = new wxBoxSizer(wxHORIZONTAL);
	centering_sizer->Add(0, 0, 1);
	centering_sizer->Add(main_sizer, 0, wxALIGN_CENTER_VERTICAL, 0);
	centering_sizer->Add(0, 0, 1);

	panel->SetSizer(centering_sizer);

	// restore vars here
	m_disable_crc_check->SetValue(m_game_vars->adv_nocrc);
	m_dont_report_stats->SetValue(m_game_vars->adv_noserversend);
	m_display_cmd_line->SetValue(m_game_vars->adv_display_cmd_line);

	return panel;
}

//////////////////////
