// main.h
// by Matt Ownby

#ifndef MAIN_H
#define MAIN_H

#include <wx/wx.h>
#include <wx/dialog.h>
#include <wx/notebook.h>
#include <wx/checkbox.h>
#include <wx/image.h>
#include <wx/url.h>
#include <wx/wfstream.h>
#include <wx/progdlg.h>
#include <wx/zstream.h>
#include <wx/memory.h>	// for debug

#include <string>
#include <vector>

using namespace std;

// this corresponds to the order of m_player_type
enum LDP_TYPES
{
	LDP_TYPE_UNKNOWN = 0,
	LDP_TYPE_VLDP = 1,
	LDP_TYPE_SERIAL = 2,
	LDP_TYPE_NONE = 3
};

// structure to hold elements that are different for each game on the list
struct game_vars
{
	bool adv_noserversend;
	bool adv_nocrc;
	bool adv_display_cmd_line;

	bool aud_nosound;
	unsigned int aud_buffer;

	int game_version_index;	// which version of this game we're playing (0 = auto, 1 = first item, etc)

	LDP_TYPES ldp_type;	// which laserdisc type we're using (see enums)
	unsigned int ldp_port;	// which serial port to use with physical laserdisc player
	bool vldp_autodownload;	// true if we are auto-downloading files for VLDP
	string vldp_framefile;	// location of VLDP framefile (if we're using VLDP)
	int vldp_download_index;	// index of the version we're getting (0 = auto, 1 = first item, etc)

	bool vid_blank_skips;
	bool vid_blank_searches;
	bool vid_blend;	// blend fields?
	bool vid_fullscreen;
	bool vid_nohwaccel;
	bool vid_scanlines;	// do scanlines?
	unsigned int vid_x, vid_y;	// size of window to create

	////////////////
	// variables that are populated dynamically, not loaded/saved

	string disc_name;	// arbitrary name of laserdisc ("astron", "lair", etc)
	string ldp_driver;	// which laserdisc driver we're using ("vldp", "noldp", "sony", etc)

};

struct conf_vars
{
	string conf_name;	// name of configuration profile
	vector <size_t> listbox_indices;	// the selected index on each listbox, sorted according to config file
};

// structure for holding all variables that we'll be sharing between classes
// I want some classes to have write access to this and others to have read-only access,
// which is why I've placed it in a struct
struct the_vars
{
	unsigned int game_index;	// which game has been selected (0 = first game on list)
	vector <game_vars> games;	// data for individual games
	vector <conf_vars> configs;	//	configurations that games can use
	unsigned int tab_index;	// which tab was last selected from config screen (0 = first)

	// global preferences here
	unsigned int ul_bytes_per_s;	// how many bytes we can upload per second (0 = unlimited)
	unsigned int dl_bytes_per_s;	// how many bytes we can download per second (0 = unlimited)
	
	// whether they have leech resistance enabled/disabled
	// 0 = auto, 1 = enabled, 2 = disabled (corresponds to order to dropdown list)
	int leech_resistance;
	
	unsigned int max_peers;	// max peers we will be connected to
	
	unsigned int bt_listen_port;	// port to listen for bit torrent connections on (default 6881)
	
};


class MyCanvas: public wxScrolledWindow
{
public:
    MyCanvas(wxWindow *parent, const wxPoint& pos, const wxSize& size);
	~MyCanvas();
	void new_image(const wxString &filename);
	void clear_image();
    void OnPaint(wxPaintEvent& event);
	DECLARE_EVENT_TABLE()
private:
	wxBitmap *m_bitmap;
	wxImage m_image;
	wxImageHandler *m_handler;
};

enum dl_status
{
		DLSTAT_ERROR = 0,	// couldn't even get a result due to a fatal error
		DLSTAT_FILE_MISSING,	// one or more of the files are simply not there
		DLSTAT_LENGTH_WRONG,	// one or more file lengths is wrong
		DLSTAT_FILE_UNVERIFIED,	// one or more of the files are there, but are unverified
		DLSTAT_CHECKSUM_WRONG,	// one or more of the files has a checksum mismatch
		DLSTAT_FILE_VERIFIED	// the files are there, and they are verified
};

// not part of a class because multiple classes may need to access it
bool spawn_executable(vector<string> cmd_line, bool wait, int *exit_code);

class XferFrame;

class MyFrame: public wxFrame
{
public:
    MyFrame(const wxString& title, const wxPoint& pos, const wxSize& size);
    void OnPreferences(wxCommandEvent &event);
    void OnClose(wxCloseEvent& event);
    void OnQuit(wxCommandEvent& event);
    void OnAbout(wxCommandEvent& event);
	void OnStart(wxCommandEvent &event);

	// Used for auto-downloading of laserdisc content.
	// returns a pointer to a struct in the control file that we've previously loaded
	// that shows which laserdisc version to try to autodownload/verify/launch.
	//  Based on which game is selected and which version index they've chosen
	// returns NULL if it can't find anything
	const struct ctrl_version *get_active_disc_version();

	// Returns a pointer to a struct in the config file that we've previous loaded
	// that shows which game version to launch.
	// 'gamever_index' is 0 if automatic, 1 for first entry, 2 for second entry, etc
	// Returns NULL if there is an error.
	const struct cfg_gamever *get_active_game_version(int gamever_index);

	// returns a pointer to the current configuration of the game version 
	// specified by 'gamever_index', which is 0 if automatic, 1 for first entry, 2 for second entry, etc
	const struct cfg_gamecfg *get_active_gamever_cfg(int gamever_index);

	// finds which version of ROMs defined in the control file correspond to
	// the version of ROMs that our build of Daphne supports
	// (returns NULL if it can't find a matching ROM version)
	const struct ctrl_version *get_rom_version();

	// returns 'game_vars' struct of actively selected game (according to game_index)
	const struct game_vars *get_active_gamevars();

	// returns 'conf_vars' struct that matches 'name'
	const struct conf_vars *get_confvars(const char *name);

	// helper for UpdateDependencies
	bool update_helper(const struct ctrl_version *v, const char *file_description);

	// checks dependencies and if allowed, tries to download any missing dependencies
	// returns true if dependencies are all met (and daphne will run without errors)
	// or false if dependencies are not met.
	// Also will pop up windows and dialog boxes informing the user of what's happening.
	bool UpdateDependencies();

	// prompts user to start a download of missing files.
	// if the download finishes, it prompts them to verify the files.
	// if verification succeeds, it returns true
	// otherwise, it returns false (including cancelation by user)
	// It also may advise user on how to resolve a problem
	// NOTE : this function both downloads and verifies because it gives different advice
	// to the user than prompt_verify does
	bool prompt_download_and_verify(const struct ctrl_version *v, const char *file_description);

	// Goes through all of the possible torrents indicated by 'v' and tries to download them
	// does md5 checking and if it passes, saves the file to disk
	// part of do_download, just separated so that function wasn't so long
	bool download_torrent(const struct ctrl_version *v);

	// does part of the work for prompt_download_and_verify
	// returns true if bit torrent was executed
	bool do_download(const struct ctrl_version *v);

	// prompts user to verify the files
	// if verification succeeds, it returns true
	// otherwise, it returns false (including cancelation by user)
	// It also may advise user on how to resolve a problem
	bool prompt_verify(const struct ctrl_version *v, const char *file_description);

	// does all the work for prompt_verify and prompt_download_and_verify
	// returns true if verification succeeds, or false if it fails or user cancels process
	bool do_verify(const struct ctrl_version *v);

	// converts 'challenge_digest' to ASCII,
	// then does a case-insensitive comparison with good_md5
	// returns 'true' if they match up
	bool compare_md5s(unsigned char *challenge_digest, string good_md5);

	// prints an error message box
	void do_error(const char *msg);

	void AddCmdLine(string s);
    void OnConfig(wxCommandEvent &event);
	void OnTransfers(wxCommandEvent &event);
	void XferFrameDying();	// helper function for OnTransfers

	void OnSelected(wxCommandEvent &event);
//	void OnNotebook(wxNotebookEvent &event);

	// functions for obtaining various access to our variables ...
	struct the_vars *get_readwrite_vars() { return m_the_vars; }
	const struct the_vars *get_readonly_vars() { return (const struct the_vars *) m_the_vars; }

	// downloads a file from host/uri with the message 'msg' into the buffer 'buf' and returns true if successful
	// host example = "www.cnn.com"
	// uri example = "/index.html"
	bool download_file(const char *url, string msg, string &buf);

	// helper function that makes a relative path from daphneloader.exe to a particular
	// subdirectory indicated by 'v'.  The '/' character is NOT appended to the end!
	// Example: could return "vldp/lair"
	string make_autodownload_path(const struct ctrl_version *v);

	// helper function that creates a *.md5 full file name including path
	string make_md5_fullpath(const struct ctrl_version *v, int which_file);

	// gets the status of the autodownload for just a single file (used by get_autodownload_status)
	dl_status get_status_one_file(const string autodownload_path, const struct ctrl_file *f);

	// gets the status of the autodownload for the version specified by 'v'
	// (see enum for possible return values)
	// returns DLSTAT_ERROR if v is NULL or empty
	dl_status get_autodownload_status(const struct ctrl_version *v);

	// downloads the latest control file, or loads it off local drive if downloading fails
	bool load_control_file();

	// loads the prefs.xml file, assumes config file is already loaded!!!
	bool load_prefs_file();

	// applies the loaded prefs from load_prefs_file
	void apply_the_vars();

	// pointer to the Transfers frame that can be displayed
	// (we want this public for easy access)
	XferFrame *m_xfer_frame;

    DECLARE_EVENT_TABLE()

private:
	vector<string> m_cmd_line;
	wxMenu *m_menuFile, *m_menuHelp, *m_menuGame;
	wxMenuBar *m_menuBar;
	wxPanel *m_panel;
	wxListBox *m_listbox;
	wxButton *m_start_button, *m_config_button;
	MyCanvas *m_canvas, *m_daphne_logo;

	wxBoxSizer *m_sizer_main;
	wxBoxSizer *m_sizer_right;
	wxBoxSizer *m_sizer_buttons;

	struct the_vars *m_the_vars;	// for sharing various variables that we need to share with other classes
};

struct ctrl_disc;	// just so we can use this before it has been defined
class MyPreConfig;	// declare before defining

class MyConfig: public wxNotebook
{
public:
	MyConfig(wxWindow *parent, wxWindowID id = -1, 
		const wxPoint& pos = wxDefaultPosition,
		const wxSize& size = wxDefaultSize,
		long style = 0);

	~MyConfig();

	void CreateInitialPages();

	// called after tabs have been created, and dialog box has been sized.. clears validators, sets start tab
	void Initialize();

	// if they change the LDP player type wxChoice
	void OnPlayerType(wxCommandEvent &event);

	// if they change the VLDP autodownload/manual wxChoice
	void OnVLDPChoices(wxCommandEvent &event);

	// if they click the 'browse' button
	void OnBrowseFramefile(wxCommandEvent &event);

	// if they change to a new tab ...
	void OnTabChange(wxNotebookEvent &event);

	// changes to a new tab on our config screen
	void ChangeTab(int which_tab);

	// called by OnTabChange and Initialize
	void HandleTabChange(int new_tab_index);

	// helper for apply_changes, also called by HandleGameVerSelection
	void apply_gamecfg_changes();

	// converts
	void apply_changes();

	// performs validation on the configuration options
	// returns true if configuration is valid (and enough stuff has been configured so a game can be played)
	bool validate(int &);

	// used to help us calculate the proper size of the notebooksizer
	void display_dynamic_possibility(int which_one);

	// if user changes to a new game version
	void OnGameVerType(wxCommandEvent &event);

private:
	// gets called by OnGameVerType or when the game page is created
	void HandleGameVerSelection(int new_ver_index);

	wxPanel *CreateGamePage();
	void resize(wxWindow *w, wxSizer *sz);
	void fix_size(wxWindow *w, unsigned int padding);
	wxStaticText *make_val(wxPanel *panel, const char *msg);
	wxPanel *CreateLaserdiscPage();
	wxPanel *CreateVideoPage();
	wxPanel *CreateAudioPage();
	wxPanel *CreateInputPage();
	wxPanel *CreateAdvancedPage();

	void HandlePlayerTypeSelection(int player_index);
	void ResetLaserdiscPageLayout();

	// if the vldp choice combox box is disturbed, this handles the new choice
	void HandleVldpChoice(int index);

	const struct the_vars *m_the_vars;	// read-only version of the vars
	const struct game_vars *m_game_vars;	// read-only version of the actively selected game vars

	////////////////////////////////////////////////////

	// Game Player Tab
	wxPanel *m_game_panel;	// this needs to be accessed from elsewhere
	wxChoice *m_gamever_choices;

	struct m_listbox_holder
	{
		wxStaticText *m_text;
		wxChoice *m_choice;
	};

	vector<m_listbox_holder> m_game_listboxes;
	wxBoxSizer *m_game_csizer;	// the main centering sizer
	wxBoxSizer *m_game_vsizer;	// the submain vertical sizer for the game page
	wxStaticBoxSizer *m_game_sbsizer;	// to separate game version drop-down from game options
	wxFlexGridSizer *m_game_gsizer;	// the flexgrid that all of the dynamic elements go into

	////////////////////////////////////////////////////

	// Laserdisc Player Tab
	wxChoice *m_player_type;
	wxStaticText *m_val_player_type;	// validator text for player type drop-down
	wxBoxSizer *m_centering_sizer;

	wxStaticText *m_none_text;
	wxBoxSizer *m_none_sizer;

	wxStaticBoxSizer *m_vldp_sizer;	// contains vldp-specific options inside, can be hidden when not in use
	wxBoxSizer *m_vldp_ff_sizer;	// manual install sizer
	wxBoxSizer *m_vldp_auto_sizer;	// automatic download sizer
	wxStaticBox *m_vldp_box;
	wxStaticText *m_vldp_choose_text;	// prompts them to choose auto-download or manual
	wxChoice *m_vldp_choices;	// auto-download or manual
	wxStaticText *m_val_vldp_choices;	// validator text for VLDP choices

	// VLDP auto download options
	wxStaticText *m_vldp_motd_header;	// the thing that explains this is a MOTD
	wxStaticText *m_vldp_motd;			// the actual MOTD itself

	wxStaticText * m_vldp_version_header;	// text that explains the version dropdown
	wxChoice *m_vldp_version;		// the version dropdown itself (which set of mpegs to download)

	// manual install VLDP options
	wxStaticText *m_vldp_framefile_text;
	wxBoxSizer *m_ldp_main_sizer;	// the main LDP sizer ...
	wxTextCtrl *m_vldp_framefile;
	wxStaticText *m_val_vldp_framefile;	// validator for framefile
	wxButton *m_vldp_browse;

	wxStaticBox *m_ldp_box;
	wxStaticBoxSizer *m_ldp_sizer;
	wxChoice *m_serial_port;
	wxChoice *m_baud_rate;

	///////////////////////////////////////////////////////

	// Video
	wxCheckBox *m_fullscreen;
	wxCheckBox *m_disable_hwaccel, *m_vldp_blank_on_skips, *m_vldp_blank_on_searches;
	wxChoice *m_screen_size;
	wxCheckBox *m_blend;
	wxCheckBox *m_scanlines;

	// Audio
	wxCheckBox *m_disable_sound, *m_sound_buffer_modify;

	// Advanced
	wxCheckBox *m_ld_latency_modify;
	wxCheckBox *m_dont_report_stats, *m_disable_crc_check, *m_display_cmd_line;
	wxTextCtrl *m_custom_settings;

	///////////////////////////////////////
	string m_ldp_driver;	// example "vldp"
	LDP_TYPES m_ldp_type;	// example "virtual"
	const ctrl_disc *m_disc;	// which disc we ended up with ...
	MyPreConfig *m_preconfig;	// point to parent class ...

	DECLARE_EVENT_TABLE()
};

class MyPreConfig : public wxDialog
{
public:
	MyPreConfig(wxWindow *parent);
	void RedoLayout();
    void OnConfigKeys(wxCommandEvent& event);
	void OnConfigButtons(wxCommandEvent& event);
    void OnOK(wxCommandEvent& event);
    void OnCancel(wxCommandEvent& event);
    void OnApply(wxCommandEvent& event);
private:
	MyConfig *m_config;
	wxButton *m_ok, *m_cancel, *m_apply;

	wxBoxSizer *m_sizer_main;
	wxBoxSizer *m_sizer_bottom_row;
	wxNotebookSizer *m_sizer_config;
	
	DECLARE_EVENT_TABLE()
};

class XferControl;

class MyApp: public wxApp
{
public:
    virtual bool OnInit();
	virtual int OnExit();
	
	// since wxApp::Yield seems to only work when called from this class
	void do_yield();
	
private:
	bool load_config(const char *filename);
	MyFrame *m_frame;
};

// wx enumerations
enum
{
    ID_Quit = 1,
    ID_Preferences,
	ID_About,
	ID_Web_Site,
	ID_Documentation,
	ID_List_Select,
	ID_Start,
	ID_Watch,
	ID_Config,
	ID_Transfers,
	ID_Notebook,
	ID_CFG_KEYS,
	ID_CFG_BUTTONS,
	ID_CFG_OK,
	ID_CFG_CANCEL,
	ID_CFG_APPLY,
	ID_GAMEVER_CHOICES,
	ID_PLAYER_TYPE,
	ID_VLDP_CHOICES,
	ID_VLDP_VERSION,
	ID_BROWSE_FRAMEFILE,

	// xfer frame enumerations
	ID_XferClose,
	ID_XferListCtrl,
	
	// prefs
	ID_GUIPREFS_APPLY,
	ID_GUIPREFS_CLOSE,
};

// macro to safely de-allocate memory
#define DEL(var) if (var) { delete(var); var = NULL; }

///

// so other modules can use it
extern MyFrame *g_mainframe;
extern MyApp *g_mainapp;
extern XferControl *g_xfer_control;

#endif	// MAIN_H
