// parse.cpp

#include "parse.h"
#include <sys/stat.h>
#include <mpolib/mpo_numstr.h>
#include <mpolib/mpo_misc.h>
#include <mpolib/mpo_xml.h>
#include <mpolib/mpo_fileio.h>

#include <ctype.h>	// linux, for isspace

#ifdef DEBUG
#include <assert.h>
#endif

const char *PREFS_VERSION = "prefs_ver_30";

//////////////////////////////////////////////////////////////////////

ini g_controlptr_ini;	// handles the control pointer .ini file
ctrl_main g_control_xml;	// handles the control XML file
cfg_main g_config_xml;	// handles the config XML file

//////////////////////////////////////////////////////////////////////

ini::ini()
: m_ini_loaded(false), m_buf(NULL), m_buf_size(0), m_buf_index(0)
{
}

ini::~ini()
{
	// de-allocate buffer if necessary
	if (m_buf)
	{
		delete [] m_buf;
		m_buf = NULL;
	}
}

void ini::use_buf(const unsigned char *buf, unsigned int buf_size)
{
	// use_buf can be called multiple times, so do a safety check
	if (m_buf)
	{
		delete [] m_buf;
		m_buf = NULL;
	}

	unsigned char *b = new unsigned char[buf_size];
	if (b)
	{
		memcpy(b, buf, buf_size);
		m_buf = b;
		m_buf_size = buf_size;
		m_buf_index = 0;
	}
}

void ini::load_ini(const char *ini_name)
{
	FILE *F;
	struct stat filestat;
	unsigned char *buf = NULL;

	F = fopen(ini_name, "rb");

	// if we were able to open the file
	if (F)
	{
		fstat(fileno(F), &filestat);	// get size of file
		buf = new unsigned char[filestat.st_size];	// allocate memory to hold file

		// if allocation is successful
		if (buf)
		{
			// if we are able to read in the whole file, then consider the INI loaded
			if (fread(buf, 1, filestat.st_size, F) == (size_t) filestat.st_size)
			{
				m_ini_loaded = true;
				m_buf = buf;
				m_buf_size = filestat.st_size;
			}
		}

		fclose(F);
	}
}

// returns true if the .INI file was loaded successfully
bool ini::ini_loaded()
{
	return m_ini_loaded;
}

// returns true if the key was found and if the index has been moved to the beginning of the line
// directly below the key
// returns false if the key wasn't found (and the index has not been changed)
bool ini::find_key(const char *keyname)
{
	unsigned int index = 0;
	bool result = false;

	// go through entire buffer looking for match
	while (index < m_buf_size)
	{
		// if we found a match, check to make sure that it's surrounded in []'s
		if (strncmp(keyname, (const char *) &m_buf[index], strlen(keyname)) == 0)
		{
			if ((index > 0) && (m_buf[index-1] == '[') && (m_buf[index + strlen(keyname)] == ']'))
			{
				m_buf_index = index;
				go_to_next_line();
				result = true;
			}
		}

		index++;
	}

	return result;
}

// moves m_buf_index to the next line
void ini::go_to_next_line()
{
	// chop off anything before the newline characters
	while ((m_buf[m_buf_index] != 0xA) && (m_buf[m_buf_index] != 0xD))
	{
		m_buf_index++;
	}

	// now chop off the newline chars
	while ((m_buf[m_buf_index] == 0xA) || (m_buf[m_buf_index] == 0xD))
	{
		m_buf_index++;
	}

	// we're not at the beginning of the next line
}

// returns a pointer to a string containing everything to the left of an equals character
const char *ini::get_left_of_equals()
{
	static string result;
	int src_index = m_buf_index;

	result = "";
	while (m_buf[src_index] != '=')
	{
		result += m_buf[src_index];
		src_index++;
	}

	go_to_next_line();

	remove_trailing_whitespace(result);
	return result.c_str();
}

// returns a pointer to a string containing everything to the right of an equals character
const char *ini::get_right_of_equals()
{
	static string result;
	int src_index = m_buf_index;

	result = "";

	// go to the = sign
	while (m_buf[src_index] != '=')
	{
		src_index++;
	}
	src_index++;	// move one character passed the equals sign

	// fill up result until we get to the end of the line
	while ((m_buf[src_index] != 0xD) && (m_buf[src_index] != 0xA))
	{
		result += m_buf[src_index];
		src_index++;
	}

	go_to_next_line();

	remove_trailing_whitespace(result);
	return result.c_str();
}

// returns the value of a variable (ie if the line is "Blah=Gwerk" then it would return Gwerk
// if you searched for Blah)
// returns an empty string ("") if it can't find the variable
// This does NOT change the current position within the .INI, so you can retrieve multiple vars without repositioning ...
string ini::get_var_value(const char *var)
{
	string result = "";
	unsigned int temp_buf_index = m_buf_index;	// this gets restored when we return ...

	// skip lines until we find the correct line
	// WARNING : this is dangerous if var does not exist at all
	while (strncmp((const char *) &m_buf[m_buf_index], var, strlen(var)) != 0)
	{
		go_to_next_line();
	}

	int src_index = m_buf_index;

	// go to the = sign
	while (m_buf[src_index] != '=')
	{
		src_index++;
	}
	src_index++;	// move one character passed the equals sign

	result = "";

	// fill up result until we get to the end of the line
	while ((m_buf[src_index] != 0xD) && (m_buf[src_index] != 0xA))
	{
		result += m_buf[src_index];
		src_index++;
	}

	go_to_next_line();

	// if we are returning any data
	if (result.length() > 0)
	{
		remove_trailing_whitespace(result);
	}

	m_buf_index = temp_buf_index;	// restore value ...

	return result;
}

// deletes trailing whitespace from a string
// This is so that string comparisons work correctly (the .INI file has at least one instance of trailing whitespace in it)
void ini::remove_trailing_whitespace(string &s)
{
	// while the end of the string is white space ...
	while ((s.length() > 0) && (isspace(s[s.length()-1])))
	{
		s.erase(s.length()-1, 1);	// erase last character from string ...
	}
}

////////////////////////////////////////////////////////////////////////////////

// XML CODE
// this simulates a class, but due to the callbacks used, we have to use global vars and
// global functions instead :(

XML_Parser g_xml_parser = NULL;
bool g_in_main = false;	// if we've passed the initial tag that says which version this control file is
bool g_in_disc = false;	// if we're in a disc subgroup
bool g_in_roms = false;	// if we're in a ROMS subgroup
bool g_in_version = false;	// if we're in a version subgroup
bool g_in_motd = false;	// if we're in a motd

void init_control()
{
		g_control_xml.m_discs.clear();	// empty vector so we can start from scratch
		g_control_xml.m_roms.clear();
}

bool parse_control_file(const unsigned char *buf, unsigned int size, string &err_msg)
{
	bool result = false;
	err_msg = "(empty message)";
	g_xml_parser = XML_ParserCreate(NULL);
	if (g_xml_parser)
	{
		XML_SetElementHandler(g_xml_parser, parse_element_start, parse_element_end);
		XML_SetCharacterDataHandler(g_xml_parser, parse_character_data);

		// reset all global flags
		g_in_main = g_in_disc = g_in_version = g_in_motd = false;
		init_control();

		// CHECKME: see if the last argument should, in fact, be a 1
		if (!XML_Parse(g_xml_parser, (const char *) buf, size, 1))
		{
			err_msg = "Parse error at line " +
				numstr::ToStr(XML_GetCurrentLineNumber(g_xml_parser));
			err_msg += ":\n";
			err_msg += XML_ErrorString(XML_GetErrorCode(g_xml_parser));
			init_control();
		}
		else
		{
			// if we have at least 1 disc in our control structure now
			if (g_control_xml.m_discs.size() > 0)
			{
				result = true;
			}

			// if we have no discs, we can assume that we never got inside the top element
			// because of a version mismatch
			else
			{
				err_msg = "DaphneLoader found an update, but it was empty.\n"
					"Perhaps you should upgrade to a newer version of DaphneLoader?";
			}
		}

		XML_ParserFree(g_xml_parser);
	}
	// else allocation failed
	else
	{
		err_msg = "Could not create parser!";
	}

	return result;
}

enum parse_type
{
	PARSE_FILE,
		PARSE_TORRENT,
		PARSE_VERSION,
		PARSE_DISC,
};

void parse_helper(parse_type type, const char **attr, void *object)
{
	ctrl_file *f = NULL;
	ctrl_torrent *t = NULL;
	ctrl_version *v = NULL;
	ctrl_disc *d = NULL;

	for (int i = 0; attr[i] != NULL; i += 2)
	{
		string at_s = mpom::str_toupper(attr[i]);	// case insensitivity
		const char *a = attr[i+1];	// for convenience

		switch (type)
		{
		case PARSE_FILE:
			f = (ctrl_file *) object;
			if (at_s == "NAME") f->m_name = a;	// file name
			else if (at_s == "MD5") f->m_md5 = a;	// md5 hash
			else if (at_s == "SIZE") f->m_size = a;	// size in bytes
			// else unknown
			break;
		case PARSE_TORRENT:
			t = (ctrl_torrent *) object;
			if (at_s == "URL") t->m_url = a;
			else if (at_s == "MD5") t->m_md5 = a;
			else if (at_s == "SIZE") t->m_size = a;
			// else unknown
			break;
		case PARSE_VERSION:
			v = (ctrl_version *) object;
			if (at_s == "NAME") v->m_name = a;
			else if (at_s == "DIR") v->m_dir = a;
			else if (at_s == "FRAMEFILE") v->m_framefile = a;
			else if (at_s == "ID") v->m_id = a;
			// else unknown
			break;
		case PARSE_DISC:
			d = (ctrl_disc *) object;
			if (at_s == "NAME") d->m_name = a;
			// else unknown
			break;
		}
	}
}

void parse_element_start(void *data, const char *el, const char **attr)
{
	string el_s = mpom::str_toupper(el);	// convert to uppercase for case-insensitive comparisons

	// if we've passed initial element
	if (g_in_main)
	{
		if (g_in_disc)
		{
			if (g_in_version)
			{
				if (el_s == "FILE")
				{
					ctrl_file f;
					f.m_md5 = f.m_name = f.m_size = "";
					parse_helper(PARSE_FILE, attr, &f);

					// convenience variables to prevent an even longer, uglier expression
					size_t d_index = g_control_xml.m_discs.size();
					size_t v_index = (g_control_xml.m_discs[d_index-1]).m_versions.size();

					// add this file structure to the file structure vector inside the last version of the current disc
					((g_control_xml.m_discs[d_index-1]).m_versions[v_index-1]).m_files.push_back(f);
				}
				else if (el_s == "TORRENT")
				{
					ctrl_torrent t;
					t.m_md5 = t.m_url = t.m_size = "";
					parse_helper(PARSE_TORRENT, attr, &t);

					// convenience variables to prevent an even longer, uglier expression
					size_t d_index = g_control_xml.m_discs.size();
					size_t v_index = (g_control_xml.m_discs[d_index-1]).m_versions.size();

					// add this torrent structure to the torrent structure vector inside the last version of the current disc
					((g_control_xml.m_discs[d_index-1]).m_versions[v_index-1]).m_torrents.push_back(t);
				}
				// else it's an unknown element, so we can ignore it
			}
			else if (el_s == "VERSION")
			{
				g_in_version = true;

				// initialize new version structure
				ctrl_version v;
				v.m_dir = "";
				v.m_rootdir = "vldp";	// since we're in the VLDP section
				v.m_files.clear();
				v.m_name = "";
				v.m_framefile = "";
				v.m_torrents.clear();
				parse_helper(PARSE_VERSION, attr, &v);

				// we can (safely?) assume that there will always be at least 1
				// disc in the vector when we make this call
				(g_control_xml.m_discs[g_control_xml.m_discs.size()-1]).m_versions.push_back(v);
			}
			else if (el_s == "MOTD")
			{
				g_in_motd = true;
				// the rest of the MOTD is handled in the character handler
			}
			// else it's an unknown element
		}
		// in roms structure
		else if (g_in_roms)
		{
			if (g_in_version)
			{
				size_t v_index = g_control_xml.m_roms.size();	// to prevent long ugly expressions

				if (el_s == "FILE")
				{
					ctrl_file f;
					f.m_md5 = f.m_name = f.m_size = "";
					parse_helper(PARSE_FILE, attr, &f);

					// add file					
					g_control_xml.m_roms[v_index-1].m_files.push_back(f);
				}
				else if (el_s == "TORRENT")
				{
					ctrl_torrent t;
					t.m_md5 = t.m_url = t.m_size = "";
					parse_helper(PARSE_TORRENT, attr, &t);

					// add torrent
					g_control_xml.m_roms[v_index-1].m_torrents.push_back(t);
				}
				// else it's an unknown element, so we can ignore it
			}
			else if (el_s == "VERSION")
			{
				// no attributes for ROM version, no need to call parse_helper
				g_in_version = true;

				ctrl_version v;
				v.m_dir = "roms";
				v.m_rootdir = ".";	// roms directory is off of daphne root
				v.m_files.clear();
				v.m_name = "";
				v.m_framefile = "";	// framefile does not apply
				v.m_torrents.clear();
				parse_helper(PARSE_VERSION, attr, &v);	// get version name

				// add new version to our roms vector
				g_control_xml.m_roms.push_back(v);
			}
		}
		// start of disc element
		else if (el_s == "DISC")
		{
			g_in_disc = true;

			// initialize new disc structure
			ctrl_disc d;
			d.m_motd = "";
			d.m_name = "";
			d.m_versions.clear();
			parse_helper(PARSE_DISC, attr, &d);
			g_control_xml.m_discs.push_back(d);	// add our new disc structure to the main one
		}
		// start of ROMS element
		else if (el_s == "ROMS") g_in_roms = true;

		// else it's an unknown element
	}
	// if it matches our version, we're in ...
	else if (el_s == "CONTROL_VER_100") g_in_main = true;

	// else it's an unknown element, so we can ignore it
}

void parse_element_end(void *data, const char *el)
{
	string el_s = mpom::str_toupper(el);

	// we don't need to check to see if these booleans are set to true, because the
	// parser will catch any syntax errors for us ...

	if (el_s == "VERSION") g_in_version = false;
	else if (el_s == "MOTD") g_in_motd = false;
	else if (el_s == "DISC") g_in_disc = false;
	else if (el_s == "ROMS") g_in_roms = false;
	else if (el_s == "CONTROL_VER_100") g_in_main = false;
}

void parse_character_data(void *data, const char *msg, int len)
{
	// if it's part of the Message Of The Day
	if (g_in_motd)
	{
		// if this is a DISC section's MOTD
		if (g_in_disc)
		{
			size_t d_index = g_control_xml.m_discs.size();
			(g_control_xml.m_discs[d_index-1]).m_motd.append(msg, len);	// add new MOTD content to our MOTD string
		}
		// else it's an unknown MOTD
	}
	// else ignore because it is unknown character data
}


// finds a disc in g_control_xml.m_discs that matches the name indicated by 'name'
// returns a pointer to that struct if successful, or NULL if no match was found
const ctrl_disc *control_find_disc(const ctrl_main *ctrl, string name)
{
	const ctrl_disc *result = NULL;
	name =  mpom::str_toupper(name);	// case insensitive

	for (size_t index = 0; index < ctrl->m_discs.size(); index++)
	{
		const ctrl_disc *tester = &ctrl->m_discs[index];
		if (mpom::str_toupper(tester->m_name) == name)
		{
			result = tester;
			break;
		}
	}

	return result;
}

bool control_download_available(const ctrl_disc *disc)
{
	bool result = false;

	// if there is at least 1 version and at least 1 torrent in that version
	if (disc && (disc->m_versions.size() > 0) && (disc->m_versions[0].m_torrents.size() > 0))
	{
		result = true;
	}
	return result;
}

///////////////////////////////////////////////////////////////////////

bool c_in_cfg = false;	//
bool c_in_gc = false;	// if we're inside <gamecfg> section
bool c_in_gc_cfg = false;	// if we're inside <gamecfg><cfg> section
bool c_in_gc_lb = false;	// if we're i nside <gamecfg><cfg><listbox>
bool c_in_gc_lb_item = false;	// if we're inside <gamecfg><cfg><listbox><item>
bool c_in_gamelist = false;
bool c_in_game = false;
bool c_in_gamever = false;
bool c_in_arg = false;
string c_arg_data = "";	// buffers all arg data we receive

void init_cfg()
{
	c_in_cfg = c_in_gamelist = c_in_game = c_in_gamever = c_in_arg = 
		c_in_gc = c_in_gc_cfg = c_in_gc_lb = c_in_gc_lb_item = false;
	c_arg_data = "";
	g_config_xml.m_games.clear();
	g_config_xml.m_games_cfg.clear();
	g_config_xml.m_rompack = "";
}

bool parse_cfg_file(const unsigned char *buf, unsigned int size, string &err_msg)
{
	bool result = false;
	err_msg = "(empty message)";
	g_xml_parser = XML_ParserCreate(NULL);
	if (g_xml_parser)
	{
		XML_SetElementHandler(g_xml_parser, cfg_el_start, cfg_el_end);
		XML_SetCharacterDataHandler(g_xml_parser, cfg_char_data);

		// reset all global flags
		init_cfg();

		// CHECKME: see if the last argument should, in fact, be a 1
		if (!XML_Parse(g_xml_parser, (const char *) buf, size, 1))
		{
			err_msg = "Parse error at line " +
				numstr::ToStr(XML_GetCurrentLineNumber(g_xml_parser));
			err_msg += ":\n";
			err_msg += XML_ErrorString(XML_GetErrorCode(g_xml_parser));
			init_cfg();
		}
		else
		{
			result = true;
		}

		XML_ParserFree(g_xml_parser);
	}
	// else allocation failed
	else
	{
		err_msg = "Could not create parser!";
	}

	return result;
}

enum cfg_type
{
	CFG_ROMPACK,
	CFG_GAME,
	CFG_GAMEVER,
	CFG_GC_CFG,	// <gamecfg><cfg>
	CFG_GC_LB,	// <gamecfg><cfg><listbox>
	CFG_GC_LB_ITEM	// <gamecfg><cfg><listbox><item>
};

void cfg_helper(cfg_type type, const char **attr, void *object)
{
	string *s = NULL;
	cfg_game *g = NULL;
	cfg_gamever *v = NULL;
	cfg_gamecfg *gc = NULL;
	cfg_listbox *lb = NULL;
	cfg_listitem *li = NULL;

	for (int i = 0; attr[i] != NULL; i += 2)
	{
		string at_s = mpom::str_toupper(attr[i]);	// case insensitivity
		const char *a = attr[i+1];	// for convenience

		switch (type)
		{
		case CFG_ROMPACK:
			s = (string *) object;
			if (at_s == "NAME") *s = a;	// file name
			// else unknown
			break;
		case CFG_GAME:
			g = (cfg_game *) object;
			if (at_s == "NAME") g->m_name = a;	// name
			else if (at_s == "DISC") g->m_disc = a;	// disc
			else if (at_s == "IMAGE") g->m_image = a;	// image
			else if (at_s == "ID") g->m_id = a;	// ID
			// else unknown
			break;
		case CFG_GAMEVER:
			v = (cfg_gamever *) object;
			if (at_s == "NAME") v->m_vername = a;	// name
			else if (at_s == "DRIVER") v->m_driver = a;	// driver
			else if (at_s == "CFG") v->m_cfg = a;	// which config to use
			else if (at_s == "ID") v->m_verid = a;	// version ID
		case CFG_GC_CFG:
			gc = (cfg_gamecfg *) object;
			if (at_s == "NAME") gc->m_name = a;	// the name of this config
			break;
		case CFG_GC_LB:
			lb = (cfg_listbox *) object;
			if (at_s == "TITLE") lb->m_title = a;
			else if (at_s == "ID") lb->m_id = a;
			else if (at_s == "TOOLTIP") lb->m_tooltip = a;
			break;
		case CFG_GC_LB_ITEM:
			li = (cfg_listitem *) object;
			if (at_s == "NAME") li->m_text = a;
			else if (at_s == "ID") li->m_id = a;
			break;
			// else unknown
		default:	// should never happen
			break;
		}
	}
}

void cfg_el_start(void *data, const char *el, const char **attr)
{
	string el_s = mpom::str_toupper(el);	// convert to uppercase for case-insensitive comparisons

	// convenience vars
	// none of these will be used if they are not > 0
	size_t c_index = g_config_xml.m_games_cfg.size() - 1;
	size_t game_index = g_config_xml.m_games.size() - 1;

	// if we've passed initial element
	if (c_in_cfg)
	{
		// if we're inside <gamelist> section
		if (c_in_gamelist)
		{
			if (c_in_game)
			{
				if (c_in_gamever)
				{
					if (el_s == "ARG")
					{
						c_in_arg = true;
					}
					// else unknown
				}
				else if (el_s == "VERSION")
				{
					cfg_gamever v;
					c_in_gamever = true;
					cfg_helper(CFG_GAMEVER, attr, &v);
					g_config_xml.m_games[game_index].m_versions.push_back(v);	// add new version
				}
				// else unknown
			}
			else if (el_s == "GAME")
			{
				cfg_game g;
				c_in_game = true;
				cfg_helper(CFG_GAME, attr, &g);
				g_config_xml.m_games.push_back(g);	// add new game entry
			}
			// else unknown
		}
		// <gamecfg>
		else if (c_in_gc)
		{
			// <gamecfg><cfg>
			if (c_in_gc_cfg)
			{
				// <gamecfg><cfg><listbox>
				if (c_in_gc_lb)
				{
					// <gamecfg><cfg><listbox><item>
					if (c_in_gc_lb_item)
					{
						if (el_s == "ARG")
						{
							c_in_arg = true;
						}
						// else unknown
					}
					else if (el_s == "ITEM")
					{
						c_in_gc_lb_item = true;
						cfg_listitem li;
						cfg_helper(CFG_GC_LB_ITEM, attr, &li);
						size_t l_index = g_config_xml.m_games_cfg[c_index].m_listboxes.size() - 1;	// must be > 0
						g_config_xml.m_games_cfg[c_index].m_listboxes[l_index].m_listitems.push_back(li);

						if (li.m_id == "")
						{
							string msg = "ListItem missing ID: ";
							msg += li.m_text;
							g_mainframe->do_error(msg.c_str());
						}
					}
				}
				else if (el_s == "LISTBOX")
				{
					cfg_listbox lb;
					c_in_gc_lb = true;
					cfg_helper(CFG_GC_LB, attr, &lb);
					// c_index must be >= 0
					g_config_xml.m_games_cfg[c_index].m_listboxes.push_back(lb);
				}
			}
			else if (el_s == "CFG")
			{
				cfg_gamecfg c;
				c_in_gc_cfg = true;
				cfg_helper(CFG_GC_CFG, attr, &c);
				g_config_xml.m_games_cfg.push_back(c);
			}
		}
		else if (el_s == "GAMELIST") c_in_gamelist = true;
		else if (el_s == "GAMECFG") c_in_gc = true;
		else if (el_s == "ROMPACK")	cfg_helper(CFG_ROMPACK, attr, &g_config_xml.m_rompack);
		// else unknown
	}
	// if it matches our version, we're in ...
	else if (el_s == "CONFIG_VER_2") c_in_cfg = true;
	// else unknown
}

void cfg_el_end(void *data, const char *el)
{
	string el_s = mpom::str_toupper(el);

	if (el_s == "GAMELIST") c_in_gamelist = false;
	else if (el_s == "CONFIG_VER_2") c_in_cfg = false;
	else if (el_s == "ARG")
	{
		// if we were in <gamelist><game><version>
		if (c_in_gamever)
		{
			size_t g_index = g_config_xml.m_games.size() - 1;	// size must be at least 1
			size_t v_index = g_config_xml.m_games[g_index].m_versions.size() - 1;	// " " "
			g_config_xml.m_games[g_index].m_versions[v_index].m_args.push_back(c_arg_data);
		}
		// else if we were in <gamecfg><cfg><listbox><item><arg>
		else if (c_in_gc_lb_item)
		{
			size_t c_index = g_config_xml.m_games_cfg.size() - 1;	// size must be at least 1
			size_t lb_index = g_config_xml.m_games_cfg[c_index].m_listboxes.size() - 1;	// " " "
			size_t li_index = g_config_xml.m_games_cfg[c_index].m_listboxes[lb_index].m_listitems.size() - 1;	// " " "
			g_config_xml.m_games_cfg[c_index].m_listboxes[lb_index].m_listitems[li_index].m_args.push_back(c_arg_data);
		}
		// else unknown place where ARG was used

		c_in_arg = false;
		c_arg_data = "";	// reset for next time
	}
	else if (el_s == "VERSION") c_in_gamever = false;
	else if (el_s == "GAME") c_in_game = false;
	else if ((c_in_gc_lb_item) && (el_s == "ITEM")) c_in_gc_lb_item = false;
	else if ((c_in_gc_lb) && (el_s == "LISTBOX")) c_in_gc_lb = false;
	else if ((c_in_gc_cfg) && (el_s == "CFG")) c_in_gc_cfg = false;
	else if ((c_in_gc) && (el_s == "GAMECFG")) c_in_gc = false;
	// else if it's ROMPACK, it doesn't matter
}

void cfg_char_data(void *data, const char *msg, int len)
{
	// if this is an argument we're receiving
	if (c_in_arg)
	{
		c_arg_data.append(msg, len);	// add to the end of it ...
	}
	// else it's unknown character data
}

///////////////////////////////////////////////////////////////////////
// PREFS here

prefs::prefs(the_vars *vars, const cfg_main *config, const ctrl_main *control)
{
	m_vars = vars;
	m_config = config;
	m_control = control;
}

bool prefs::parse(const unsigned char *buf, unsigned int size, string &err_msg)
{
	mpo_xml x;
	bool result = false;

	// if the initial parsing succeeded
	if (x.parse(buf, size, err_msg))
	{
		const xml_element *el = x.get_parsed_element();

		// check to make sure these are the correct preferences
		if (mpom::str_case_eq(el->name,PREFS_VERSION))
		{
			result = true;	// now we assume true unless we explicitly fail

			// go through children whilst we've not got any errors
			for (size_t i = 0; (i < el->children.size()) && result; i++)
			{
				string e = mpom::str_toupper(el->children[i].name);
				result = do_section(e.c_str(), &el->children[i], err_msg);
			}
		}
		else
		{
			err_msg = "Preference file version mismatch";
		}
	}
	// else it failed, and err_msg will say something

	return result;
}

// recursive function
bool prefs::do_section(const char *fulltag, const xml_element *el, string &err_msg)
{
	bool result = true;	// kind of hard to fail here ...
	size_t i = 0;

	string name = mpom::str_toupper(fulltag);

	// parse the game attributes as long as we've got no error
	for (i = 0; (i < el->attributes.size()) && result; i++)
	{
		struct game_vars *g = &m_vars->games[m_game_index];	// for convenience, this MUST come inside this 'for' loop!
		string a = mpom::str_toupper(el->attributes[i].name);	// for convenience
		string v = el->attributes[i].val;	// for convenience

		if (name == "GLOBALS")
		{
			if (a == "TAB")
			{
				m_vars->tab_index = numstr::ToUint32(v.c_str());

				// safety range check
				if (m_vars->tab_index > 5) m_vars->tab_index = 0;
			}						
		}
		// global bit torrent group section
		else if (name == "GLOBALS/BTG")
		{
			if (a == "BT_LISTEN_PORT")
			{
				m_vars->bt_listen_port = numstr::ToUint32(v.c_str());
				// this value will be validated by XferControl::apply_cur_vars
			}

			else if (a == "DL_BYTES_PER_S")
			{
				m_vars->dl_bytes_per_s = numstr::ToUint32(v.c_str());
				// this value will be validated by XferControl::apply_cur_vars
			}
			else if (a == "LEECH_RESISTANCE")
			{
				m_vars->leech_resistance = numstr::ToUint32(v.c_str());
				// this value will be validated by XferControl::apply_cur_vars
			}
			else if (a == "MAX_PEERS")
			{
				m_vars->max_peers = numstr::ToUint32(v.c_str());
				// this value will be validated by XferControl::apply_cur_vars
			}
			
			else if (a == "UL_BYTES_PER_S")
			{
				m_vars->ul_bytes_per_s = numstr::ToUint32(v.c_str());
				// this value will be validated by XferControl::apply_cur_vars
			}
			// else unknown, ignore
		}

		else if (name == "GAMES")
		{
			if (a == "SELECTED")
			{
				// try to find the game index that goes with this game ID
				if (find_game_index(el->attributes[i].val))
				{
					m_vars->game_index = m_game_index;	// we've found the game to select!
				}
				// else the default game shall be selected
			}
			// else unknown attribute
		}

		else if (name == "GAMES/GAME")
		{
			if (a == "ID")
			{
				// this sets the game index for all child recurses
				result = find_game_index(el->attributes[i].val);
				if (!result)
				{
					err_msg = "Error in prefs file. Cannot locate game section ";
					err_msg += el->attributes[i].val;
				}
				// if result is true, then m_game_index is now relevant
			}
			else if (a == "SELECTED")
			{
				// go through all games and see if we have a match
				for (size_t ver_index = 0; ver_index < m_config->m_games[m_game_index].m_versions.size(); ver_index++)
				{
					// if we've found the game in question
					if (mpom::str_case_eq(v,
						m_config->m_games[m_game_index].m_versions[ver_index].m_verid))
					{
						g->game_version_index = ver_index + 1;	// + 1 because 0 means automatic, 1 means first game, etc
						break;
					}
				}
				// if we didn't find the game, no error needs to be generated
			}
		}

		else if (name == "GAMES/GAME/LDP")
		{
			if (a == "TYPE")
			{
				if (v == "virtual") g->ldp_type = LDP_TYPE_VLDP;
				else if (v == "serial") g->ldp_type = LDP_TYPE_SERIAL;
				else if (v == "none") g->ldp_type = LDP_TYPE_NONE;
				else g->ldp_type = LDP_TYPE_UNKNOWN;
			}
			else if (a == "AUTODOWNLOAD") setone(v, g->vldp_autodownload);
			else if (a == "FRAMEFILE")
			{
				g->vldp_framefile = v;
			}
			else if (a == "DOWNLOAD_VER")
			{
				const ctrl_disc *d = control_find_disc(m_control, m_config->m_games[m_game_index].m_disc);

				if (d != NULL)
				{
					// go through all VLDP disc versions to see if we have a match
					for (size_t ver_index = 0; ver_index < d->m_versions.size(); ver_index++)
					{
						// if we've found the version and the disc in question
						if (mpom::str_case_eq(v,
							d->m_versions[ver_index].m_id))
						{
							g->vldp_download_index = ver_index + 1;	// + 1 because 0 means automatic, 1 means first version, etc
							break;
						}
					}
				}
				// else disc can't be found, ignore
			}
			else if (a == "PORT")
			{
				g->ldp_port = numstr::ToUint32(v.c_str());
			}
			else if (a == "DRIVER")
			{
				g->ldp_driver = v;
			}
			// else unknown
		}
		else if (name == "GAMES/GAME/VIDEO")
		{
			if (a == "FULLSCREEN") setone(v, g->vid_fullscreen);
			else if (a == "X") g->vid_x = numstr::ToUint32(v.c_str());
			else if (a == "Y") g->vid_y = numstr::ToUint32(v.c_str());
			else if (a == "NOHWACCEL") setone(v, g->vid_nohwaccel);
			else if (a == "BLANK_SEARCHES") setone(v, g->vid_blank_searches);
			else if (a == "BLANK_SKIPS") setone(v, g->vid_blank_skips);
			else if (a == "BLEND") setone(v, g->vid_blend);
			else if (a == "SCANLINES") setone(v, g->vid_scanlines);
			// else unknown
		}
		else if (name == "GAMES/GAME/AUDIO")
		{
			if (a == "NOSOUND") setone(v, g->aud_nosound);
			else if (a == "BUFFER") g->aud_buffer = numstr::ToUint32(v.c_str());
			// else unknown
		}
		else if (name == "GAMES/GAME/ADVANCED")
		{
			// NOTE : I want to move some of these advanced options into global prefs,
			// so I am ignoring them for now
		}

		// config section
		else if (name == "CFG")
		{
			if (a == "NAME") m_cfg_section = v;
		}

		// else unknown
	} // end parsing all attributes

	// TEXT-ONLY SECTION
	// item selected in cfg section
	if (name == "CFG/ITEMSELECTED")
	{
		// find config section, go through every listbox, go through every listitem
		for (size_t cfg_index = 0; cfg_index < m_config->m_games_cfg.size(); cfg_index++)
		{
			// if we've found the config section in question
			if (mpom::str_case_eq(m_config->m_games_cfg[cfg_index].m_name, m_cfg_section))
			{
				bool found = false;
				const cfg_gamecfg *gc = &m_config->m_games_cfg[cfg_index];
				// go through every listbox
				for (size_t lb_index = 0; (lb_index < gc->m_listboxes.size()) && (!found); lb_index++)
				{
					const cfg_listbox *lb = &gc->m_listboxes[lb_index];
					// and go through every listitem
					for (size_t li_index = 0; (li_index < lb->m_listitems.size()) && (!found); li_index++)
					{
						// if we found a match, then select it
						if (mpom::str_case_eq(lb->m_listitems[li_index].m_id, el->text))
						{
							// + 1 because 0 is automatic, 1 is first item, etc
							m_vars->configs[cfg_index].listbox_indices[lb_index] = li_index + 1;
							found = true;
						}
						// else keep going
					}
				}

				// a little debug message to help us out
				if (!found)
				{
					result = false;
					err_msg = "ListItem was not found: ";
					err_msg += el->text;
				}
				break;
			}
			// else keep going ...
		}
	} // end if CFG/ITEMSELECTED

	// parse all children as long as result is true
	for (i = 0; (i < el->children.size()) && result; i++)
	{
		string cname = mpom::str_toupper(el->children[i].name);
		result = do_section((name + "/" + cname).c_str(), &el->children[i], err_msg);
	}

	return result;
}

bool prefs::find_game_index(const string &game_id)
{
	bool result = false;

	// go through all games and see if we have a match
	for (m_game_index = 0; m_game_index < m_config->m_games.size(); m_game_index++)
	{
		// if we've found the game in question
		if (mpom::str_case_eq(game_id,
			m_config->m_games[m_game_index].m_id))
		{
			result = true;
			break;
		}
	}

	// if no match could be found, reset to safe default
	if (!result) m_game_index = 0;
	
	return result;
}

void prefs::setone(const string &s, bool &b)
{
	if (numstr::ToUint32(s.c_str()) == 1) b = true;
	else b = false;
}

bool prefs::save(const char *filename)
{
	bool result = false;
	string buf = "";	// the buffer we'll create our XML output in before saving
	size_t i = 0;	// general purpose

	buf = "<";
	buf += PREFS_VERSION;
	buf += ">\n";

	// <globals> section
	buf += "\t<globals tab=\"";
	buf += numstr::ToStr(m_vars->tab_index) + "\">\n";
	buf += "\t\t<btg ";
	buf += "ul_bytes_per_s=\"" + numstr::ToStr(m_vars->ul_bytes_per_s) + "\" ";
	buf += "dl_bytes_per_s=\"" + numstr::ToStr(m_vars->dl_bytes_per_s) + "\" ";
	buf += "leech_resistance=\"" + numstr::ToStr(m_vars->leech_resistance) + "\" ";
	buf += "max_peers=\"" + numstr::ToStr(m_vars->max_peers) + "\" ";
	buf += "bt_listen_port=\"" + numstr::ToStr(m_vars->bt_listen_port) + "\" ";
	buf += "/>\n";
	buf += "\t</globals>\n";
	// end <globals>

	// <games> section
	buf += "\t<games";

	// safety check, if m_games were ever empty (bad config file, for example) we don't want to
	//  crash
	if (!m_config->m_games.empty())
	{
		buf += " selected=\"";
		buf += m_config->m_games[m_vars->game_index].m_id;
		buf += "\"";
	}
	buf += ">\n";

	// do each game
	//  (if m_config->m_games is empty, it's ok, we won't crash)
	for (i = 0; i < m_config->m_games.size(); i++)
	{
		const game_vars *g = &m_vars->games[i];	// for convenience
		const cfg_game *c = &m_config->m_games[i];	// for convenience

		buf += "\t\t<game id=\"";
		buf += c->m_id + "\" ";

		// if a game version has been selected
		if (g->game_version_index > 0)
		{
			buf += "selected=\"";
			// -1 to adjust for automatic choice
			buf += c->m_versions[g->game_version_index-1].m_verid;
			buf += "\" ";
		}
		buf += ">\n";

		// GAME LDP
		buf += "\t\t\t<ldp type=\"";
		switch (g->ldp_type)
		{
		case LDP_TYPE_VLDP:
			buf += "virtual";
			break;
		case LDP_TYPE_SERIAL:
			buf += "serial";
			break;
		case LDP_TYPE_NONE:
			buf += "none";
			break;
		default:
			buf += "unknown";
			break;
		}
		buf += "\" autodownload=\"" + b2i(g->vldp_autodownload) + "\" ";
		buf += "framefile=\"" + g->vldp_framefile + "\" ";

		// if user has specified which version to get
		if (g->vldp_download_index > 0)
		{
			const ctrl_disc *d = control_find_disc(m_control, c->m_disc);
			// safety check, this should never fail
			if (d)
			{
				// -1 to adjust for 'auto' being at index 0
				buf += "download_ver=\"" + d->m_versions[g->vldp_download_index-1].m_id + "\" ";
			}
		}

		// ignored by VLDP, but we may as well add it so it is remembered in case they go back and forth
		buf += "port=\"" + numstr::ToStr(g->ldp_port) + "\" ";
		buf += "driver=\"" + g->ldp_driver + "\" ";

		buf += "/>\n";	// end LDP section

		// GAME VIDEO
		buf += "\t\t\t<video fullscreen=\"" + b2i(g->vid_fullscreen) + "\" ";
		buf += "x=\"" + numstr::ToStr(g->vid_x) + "\" ";
		buf += "y=\"" + numstr::ToStr(g->vid_y) + "\" ";
		buf += "nohwaccel=\"" + b2i(g->vid_nohwaccel) + "\" ";
		buf += "blank_searches=\"" + b2i(g->vid_blank_searches) + "\" ";
		buf += "blank_skips=\"" + b2i(g->vid_blank_skips) + "\" ";
		buf += "blend=\"" + b2i(g->vid_blend) + "\" ";
		buf += "scanlines=\"" + b2i(g->vid_scanlines) + "\" ";
		buf += "/>\n";	// end VIDEO

		// GAME AUDIO
		buf += "\t\t\t<audio ";
		buf += "nosound=\"" + b2i(g->aud_nosound) + "\" ";
		buf += "buffer=\"" + numstr::ToStr(g->aud_buffer) + "\" ";
		buf += "/>\n";	// end AUDIO

		// GAME ADVANCED
		// TODO : advanced!

		buf += "\t\t</game>\n";	// end game section
	} // end for loop for games

	buf += "\t</games>\n";	// end GAMES section

	// DO EACH GAME DYNAMIC CONFIG
	for (i = 0; i < m_config->m_games_cfg.size(); i++)
	{
		const cfg_gamecfg *c = &m_config->m_games_cfg[i];	// for convenience
		const conf_vars *cv = &m_vars->configs[i];	// for convenience

		buf += "\t<cfg name=\"" + c->m_name + "\">\n";

		// do all selected items
		for (size_t lb_index = 0; lb_index < c->m_listboxes.size(); lb_index++)
		{
			// if they haven't just chosen 'auto'
			if (cv->listbox_indices[lb_index] > 0)
			{
				buf += "\t\t<itemselected>";
				// -1 to adjust for index 0 being 'auto'
				buf += c->m_listboxes[lb_index].m_listitems[cv->listbox_indices[lb_index]-1].m_id;
				buf += "</itemselected>\n";
			}
		}

		buf += "\t</cfg>\n";	// end CONFIG
	}

	buf += "</";
	buf += PREFS_VERSION;
	buf += ">\n";	// end prefs file!

	// NOW WRITE FILE TO DISC!
	mpo_io *io = mpo_open(filename, MPO_OPEN_CREATE);

	if (io)
	{
		if (mpo_write(buf.data(), buf.size(), NULL, io))
		{
			result = true;
		}
		mpo_close(io);
	}

	return result;
}

string prefs::b2i(bool b)
{
	string result = "0";
	if (b) result = "1";
	return result;
}
