#ifndef PARSE_H
#define PARSE_H

// parse.h
// by Matt Ownby

#include <string>
#include <vector>
#include <expat.h>	// http://expat.sourceforge.net
#include <mpolib/mpo_xml.h>
#include "main.h"

using namespace std;

//////////////////////////////////////

// handles .INI-styled files
class ini
{
public:
	ini();
	~ini();
	void use_buf(const unsigned char *buf, unsigned int buf_size);
	void load_ini(const char *ini_name);
	bool ini_loaded();
	bool find_key(const char *);
	void go_to_next_line();
	const char *get_right_of_equals();
	const char *get_left_of_equals();
	string get_var_value(const char *var);

private:
	void remove_trailing_whitespace(string &s);

	bool m_ini_loaded;
	unsigned char *m_buf;
	unsigned int m_buf_size;
	unsigned int m_buf_index;
};

// START XML CONFIG STUFF

// handles multiple file entries
struct ctrl_file
{
	string m_name;
	string m_md5;
	string m_size;
};

// handles multiple torrent entries
struct ctrl_torrent
{
	string m_url;
	string m_md5;
	string m_size;
};

// Control file 'VERSION' area
struct ctrl_version
{
	vector <ctrl_file> m_files;
	vector <ctrl_torrent> m_torrents;
	string m_name;	// formal name of this version
	string m_id;	// unique ID of this version (this must never change!)
	string m_dir;		// the subdirectory under the rootdir (m_rootdir + /" + m_dir + "/" + files)
	string m_rootdir;	// for VLDP this will probably be "vldp", for ROMs this will be "roms"
	string m_framefile;	// the name of the framefile for this version (there can be only one)
};

// Control file 'DISC' area
struct ctrl_disc
{
	string m_name;
	string m_motd;
	vector <ctrl_version> m_versions;
};

// the main control structure to hold all of our control file configuration
struct ctrl_main
{
	vector <ctrl_disc> m_discs;
	vector <ctrl_version> m_roms;	// different rom versions for different daphne builds
};

////////////////////
bool parse_control_file(const unsigned char *buf, unsigned int size, string &err_msg);
void parse_element_start(void *data, const char *el, const char **attr);
void parse_element_end(void *data, const char *el);
void parse_character_data(void *data, const char *msg, int len);
const ctrl_disc *control_find_disc(const struct ctrl_main *ctrl, string name);

// returns true if a download is available for the indicated disc
bool control_download_available(const ctrl_disc *disc);

///////////////////////////

struct cfg_gamever
{
	string m_vername;	// the name of this version
	string m_verid;	// ID of this version (unique, must never change)
	string m_driver;	// which game driver this version uses ("lair", "cliff", "sdq", etc)
	string m_cfg;	// which config (cfg_gamecfg.m_name) this version uses
	vector <string> m_args;	// any additional arguments this version needs when launched
};

struct cfg_game
{
	string m_name;	// name of the game (formal title, can change)
	string m_id;	// ID of the game (unique, must never change)
	string m_disc;	// which disc this game uses (corresponds to control file)
	string m_image;	// which image to use to portray this game in the menu
	vector <cfg_gamever> m_versions;	// the versions of this game available
};

struct cfg_listitem
{
	string m_text;	// what this listitem says
	string m_id;	// unique ID for this list item
	vector<string> m_args;	// any additional arguments associated with this list item
};

struct cfg_listbox
{
	string m_title;	// title of the listbox (descriptive text)
	string m_id;	// unqiue identifier for this listbox
	string m_tooltip;	// an optional tooltip
	vector<cfg_listitem> m_listitems;
};

struct cfg_gamecfg
{
	string m_name;	// name of configuration (see cfg_gamever.m_cfg)
	vector <cfg_listbox> m_listboxes;	// listboxes containing options
};

struct cfg_main
{
	string m_rompack;	// name of rom pack
	vector <cfg_game> m_games;
	vector <cfg_gamecfg> m_games_cfg;
	// ADDME : serial LDP stuff
};

bool parse_cfg_file(const unsigned char *buf, unsigned int size, string &err_msg);
void cfg_el_start(void *data, const char *el, const char **attr);
void cfg_el_end(void *data, const char *el);
void cfg_char_data(void *data, const char *msg, int len);

//////////////////////

//	BEGIN XML PREFS STUFF

class prefs
{
public:
	prefs(the_vars *vars, const cfg_main *config_xml, const ctrl_main *control_xml);
	bool parse(const unsigned char *buf, unsigned int size, string &err_msg);

	// saves preferences to an XML file on disc
	bool save(const char *filename);
private:
	// 'game_index' should be 0 when first calling this function, fulltag should be "games"
	bool do_section(const char *fulltag, const xml_element *el, string &err_msg);

	bool find_game_index(const string &game);

	// helper function to convert string ints to bools
	void setone(const string &s, bool &b);

	// helper function to convert bool to "0" or "1"
	string b2i(bool b);

	size_t m_game_index;	// temp variable set by find_game_index
	string m_cfg_section;	// name of configuration section we're currently in

	the_vars *m_vars;
	const cfg_main *m_config;
	const ctrl_main *m_control;
};

//	END XML PREFS STUFF

// END XML CONFIG STUFF

/////////////////////////

// so main.cpp can conveniently access this
extern ini g_controlptr_ini;	// handles the control pointer .ini file
extern ctrl_main g_control_xml;	// handles the control XML file
extern cfg_main g_config_xml;	// handles the config XML file

#endif // PARSE_H
