// GUIPrefs.cpp by Matt Ownby

#include "prefs.h"

BEGIN_EVENT_TABLE(GUIPrefs, wxDialog)
EVT_BUTTON(ID_GUIPREFS_APPLY, GUIPrefs::OnApply)
EVT_BUTTON(ID_GUIPREFS_CLOSE, GUIPrefs::OnClose)
END_EVENT_TABLE()

GUIPrefs::GUIPrefs(wxWindow *parent)
: wxDialog(parent, -1, wxString(_T("Preferences")))
{
	m_panel = new wxPanel(this);	// create panel
	
	m_gsizer = new wxFlexGridSizer(6, 4, 5, 5);	// rows, cols, vgap, hgap
	
	// line 1
	AddTextFlex(NULL, "");
	AddTextFlex(NULL, "New Value");
	AddTextFlex(NULL, "Active Value");
	AddTextFlex(NULL, "Recommended Value");
	
	// line 2
	AddTextFlex(NULL, "Max Upload Bytes/Sec");
	AddInputFlex(&m_txtMaxUlBytes, "?");
	AddTextFlex(&m_lblMaxUlBytes, "?");	// actual value
	AddTextFlex(NULL, "80% of up capacity (0 means no limit)");
		
	AddTextFlex(NULL, "Max Download Bytes/Sec");
	AddInputFlex(&m_txtMaxDlBytes, "?");
	AddTextFlex(&m_lblMaxDlBytes, "?");	// actual value
	AddTextFlex(NULL, "0 (0 means no limit)");
	
	AddTextFlex(NULL, "Leech Resistance");
	wxString leech_choices[] = 
	{
		"Auto",
		"Enabled",
		"Disabled"
	};
	int choices_count = sizeof(leech_choices) / sizeof(wxString);
	m_lstLeechRes = new wxChoice(m_panel, -1, wxDefaultPosition, wxDefaultSize, choices_count, leech_choices);
	m_gsizer->Add(m_lstLeechRes, 0, wxEXPAND, 0);
	AddTextFlex(&m_lblLeechRes, "?");	// actual value
	AddTextFlex(NULL, "Auto");

	AddTextFlex(NULL, "Max Peers");
	AddInputFlex(&m_txtMaxPeers, "?");
	AddTextFlex(&m_lblMaxPeers, "?");	// actual value
	AddTextFlex(NULL, "50");

	AddTextFlex(NULL, "BitTorrent Listen Port");
	AddInputFlex(&m_txtListenPort, "?");
	AddTextFlex(&m_lblListenPort, "?");	// actual value
	AddTextFlex(NULL, "0 (means to try 6881-6889)");
	
	m_btnApply = new wxButton(m_panel, ID_GUIPREFS_APPLY, "Apply");
	m_btnClose = new wxButton(m_panel, ID_GUIPREFS_CLOSE, "Close");
	
	m_btnsizer = new wxBoxSizer(wxHORIZONTAL);

	m_btnsizer->Add(m_btnApply, 0, wxALL, 4);
	m_btnsizer->Add(m_btnClose, 0, wxALL, 4);
	
	// MAIN SIZERS
	m_vsizer = new wxBoxSizer(wxVERTICAL);
	m_vsizer->Add(m_gsizer, 0, wxLEFT | wxTOP, 10);
	m_vsizer->Add(m_btnsizer, 0, wxALIGN_CENTER | wxTOP, 10);

	m_panel->SetSizer(m_vsizer);

	m_panel->Fit();	// causes contents to fit together
	this->Fit();	// causes frame window to fit to smallest size around contents

	refresh_controls();

}

void GUIPrefs::OnClose(wxCommandEvent &event)
{
	EndModal(0);
}

void GUIPrefs::OnApply(wxCommandEvent &event)
{
	apply_changes();
	refresh_controls();
}

// helper function
void GUIPrefs::AddTextFlex(wxStaticText **txt, const char *s)
{
	wxStaticText *text = NULL;
	
	text = new wxStaticText(m_panel, -1, _T(s));
	m_gsizer->Add(text, 0, wxEXPAND, 0);

	// if we want to hang onto this ptr
	if (txt != NULL) *txt = text;
}

void GUIPrefs::AddInputFlex(wxTextCtrl **inp, const char *s)
{
	wxTextCtrl *txt = NULL;
	
	txt = new wxTextCtrl(m_panel, -1, s);	// this will be part of an expanding sizer so we need not define a width
	m_gsizer->Add(txt, 0, wxEXPAND, 0);
	
	// if we want to save this pointer
	if (inp != NULL) *inp = txt;
}

void GUIPrefs::refresh_controls()
{
	// we will only be reading from these, so we grab the readonly version to minimize bugs
	const struct the_vars *v = g_mainframe->get_readonly_vars();
	
	// max up bytes
	refresh_helper(v->ul_bytes_per_s, m_lblMaxUlBytes, m_txtMaxUlBytes, true);

	// max dl bytes
	refresh_helper(v->dl_bytes_per_s, m_lblMaxDlBytes, m_txtMaxDlBytes, true);
		
	// leech resistance
	// NOTE : we don't check to see if leech_resistance is out of range, because that check should be done any time
	// leech_resistance is modified.
	m_lstLeechRes->SetSelection(v->leech_resistance);
	m_lblLeechRes->SetLabel(m_lstLeechRes->GetStringSelection());	// just mirror what is printed

	// max peers
	refresh_helper(v->max_peers, m_lblMaxPeers, m_txtMaxPeers, false);

	// listen port
	refresh_helper(v->bt_listen_port, m_lblListenPort, m_txtListenPort, false);

	m_vsizer->Layout();	// this is supposed to update the controls to fit, but having ploblems on win32
}

void GUIPrefs::refresh_helper(unsigned int val, wxStaticText *lbl, wxTextCtrl *txt, bool xfer)
{
	string s = "";	// temp string
	
	// max up bytes
	s = numstr::ToStr(val);
	txt->SetValue(s.c_str());

	// if this is a transfer speed, then we can show the kb/s or mb/s or whatever equivalent for user convenience	
	if (xfer)
	{
		// if it's 0, we want to notify the user that this means there is no limit
		if (val == 0) s += " (unlimited)";
		// if they've actually provided an amount, we want to convert it to an indicative value so the user realizes this is bytes
		//  per second, not kilobytes or anything else
		else
		{
			s += " (" + numstr::ToUnitStr(val) + "/s)";
		}
	}
	lbl->SetLabel(s.c_str());
}

void GUIPrefs::apply_changes()
{
	struct the_vars *v = g_mainframe->get_readwrite_vars();	// we'll be making changes here, so we get read/write version
	
	// another function will handle the majority of the validation here, so we need not do much
	v->ul_bytes_per_s = numstr::ToUint32(m_txtMaxUlBytes->GetValue().c_str());
	v->dl_bytes_per_s = numstr::ToUint32(m_txtMaxDlBytes->GetValue().c_str());
	v->leech_resistance = m_lstLeechRes->GetSelection();
	v->max_peers = numstr::ToUint32(m_txtMaxPeers->GetValue().c_str());
	v->bt_listen_port = numstr::ToUint32(m_txtListenPort->GetValue().c_str());
	
	// do validation and make changes to xfer module
	g_xfer_control->apply_cur_vars();
}
